/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.membranepackingalgorithm;

import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import org.cellmicrocosmos.cm2.constants.CM2_CONSTANTS;
import org.cellmicrocosmos.cm2.dataobjects.Atom_3D;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.dataobjects.Particle;


/**
 * This class is a wrapper for an Atom that is contained by a molecule.
 * Represents also possible coarse-grain-particles so that they can be treated
 * like any other atom
 * 
 * @author Tim Dingersen dackelnose@gmx.net
 * 
 */
public class Atom {

	private Atom_3D originalAtom;
	private Molecule parentMolecule;
	public Point3f position;
	private int index;
	public Bond[] bonds;
	private MembranePackingAlgorithm algo;

	/**
	 * !NOT MEANT TO BE USED BY ALGORITHM DEVELOPERS!
	 */
	public Atom(Atom_3D originalAtom, Molecule parentMolecule, int index,
			MembranePackingAlgorithm algo) {
		this.originalAtom = originalAtom;
		this.parentMolecule = parentMolecule;
		this.index = index;
		this.algo = algo;
	}

	/**
	 * Sets and applies a new position to this atom. The visual position will be
	 * automatically updated
	 * 
	 * @param position
	 *            the new absolute position of the atom, measured from the
	 *            molecule's center point
	 */
	public void setPosition(Point3f position) {

		algo.dock();

		Point3f prev = this.parentMolecule.getOriginalMolecule().atoms[index]
				.getPosition();

		this.parentMolecule.getOriginalMolecule().applyAtomPosition(index,
				position);

		this.position = position;
		this.getParentalMolecule().isUpdated = false;

		if (this.parentMolecule.getOriginalMolecule().getRenderMode() == Molecule_3D.RENDER_MODE_COVALENT
				|| this.parentMolecule.getOriginalMolecule().getRenderMode() == Molecule_3D.RENDER_MODE_BALLS_N_STICKS
				|| this.parentMolecule.getOriginalMolecule().getRenderMode() == Molecule_3D.RENDER_MODE_VAN_DER_WAALS) {

			if (this.parentMolecule.getOriginalMolecule().atoms[index].branchGroup != null) {

				for (int j = 0; j < this.parentMolecule.getOriginalMolecule().atoms[index].branchGroup
						.numChildren(); j++) {
					final TransformGroup tg = (TransformGroup) this.parentMolecule
							.getOriginalMolecule().atoms[index].branchGroup
							.getChild(j);
					final Transform3D t3d = new Transform3D();
					final Vector3f v = new Vector3f(this.parentMolecule
							.getOriginalMolecule().atoms[index].position);
					v.scale(CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE);
					t3d.setTranslation(v);
					tg.setTransform(t3d);
				}
			}

		}

		if (this.originalAtom instanceof Particle) {

			Atom_3D[] toMove = ((Particle) this.originalAtom).includedAtoms
					.toArray(new Atom_3D[0]);

			Point3f move = this.parentMolecule.getOriginalMolecule().atoms[index]
					.getPosition();

			move.sub(prev);
			for (int i = 0; i < toMove.length; i++) {

				toMove[i].position.add(move);

			}

		}

	}

	/**
	 * Gets a copy of the atom's current position inside the molecule
	 * 
	 * @return a copy of the position field
	 */
	public Point3f getPosition() {
		return new Point3f(this.position);
	}

	/**
	 * Returns the molecule instance that hosts this atom
	 * 
	 * @return the molecule instance
	 */
	public Molecule getParentalMolecule() {
		return this.parentMolecule;
	}

	/**
	 * Returns the record name of this atom's pdb-file entry
	 * 
	 * @return the record name
	 */
	public String getRecordName() {
		return this.originalAtom.recordName;
	}

	/**
	 * Returns the name of this atom in its pdb-file entry
	 * 
	 * @return the atom name
	 */
	public String getAtomName() {
		return this.originalAtom.name;
	}

	/**
	 * Returns the name of the atom's residue as specifed in its pdb-file entry
	 * 
	 * @return the residue name
	 */
	public String getResidueName() {
		return this.originalAtom.residueName;
	}

	/**
	 * Returns the id of the atom's peptide chain as specifed in its pdb-file
	 * entry
	 * 
	 * @return the identifier of the specified chain in the pdb file
	 */
	public String getChainID() {
		return this.originalAtom.chainID;
	}

	/**
	 * Returns the field occupancy as specifed in the pdb-file entry
	 * 
	 * @return the occupancy string
	 */
	public String getOccupancy() {
		return this.originalAtom.occupancy;
	}

	/**
	 * Returns the field tempfactor as specifed in the pdb-file entry
	 * 
	 * @return the temp factor string
	 */
	public String getTempFactor() {
		return this.originalAtom.tempFactor;
	}

	/**
	 * Returns the field element as specifed in the pdb-file entry
	 * 
	 * @return the element sysmbol string
	 */
	public String getElement() {
		return this.originalAtom.element;
	}

	/**
	 * Returns the fieldcharge as specifed in the pdb-file entry
	 * 
	 * @return the charge string
	 */
	public String getCharge() {
		return this.originalAtom.charge;
	}

	/**
	 * Returns the determined covalent radius of this atom
	 * 
	 * @return the covalent radius
	 */
	public float getCovalentRadiusInAngstrom() {
		return this.originalAtom.radiusCov;
	}

	/**
	 * Returns the determined Van-der-Waals-radius of this atom
	 * 
	 * @return the Van-der-Waals-radius
	 */
	public float getVanDerWaalsRadiusInAngstrom() {
		return this.originalAtom.radiusVDW;
	}

	/**
	 * Returns a list of all bond objects associated with this atom. Each bond
	 * represents a bonded neighbor of this atom
	 * 
	 * @return the list of all bond objects of this atom
	 */
	public Bond[] getAllInternalBonds() {
		return this.bonds;
	}

	/**
	 * Determines, wether this atom is actually a coarse-grain-particle
	 * 
	 * @return true if this atom is a coarse-grain-particle, false if it is an
	 *         usual atom
	 */
	public boolean isCoarseGrainParticle() {
		return this.originalAtom instanceof Particle;
	}

	/**
	 * Returns the particle identifier of this particle
	 * 
	 * @return the particle identifier or null if this is an usual atom
	 */
	public String getParticleID() {
		if (this.originalAtom instanceof Particle) {
			return ((Particle) this.originalAtom).particleIdentifier;
		} else
			return null;
	}

	/**
	 * Returns the string that serves as a possible key for parameters.
	 * Particles return their particle identifier here while usual atoms return
	 * their element-string
	 * 
	 * @return the particle id if this atom is a particle, the element-string if
	 *         it is an usual atom
	 */
	public String getParameterKey() {
		String ret = null;
		if (this.originalAtom instanceof Particle)

			ret = ((Particle) this.originalAtom).particleIdentifier;
		else
			ret = this.originalAtom.element;

		if (ret == null || ret.length() < 1)
			return null;
		return ret;
	}

	private double energy = 0;

	private Object userData;

	/**
	 * Can be used to store an energy value for this atom
	 * 
	 * @param en
	 *            the energy value to store
	 */
	public void setEnergy(double en) {
		this.energy = en;

	}

	/**
	 * Returns the stored energy value
	 * 
	 * @return the stored energy value
	 */
	public double getEnergy() {
		return this.energy;
	}

	/**
	 * Can be used to store any Object to associate it with this atom
	 * 
	 * @param userData
	 *            any Object
	 */
	public void setUserData(Object userData) {
		this.userData = userData;

	}

	/**
	 * Returns the object that has been stored by setUserData()
	 * 
	 * @return the stored object
	 */
	public Object getUserData() {
		return this.userData;
	}

	/**
	 * Returns the atomic mass of this atom. In case the atom is a particle, it
	 * will return the sum of all atomic masses of the hidden atoms
	 * 
	 * @return the mass value of this atom
	 */
	public float getAtomicMass() {
		if (this.isCoarseGrainParticle()) {
			return ((Particle) this.originalAtom).getMassValue();
		} else {
			return this.originalAtom.massValue;
		}

	}

}
