/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.coarsegrainsupport;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Vector;

import javax.vecmath.Point3f;

import org.cellmicrocosmos.cm2.constants.CM2_CONSTANTS;
import org.cellmicrocosmos.cm2.dataobjects.Atom_3D;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.dataobjects.MyBounds;
import org.cellmicrocosmos.cm2.dataobjects.Particle;
import org.cellmicrocosmos.cm2.engine.DataPool;
import org.cellmicrocosmos.cm2.io.PdbReader_3D;
import org.cellmicrocosmos.cm2.misc.Cm2_Toolkit;


/**
 * Manages all Input / Output operations for Molecule Profiles and is generely responsible for their creation 
 * @author Tim Dingersen
 * dackelnose@gmx.net
 *
 */
public class Molecule_Profile_IO {

	public static File[] getAllProfiles(Molecule_3D mol, String pdbCode) {
		String modelext = "";
		if (mol.getModelSet() > 0) {
			modelext = File.separator + "Model_" + mol.getModelSet();
		}
		File dir = new File(DataPool.workspace.getAbsolutePath()
				+ File.separator
				+ CM2_CONSTANTS.IO.MOLECULE_PROFILES_DIR.getPath()
				+ File.separator + pdbCode.toUpperCase() + modelext);
		return dir.listFiles();
	}

	public static String[] getAllProfileNames(Molecule_3D mol, String pdbCode) {

		File[] dir = Molecule_Profile_IO.getAllProfiles(mol, pdbCode);
		if (dir == null)
			return null;
		String[] ret = new String[dir.length];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = Cm2_Toolkit.getSimpleName(dir[i]);
		}
		return ret;
	}

	public static boolean existsProfile(Molecule_3D mol, String name,
			String pdbCode) {
		String modelext = "";
		if (mol.getModelSet() > 0) {
			modelext = File.separator + "Model_" + mol.getModelSet();
		}
		File dir = new File(DataPool.workspace.getAbsolutePath()
				+ File.separator
				+ CM2_CONSTANTS.IO.MOLECULE_PROFILES_DIR.getPath()
				+ File.separator + pdbCode.toUpperCase() + modelext);
		File[] files = dir.listFiles();

		for (int i = 0; i < files.length; i++) {
			if (files[i].getName().equals(name))
				return true;
		}
		return false;
	}

	public static void saveProfile(Molecule_3D mol, String name)
			throws IOException {

		String modelext = "";
		if (mol.getModelSet() > 0) {
			modelext = File.separator + "Model_" + mol.getModelSet();
		}
		File dir = new File(DataPool.workspace.getAbsolutePath()
				+ File.separator
				+ CM2_CONSTANTS.IO.MOLECULE_PROFILES_DIR.getPath()
				+ File.separator + mol.getPDBcode().toUpperCase() + modelext);
		if (!dir.exists()) {
			dir.mkdirs();
		}

		File file = new File(dir.getAbsolutePath() + File.separator + name
				+ ".cm2mp");

		Vector<Atom_3D> atoms = new Vector<Atom_3D>();
		// Vector<Constraint>constraints = new Vector<Constraint>();

		BufferedWriter bw = new BufferedWriter(new FileWriter(file));
		bw.write("POSITIONS");
		bw.newLine();
		for (int i = 0; i < mol.allAtomsSorted.length; i++) {
			Atom_3D atom = mol.allAtomsSorted[i];
			atoms.addElement(atom);
			bw.write(atom.position.x + " " + atom.position.y + " "
					+ atom.position.z);
			bw.newLine();
		}
		bw.write("/POSITIONS");
		bw.newLine();

		for (int i = 0; i < mol.atoms.length; i++) {
			if (mol.atoms[i] instanceof Particle) {

				Particle p = (Particle) mol.atoms[i];
				bw.write("PARTICLE");
				bw.newLine();
				bw.write(p.particleIdentifier);
				bw.newLine();
				bw.write(p.name);
				bw.newLine();
				bw.write(p.element);
				bw.newLine();

				for (int j = 0; j < p.includedAtoms.size(); j++) {
					bw.write(atoms.indexOf(p.includedAtoms.get(j)) + " ");
				}
				bw.newLine();
				bw.write(p.color.getRed() + " " + p.color.getGreen() + " "
						+ p.color.getBlue());
				bw.newLine();
				bw.write("/PARTICLE");
				bw.newLine();

			}
		}

		// bw.write("BONDS");
		// bw.newLine();
		// for(int i=0;i<mol.activeConstraints.size();i++) {
		// Constraint c = mol.activeConstraints.get(i);
		// bw.write(atoms.indexOf(c.atomA)+" "+atoms.indexOf(c.atomB));
		// bw.newLine();
		// }
		// bw.write("/BONDS");
		// bw.newLine();

		bw.close();
	}

	public static MoleculeProfile readProfile(Molecule_3D mol, String name,
			String code) throws Exception {
		if (name.equals("Default Profile"))
			return Molecule_Profile_IO.readDefaultProfile(mol);

		String modelext = "";
		if (mol.getModelSet() > 0) {
			modelext = File.separator + "Model_" + mol.getModelSet();
		}

		File dir = new File(DataPool.workspace.getAbsolutePath()
				+ File.separator
				+ CM2_CONSTANTS.IO.MOLECULE_PROFILES_DIR.getPath()
				+ File.separator + code.toUpperCase() + modelext);

		File file = new File(dir.getAbsolutePath() + File.separator + name
				+ ".cm2mp");

		Vector<Point3f> positions = new Vector<Point3f>();
		Vector<String> particleIDs = new Vector<String>();
		Vector<String> particleAtomNames = new Vector<String>();
		Vector<Color> colors = new Vector<Color>();
		Vector<String> particleElementNames = new Vector<String>();
		Vector<Integer[]> particleMemberIndices = new Vector<Integer[]>();

		BufferedReader br = new BufferedReader(new FileReader(file));

		while (true) {
			String s = br.readLine();
			if (s == null)
				break;

			if (s.trim().equals("POSITIONS")) {
				while (true) {
					s = br.readLine();
					if (s.trim().equals("/POSITIONS"))
						break;

					String[] numbers = s.split(" ");
					try {
						positions.addElement(new Point3f(Float
								.parseFloat(numbers[0].trim()), Float
								.parseFloat(numbers[1].trim()), Float
								.parseFloat(numbers[2].trim())));
					} catch (Exception e) {
						continue;
					}
				}
			} else if (s.trim().equals("PARTICLE")) {
				try {
					String id = br.readLine().trim();
					String aname = br.readLine().trim();
					String aelement = br.readLine().trim();
					String[] indicesAsString = br.readLine().split(" ");
					Integer[] indices = new Integer[indicesAsString.length];
					for (int i = 0; i < indicesAsString.length; i++) {
						indices[i] = Integer.parseInt(indicesAsString[i]);
					}

					Color color;
					try {
						String[] cs = br.readLine().trim().split(" ");
						color = new Color(Integer.parseInt(cs[0]), Integer
								.parseInt(cs[1]), Integer.parseInt(cs[2]));

					} catch (Exception e) {
						color = Color.gray;
					}

					particleIDs.addElement(id);
					colors.addElement(color);
					particleAtomNames.addElement(aname);
					particleElementNames.addElement(aelement);
					particleMemberIndices.addElement(indices);
				} catch (Exception e) {
					continue;
				}

			}

		}

		if (mol.allAtomsSorted.length != positions.size())
			throw new Exception(
					"Molecule and Profile have different Atom Counts");

		br.close();

		MoleculeProfile mp = new MoleculeProfile();
		mp.mol = mol;
		mp.name = name;
		mp.pdbCode = mol.getPDBcode();
		mp.positions = positions.toArray(new Point3f[0]);
		mp.particleIDs = particleIDs.toArray(new String[0]);
		mp.colors = colors.toArray(new Color[0]);
		mp.particleAtomNames = particleAtomNames.toArray(new String[0]);
		mp.particleElementNames = particleAtomNames.toArray(new String[0]);

		mp.particleMemberIndices = particleMemberIndices
				.toArray(new Integer[0][]);

		return mp;

	}

	public static MoleculeProfile readDefaultProfile(Molecule_3D mol)
			throws Exception {

		PdbReader_3D reader = new PdbReader_3D();
		String[] pdbFile = DataPool.getRawPdbData(mol.getPDBFile()
				.getAbsolutePath());

		String[][] contents = reader.readModel(pdbFile, mol.getModelSet());

		final Vector<Point3f> allPositions = new Vector<Point3f>();

		for (int i = 0; i < contents[0].length; i++) {

			try {

				final Atom_3D atom = new Atom_3D(contents[0][i]);

				allPositions.addElement(atom.position);

			} catch (final Exception e) {

			}

		}

		MoleculeProfile mp = new MoleculeProfile();
		mp.mol = mol;
		mp.name = "Default Profile";
		mp.pdbCode = mol.getPDBcode();
		mp.positions = allPositions.toArray(new Point3f[0]);
		MyBounds mb = new MyBounds(mp.positions);
		for (int i = 0; i < mp.positions.length; i++) {
			mp.positions[i].sub(mb.center);
		}

		return mp;
	}

}
