/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.functionality3d;

import javax.media.j3d.BoundingSphere;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.View;
import javax.vecmath.AxisAngle4f;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import org.cellmicrocosmos.cm2.constants.CM2_CONSTANTS;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.engine.DataPool;
import org.cellmicrocosmos.cm2.engine.DataPool.J3DSettings;
import org.cellmicrocosmos.cm2.interfaces.MembraneViewListener;


import com.sun.j3d.utils.behaviors.vp.OrbitBehavior;
import com.sun.j3d.utils.universe.SimpleUniverse;

/**
 * Handles the camera in the 3d view.
 * 
 * @author Tim Dingersen dackelnose@gmx.net
 * @author Sebastian Schneider
 * 
 */
public class CameraMode {

	public static final int EXTRA = 0;

	public static final int INTRA = 1;

	public static final int FREE = -1;

	private int mode = EXTRA;

	private OrbitBehaviorAdvanced orbit;

	private SimpleUniverse universe;

	
	private MembraneViewListener listener;



	private static Canvas3D canvas;

	

	/**
	 * Equips the canvas with a camera mode
	 * 
	 * @param canvas
	 * @param universe
	 */
	public CameraMode(Canvas3D canvas, SimpleUniverse universe) {

		CameraMode.canvas=canvas;
		this.universe = universe;
		orbit = new OrbitBehaviorAdvanced(canvas, OrbitBehavior.REVERSE_ALL, listener);
		orbit.setSchedulingBounds(new BoundingSphere(
				new Point3d(0.0, 0.0, 0.0), 100000.0));
		this.setMouseSensitivity(DataPool.J3DSettings.mouseFactor);
		universe.getViewingPlatform().setViewPlatformBehavior(orbit);

		universe.getViewer().getView().setFrontClipPolicy(CM2_CONSTANTS.J3D.VIEW_FRONT_CLIP_POLICY);
		universe.getViewer().getView().setBackClipPolicy(CM2_CONSTANTS.J3D.VIEW_BACK_CLIP_POLICY);
		universe.getViewer().getView().setFrontClipDistance(CM2_CONSTANTS.J3D.VIEW_FRONT_CLIP_DISTANCE);
		universe.getViewer().getView().setBackClipDistance(CM2_CONSTANTS.J3D.VIEW_BACK_CLIP_DISTANCE);
		universe.getViewer().getView().setSceneAntialiasingEnable(DataPool.J3DSettings.antialiasing);
		if (J3DSettings.parallelProjection) {
			universe.getViewer().getView().setProjectionPolicy(View.PARALLEL_PROJECTION);
			universe.getViewer().getView().setScreenScalePolicy(View.SCALE_EXPLICIT);
		} else {
			universe.getViewer().getView().setProjectionPolicy(View.PERSPECTIVE_PROJECTION);
			universe.getViewer().getView().setScreenScalePolicy(View.SCALE_SCREEN_SIZE);
		}
	}

	/**
	 * Sets the mousesensitivity synchroniously for rotation, zoom and
	 * translation
	 * 
	 * @param mouseFactor
	 */
	public void setMouseSensitivity(double mouseFactor) {

		orbit.setZoomFactor(mouseFactor + 1d);
		orbit.setRotFactors(mouseFactor, mouseFactor);
		orbit.setTransFactors(mouseFactor, mouseFactor);
	}

	/**
	 * sets the view by the mode. modes are: 1. FREE : a free, moveable view 2.
	 * EXTRA : a view from above the membrane on the membrane (extracellular
	 * side) 3. INTRA : a view from under the membrane on the membrane
	 * (intracellular side)
	 * 
	 * @param mode
	 * @param center 
	 */
	public void setViewByMode(int mode, Point3f center) {
		this.mode = mode;

		float mwidth = (float)DataPool.getMembraneModel().getSize().getWidth()
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;
		float mheight=
			(float) DataPool.getMembraneModel().getSize().getHeight()
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;
		

		
		if(center==null)
				center=new Point3f(mwidth/2,0,mheight/2);

		
		float Hc=Math.max(mwidth,mheight);

		
		Hc*=1.6f;
		Hc+=center.y;
		
		
		
		
		
		
		if (mode == Membrane3D.FREE) {

			orbit.setRotateEnable(true);

			this.mode = FREE;

		} else if (mode == Membrane3D.EXTRA) {

			TransformGroup homeTG;
			Transform3D home = new Transform3D();

			homeTG = universe.getViewingPlatform().getViewPlatformTransform();

			homeTG.getTransform(home);

			Vector3f v = new Vector3f(center.x, (float) Hc,
					center.z);

			home.setTranslation(v);
			home.setRotation(new AxisAngle4f(1, 0, 0, (float) Math
					.toRadians(270)));

			homeTG.setTransform(home);

			orbit.setRotateEnable(false);
			this.mode = EXTRA;
		} else if (mode == Membrane3D.INTRA) {

			TransformGroup homeTG;
			Transform3D home = new Transform3D();

			homeTG = universe.getViewingPlatform().getViewPlatformTransform();

			homeTG.getTransform(home);

			Vector3f v = new Vector3f(center.x, -(float) Hc,
					center.x);

			home.setTranslation(v);
			home.setRotation(new AxisAngle4f(1, 0, 0, (float) Math
					.toRadians(90)));

			homeTG.setTransform(home);

			orbit.setRotateEnable(false);
			this.mode = INTRA;
		}
		if (mode > -1)
			listener.displayedLayerChanged(mode);
		
		orbit.setRotationCenter(new Point3d(center));
	}

	

	public void addMembraneViewListener(MembraneViewListener listener) {
		this.listener = listener;
		orbit.setMembraneViewListener(listener);
	}

	/**
	 * Returns the currently active cameramode
	 * 
	 * @return
	 */
	public int getMode() {
		return mode;
	}

	/**
	 * en- or disables the movable camera.
	 * 
	 * @param b
	 */
	public void setEnabled(boolean b) {
		
		if (b) {
			if (this.mode == FREE)
				orbit.setRotateEnable(true);
			else
				orbit.setRotateEnable(false);

			orbit.setTranslateEnable(true);
			orbit.setZoomEnable(true);
		} else {

			orbit.setRotateEnable(false);
			orbit.setTranslateEnable(false);
//			orbit.setZoomEnable(true);
		}

	}
	
	public void setZoomEnable(boolean b) {
		orbit.setZoomEnable(b);
	}
	public boolean isEnabled() {
		return orbit.getRotateEnable();
	}

	

	public void forwardButton(int bt,Molecule_3D ref) {
		
		

		


		Point3f center=new Point3f(ref.getPosition());
		center.scale(CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE);

			TransformGroup homeTG;
			Transform3D home = new Transform3D();

			homeTG = universe.getViewingPlatform().getViewPlatformTransform();

			homeTG.getTransform(home);

			Vector3f v;
			
			Point3f lower=ref.getLower(false);
			Point3f upper=ref.getUpper(false);
			
			float dist = lower.distance(upper);
			
			if(bt==0) {
				v= new Vector3f( center.x, center.y,
						center.z-dist*2);
				home.setRotation(new AxisAngle4f(0, 1, 0, (float) Math
						.toRadians(180)));
			}
					
			else if(bt==1) {
				v= new Vector3f( center.x+dist*2, center.y,
						center.z);
				home.setRotation(new AxisAngle4f(0, 1, 0, (float) Math
						.toRadians(90)));
			}
					
			else if(bt==2) {
				v= new Vector3f( center.x, center.y,
						center.z+dist*2);
				home.setRotation(new AxisAngle4f(1, 0, 0, (float) Math
						.toRadians(0)));
			}
					
			else if(bt==3) {
				v= new Vector3f( center.x-dist*2, center.y,
						center.z);
				home.setRotation(new AxisAngle4f(0, 1, 0, (float) Math
						.toRadians(270)));
			}
					
			else {
				v= new Vector3f( center.x,  center.y+dist*2,
						center.z);
				home.setRotation(new AxisAngle4f(1, 0, 0, (float) Math
						.toRadians(270)));
			}
					
			

			home.setTranslation(v);
			

			homeTG.setTransform(home);

		
		
		
		
	
		
		
		
	}

	public void moveY(float diff) {
		Transform3D home = new Transform3D();

		TransformGroup homeTG = universe.getViewingPlatform().getViewPlatformTransform();

		homeTG.getTransform(home);
//		Transform3D newone = new Transform3D(home);
		
		Vector3f v = new Vector3f();
		home.get(v);
		
		v=new Vector3f(v.x,v.y+diff,v.z);
		home.setTranslation(v);
		
		
		
		homeTG.setTransform(home);
	}

	public void setRotateEnable(boolean b) {
	this.orbit.setRotateEnable(b);
		
	}
	
	public static Point3f getCurrentEyePosition() {
		Point3d eye_pos = new Point3d();

		canvas.getCenterEyeInImagePlate(eye_pos);

		Transform3D motion = new Transform3D();
		canvas.getImagePlateToVworld(motion);
		motion.transform(eye_pos);
		return new Point3f(eye_pos);
	}
	

}
