/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.functionality3d;

import java.awt.Dimension;
import java.util.Vector;

import javax.media.j3d.BoundingBox;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3d;

import org.cellmicrocosmos.cm2.constants.CM2_CONSTANTS;
import org.cellmicrocosmos.cm2.dataobjects.Atom_3D;
import org.cellmicrocosmos.cm2.dataobjects.MembraneModel;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.engine.DataPool;
import org.j3d.geom.IntersectionUtils;


/**
 * Offers various Methods to check Objects for Intersections. All tests work
 * completely offscreen, meaning that the tested Molecule instances dont have to
 * be added into the 3d view. The position and rotation of the molecule to apply
 * the test for is retrieved from the molecule itself; so these values have to
 * be set up properly
 * 
 * How the intersection tests between molecules work:
 * 
 * If the test contains a check for collision with the border, then this will be
 * the first step; if that test succeeds already, all following tests become
 * needless. If the test doesn't contain a border check, then there has to be
 * taken care of elsewhere.
 * 
 * The second test is a test wether or not the bounding box of the objects
 * collide. If not, the detailed test is needless. This early rejection saves a
 * lot of computation.
 * 
 * If there are any colliding bounds, a detailed test will be performed on the
 * referring objects then. This detailed test include: -an intersection test
 * between the shapes and -an inside test, performed for both candidates
 * 
 * @author Tim Dingersen dackelnose@gmx.net
 * 
 */
public class IntersectManager {

	/*
	 * used as returnvalues for the enhanced border collision test
	 */
	public static final int BORDER_NORTH = 0;

	public static final int BORDER_EAST = 1;

	public static final int BORDER_SOUTH = 2;

	public static final int BORDER_WEST = 3;

	public static final int BORDER_MAX_Y = 4;

	public static final int BORDER_MIN_Y = 5;

	/*
	 * The Vector that collided objects get added to after a check was performed
	 */
	private Vector<Molecule_3D> collisionObjects = new Vector<Molecule_3D>();

	// used for the ray-trianglearray-check
	final private IntersectionUtils utils = new IntersectionUtils();

	// the 6 orthogonal directions in a 3dimensional universe. Used for
	// inside-checking
	public final static Vector3d[] possibleDirections = new Vector3d[] {
			new Vector3d(1, 0, 0), new Vector3d(-1, 0, 0),
			new Vector3d(0, 0, 1), new Vector3d(0, 0, -1),
			new Vector3d(0, 1, 0), new Vector3d(0, -1, 0) };

	/**
	 * Tests, if a molecule is completely inside the defined membrane bounds
	 * Returns false if not, no matter if the molecule intersects with the
	 * border or if it is completely outside the range
	 * 
	 * @param mol
	 * @param membraneSize
	 * @return
	 */
	public boolean intercects_Molecule_Border(final Molecule_3D mol,
			final Dimension membraneSize) {

		final float width = membraneSize.width
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;
		final float height = membraneSize.height
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;

		final Point3f lower = mol.getLower(false);
		final Point3f upper = mol.getUpper(false);

		if (lower.x > 0f
				&& upper.x < width
				&& lower.z > 0
				&& upper.z < height
				&& lower.y > CM2_CONSTANTS.MEMBRANE_SIZES.MIN_Y_TRANSLATION // [old from Tim, makes no sense?!] /10f) 
						* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE
				&& upper.y < CM2_CONSTANTS.MEMBRANE_SIZES.MAX_Y_TRANSLATION
						* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE)
			return false;

		// Point3f[] new1 = mol.getAlignedTrianglePoints();
		//
		// for (int i = 0; i < new1.length; i++) {
		//
		// if (new1[i].x < 0
		// || new1[i].z < 0
		// || new1[i].x >= width
		// || new1[i].z >= height
		// || Math.abs(new1[i].y) >=
		// CM2_CONSTANTS.MEMBRANE_SIZES.MAX_Y_TRANSLATION
		// * CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE) {
		//
		// return true;
		// }
		//
		// }

		return true;
	}
	
	/**
	 * Tests, if a molecule is inside the upper limit.
	 * Returns false if not, no matter if the molecule intersects with the
	 * border or if it is completely outside the range
	 * 
	 * @param mol
	 * @param membraneSize
	 * @return
	 */
	public boolean intercects_Molecule_Sky(final Molecule_3D mol) {

		final Point3f upper = mol.getUpper(false);

		if (upper.y < CM2_CONSTANTS.MEMBRANE_SIZES.MAX_Y_TRANSLATION
						* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE)
			return false;

		return true;
	}
	
	/**
	 * Tests, if a molecule is inside the lower limit.
	 * Returns false if not, no matter if the molecule intersects with the
	 * border or if it is completely outside the range
	 * 
	 * @param mol
	 * @param membraneSize
	 * @return
	 */
	public boolean intercects_Molecule_Earth(final Molecule_3D mol) {

		final Point3f lower = mol.getLower(false);

		if (lower.y > CM2_CONSTANTS.MEMBRANE_SIZES.MIN_Y_TRANSLATION
						* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE)
			return false;

		return true;
	}

	/**
	 * An enhanced version of the molecule_border-test. Doesn't terminate after
	 * the first collision has been detected, instead all point are tested, and
	 * the identifiers (see static values) of all involved borders are returned
	 * in an array
	 * 
	 * @param mol
	 * @param membraneSize
	 * @return An empty array if no collision occured. Else the referring
	 *         numbers for the involved borders can be found inside
	 */
	public int[] intercects_Molecule_Border_In_Detail(final Molecule_3D mol,
			final Dimension membraneSize) {

		final float width = membraneSize.width
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;
		final float height = membraneSize.height
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;

		final Point3f lower = mol.getLower(false);
		final Point3f upper = mol.getUpper(false);

		if (lower.x > 0f
				&& upper.x < width
				&& lower.z > 0
				&& upper.z < height
				&& lower.y > CM2_CONSTANTS.MEMBRANE_SIZES.MIN_Y_TRANSLATION
						* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE
				&& upper.y < CM2_CONSTANTS.MEMBRANE_SIZES.MAX_Y_TRANSLATION
						* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE)
			return new int[0];

		// Point3f[] new1 = mol.getAlignedTrianglePoints();

		final Vector<Integer> results = new Vector<Integer>();
		results.setSize(6);
		// for (int i = 0; i < new1.length; i++) {
		//
		// if (new1[i].x < 0)
		// results.set(BORDER_WEST, BORDER_WEST);
		// if (new1[i].z < 0)
		// results.set(BORDER_NORTH, BORDER_NORTH);
		// if (new1[i].x >= width)
		// results.set(BORDER_EAST, BORDER_EAST);
		// if (new1[i].z >= height)
		// results.set(BORDER_SOUTH, BORDER_SOUTH);
		//
		// if (new1[i].y >= CM2_CONSTANTS.MEMBRANE_SIZES.MAX_Y_TRANSLATION
		// * CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE)
		// results.set(BORDER_MAX_Y, BORDER_MAX_Y);
		//
		// if (new1[i].y <= -CM2_CONSTANTS.MEMBRANE_SIZES.MAX_Y_TRANSLATION
		// * CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE)
		// results.set(BORDER_MIN_Y, BORDER_MIN_Y);
		//
		// }

		if (lower.x < 0)
			results.set(IntersectManager.BORDER_WEST,
					IntersectManager.BORDER_WEST);
		if (lower.z < 0)
			results.set(IntersectManager.BORDER_NORTH,
					IntersectManager.BORDER_NORTH);
		if (upper.x >= width)
			results.set(IntersectManager.BORDER_EAST,
					IntersectManager.BORDER_EAST);
		if (upper.z >= height)
			results.set(IntersectManager.BORDER_SOUTH,
					IntersectManager.BORDER_SOUTH);

		if (upper.y >= CM2_CONSTANTS.MEMBRANE_SIZES.MAX_Y_TRANSLATION
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE)
			results.set(IntersectManager.BORDER_MAX_Y,
					IntersectManager.BORDER_MAX_Y);

		if (lower.y <= CM2_CONSTANTS.MEMBRANE_SIZES.MIN_Y_TRANSLATION
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE)
			results.set(IntersectManager.BORDER_MIN_Y,
					IntersectManager.BORDER_MIN_Y);

		results.trimToSize();
		final Integer[] integers = results.toArray(new Integer[0]);

		final int[] retval = new int[integers.length];

		for (int i = 0; i < integers.length; i++) {
			if (integers[i] != null) {
				retval[i] = integers[i].intValue();
			} else
				retval[i] = -1;
		}

		return retval;
	}

	public boolean intercectsBorderOrAnyMolecule(final Molecule_3D mol,
			final boolean breakAfterFirstHit) {
		final boolean goon = this.intercects_Molecule_Border(mol, DataPool
				.getMembraneModel().getSize());
		if (breakAfterFirstHit && goon)
			return true;
		return this.intersectsMoleculeAnyMolecules(mol, breakAfterFirstHit);

	}

	/**
	 * Tests, if a molecule collides with either the border or any Protein that
	 * is already added to the membrane.
	 * 
	 * 
	 * @param mol
	 * @param breakAfterFirstHit
	 *            if false, all other proteins will be checked for collision.
	 *            All collision results will be added into a list that can be
	 *            used afterwards. If its just necessary to test wether a
	 *            collision occures at all, then it should be set true, to
	 *            reduce costs
	 * @return
	 */
	public boolean intercectsWithBorderOrAnyProtein(final Molecule_3D mol,
			final boolean breakAfterFirstHit) {
		this.collisionObjects.clear();
		boolean hit = false;

		if (this.intercects_Molecule_Border(mol, DataPool.getMembraneModel()
				.getSize())) {

			return true;
		}

		if (DataPool.J3DSettings.atomCollisionCheck == false
				&& DataPool.J3DSettings.shapeCollisionCheck == false)
			return false;

		final Vector<Molecule_3D> collidedRaw = new Vector<Molecule_3D>();
		if (hit == false) {

			final MembraneModel model = DataPool.getMembraneModel();
			final int length = model.getProteinCount();
			final BoundingBox extendedBB = this.getExtendedBoundingBox(mol
					.getAlignedGeometricBounds());

			for (int i = 0; i < length; i++) {

				final Molecule_3D tmP = model.getProteinAt(i);

				if (tmP == mol)
					continue;
				if (this
						.getExtendedBoundingBox(tmP.getAlignedGeometricBounds())
						.intersect(extendedBB)) {

					collidedRaw.add(tmP);

				}

			}
		}

		for (int i = 0; i < collidedRaw.size(); i++) {

			final boolean b = (DataPool.J3DSettings.atomCollisionCheck ? this
					.intersectsAtoms(mol, collidedRaw.get(i)) : false)
					|| (DataPool.J3DSettings.shapeCollisionCheck ? this
							.intersects_Molecule_Molecule_Detailed(mol,
									collidedRaw.get(i)) : false);

			if (b) {

				this.collisionObjects.add(collidedRaw.get(i));
				hit = true;

				if (breakAfterFirstHit)
					break;
			}

		}

		return hit;
	}

	/**
	 * Tests, if a molecule collides with any molecule in a list that also has
	 * to be passed. Very useful when the possible collision candidates are
	 * known because that way reduces a lot of computation
	 * 
	 * @param mol
	 * @param breakAfterFirstHit
	 *            if false, all other proteins will be checked for collision.
	 *            All collision results will be added into a list that can be
	 *            used afterwards. If its just necessary to test wether a
	 *            collision occures at all, then it should be set true, to
	 *            reduce costs
	 * @return
	 */
	public boolean intersectsMoleculeMoleculesVector(final Molecule_3D mol,
			final Vector<?> molecules, final boolean breakAfterFirstHit) {
		this.collisionObjects.clear();
		boolean hit = false;

		if (DataPool.J3DSettings.atomCollisionCheck == false
				&& DataPool.J3DSettings.shapeCollisionCheck == false)
			return false;

		final Vector<Molecule_3D> collidedRaw = new Vector<Molecule_3D>();

		final int length = molecules.size();

		final BoundingBox extendedBB = this.getExtendedBoundingBox(mol
				.getAlignedGeometricBounds());

		for (int i = 0; i < length; i++) {
			if (molecules.get(i) instanceof Molecule_3D == false)
				continue;

			final Molecule_3D tmP = (Molecule_3D) molecules.get(i);

			if (tmP == mol)
				continue;
			if (this.getExtendedBoundingBox(tmP.getAlignedGeometricBounds())
					.intersect(extendedBB)) {

				collidedRaw.addElement(tmP);

			}

		}

		for (int i = 0; i < collidedRaw.size(); i++) {

			final boolean b = (DataPool.J3DSettings.atomCollisionCheck ? this
					.intersectsAtoms(mol, collidedRaw.get(i)) : false)
					|| (DataPool.J3DSettings.shapeCollisionCheck ? this
							.intersects_Molecule_Molecule_Detailed(mol,
									collidedRaw.get(i)) : false);

			if (b) {

				this.collisionObjects.addElement(collidedRaw.get(i));
				hit = true;
				if (breakAfterFirstHit)
					break;
			}

		}

		return hit;
	}

	/**
	 * Tests, if a molecule collides with any Protein that is already added to
	 * the membrane.
	 * 
	 * @param mol
	 * @param breakAfterFirstHit
	 *            if false, all other proteins will be checked for collision.
	 *            All collision results will be added into a list that can be
	 *            used afterwards. If its just necessary to test wether a
	 *            collision occures at all, then it should be set true, to
	 *            reduce costs
	 * @return
	 */
	public boolean intersectsMoleculeAnyProtein(final Molecule_3D mol,
			final boolean breakAfterFirstHit) {
		final MembraneModel model = DataPool.getMembraneModel();

		return this.intersectsMoleculeMoleculesVector(mol, model
				.getAllProteins(), breakAfterFirstHit);

	}

	/**
	 * Tests, if a molecule collides with any Molecule that is already added to
	 * the membrane, inclusive lipids.
	 * 
	 * @param mol
	 * @param breakAfterFirstHit
	 *            if false, all other molecules will be checked for collision.
	 *            All collision results will be added into a list that can be
	 *            used afterwards. If its just necessary to test wether a
	 *            collision occures at all, then it should be set true, to
	 *            reduce costs
	 * @return
	 */
	public boolean intersectsMoleculeAnyMolecules(final Molecule_3D mol,
			final boolean breakAfterFirstHit) {
		final MembraneModel model = DataPool.getMembraneModel();

		boolean b1 = false;
		boolean b2 = false;
		boolean b3 = false;

		final Vector<Molecule_3D> coll = new Vector<Molecule_3D>();

		b1 = this.intersectsMoleculeMoleculesVector(mol,
				model.getAllProteins(), breakAfterFirstHit);
		for (int i = 0; i < this.collisionObjects.size(); i++)
			coll.addElement(this.collisionObjects.get(i));

		if ((b1 == true) && breakAfterFirstHit)
			return true;

		b2 = this.intersectsMoleculeMoleculesVector(mol, model
				.getAllLipidsExtra(), breakAfterFirstHit);
		for (int i = 0; i < this.collisionObjects.size(); i++)
			coll.addElement(this.collisionObjects.get(i));

		if ((b2 == true) && breakAfterFirstHit)
			return true;

		b3 = this.intersectsMoleculeMoleculesVector(mol, model
				.getAllLipidsIntra(), breakAfterFirstHit);
		for (int i = 0; i < this.collisionObjects.size(); i++)
			coll.addElement(this.collisionObjects.get(i));

		this.collisionObjects = coll;
		return b1 || b2 || b3;
	}

	/**
	 * Returns the list of Objects, that have collided during the last run of a
	 * test. Does only contain any objects if, of course at least one collision
	 * occured and the "breakAfterFirstHit" boolean was set to false for that
	 * last test
	 * 
	 * @return
	 */
	public Molecule_3D[] getCollisionObjects() {
		return this.collisionObjects.toArray(new Molecule_3D[0]);
	}

	/**
	 * Tests the collision between 2 objects in a detailed way. This method has
	 * to be used each time two bounding boxes intersect. In this method, the
	 * first object will be devided into triangles. The second one will
	 * represent the geometry to test these triangles angainst. For that, the
	 * intersect triangle.shape method will be used until a collision gets
	 * detected. If so, the test returns. If not: Afterwards a inside test will
	 * be performed on the objects that tests, wether the first object may be
	 * completely inside the second object's hull. If this returns false, the
	 * same test is repeated another time with reversed roles. If that test also
	 * returns false, an overlap can definitely be rejected
	 * 
	 * @param first
	 * @param second
	 * @return
	 */

	private boolean intersects_Molecule_Molecule_Detailed(
			final Molecule_3D first, final Molecule_3D second) {

		final Point3f[] new1 = first.getAlignedTrianglePoints();
		final Point3f[] new2 = second.getAlignedTrianglePoints();

		Vector<Triangle> TRS1 = new Vector<Triangle>();
		Vector<Triangle> TRS2 = new Vector<Triangle>();

		final int l1 = new1.length / 3;

		int c = -1;
		BoundingBox bb = second.getAlignedGeometricBounds();
		for (int i = 0; i < l1; i++) {
			final Point3f p1 = new1[++c];
			final Point3f p2 = new1[++c];
			final Point3f p3 = new1[++c];

			final Point3d[] A = { new Point3d(p1), new Point3d(p2),
					new Point3d(p3) };

			final Point3d lower = new Point3d(A[0]);
			final Point3d upper = new Point3d(A[0]);

			for (int j = 0; j < 3; j++) {
				lower.x = Math.min(lower.x, A[j].x);
				lower.y = Math.min(lower.y, A[j].y);
				lower.z = Math.min(lower.z, A[j].z);

				upper.x = Math.max(upper.x, A[j].x);
				upper.y = Math.max(upper.y, A[j].y);
				upper.z = Math.max(upper.z, A[j].z);
			}

			if (bb.intersect(new BoundingBox(lower, upper)))
				TRS1.add(new Triangle(p1, p2, p3, lower, upper));
		}
		final int l2 = new2.length / 3;

		c = -1;
		bb = first.getAlignedGeometricBounds();
		for (int i = 0; i < l2; i++) {
			final Point3f p1 = new2[++c];
			final Point3f p2 = new2[++c];
			final Point3f p3 = new2[++c];

			final Point3d[] A = { new Point3d(p1), new Point3d(p2),
					new Point3d(p3) };

			final Point3d lower = new Point3d(A[0]);
			final Point3d upper = new Point3d(A[0]);

			for (int j = 0; j < 3; j++) {
				lower.x = Math.min(lower.x, A[j].x);
				lower.y = Math.min(lower.y, A[j].y);
				lower.z = Math.min(lower.z, A[j].z);

				upper.x = Math.max(upper.x, A[j].x);
				upper.y = Math.max(upper.y, A[j].y);
				upper.z = Math.max(upper.z, A[j].z);
			}

			if (bb.intersect(new BoundingBox(lower, upper)))
				TRS2.add(new Triangle(p1, p2, p3, lower, upper));
		}

		final Vector<Triangle> TRS1temp = new Vector<Triangle>();
		final int s1 = TRS1.size();
		int size = TRS1.size();
		int size2 = TRS2.size();
		for (int i = 0; i < size; i++) {
			final Triangle t = TRS1.get(i);

			for (int j = 0; j < size2; j++) {
				if (t.boundingbox.intersect(TRS2.get(j).boundingbox)) {
					TRS1temp.addElement(t);
					break;
				}
			}
		}

		TRS1 = TRS1temp;

		final Vector<Triangle> TRS2temp = new Vector<Triangle>();
		final int s2 = TRS2.size();
		size = TRS2.size();
		size2 = TRS1.size();
		for (int i = 0; i < size; i++) {
			final Triangle t = TRS2.get(i);

			for (int j = 0; j < size2; j++) {
				if (t.boundingbox.intersect(TRS1.get(j).boundingbox)) {
					TRS2temp.addElement(t);
					break;
				}
			}
		}

		TRS2 = TRS2temp;

		Triangle[] trarray = TRS1.toArray(new Triangle[0]);
		int l3 = trarray.length;
		int l4 = TRS2.size();
		float[] tar = new float[l4 * 9];
		c = -1;
		for (int i = 0; i < l4; i++) {
			final Triangle t = TRS2.get(i);
			tar[++c] = (float) t.a.x;
			tar[++c] = (float) t.a.y;
			tar[++c] = (float) t.a.z;
			tar[++c] = (float) t.b.x;
			tar[++c] = (float) t.b.y;
			tar[++c] = (float) t.b.z;
			tar[++c] = (float) t.c.x;
			tar[++c] = (float) t.c.y;
			tar[++c] = (float) t.c.z;
		}

		for (int i = 0; i < l3; i++) {

			if (this.intersects_Triangle_Shape(trarray[i], tar)) {
				return true;
			}
		}

		if (s1 == l1) {
			final int l5 = new2.length;
			final float[] hull = new float[l5 * 3];
			c = -1;
			for (int i = 0; i < l5; i++) {
				hull[++c] = new2[i].x;
				hull[++c] = new2[i].y;
				hull[++c] = new2[i].z;
			}
			if (this.isCompletelyInside(first, hull)) {
				return true;
			}
		}

		trarray = TRS2.toArray(new Triangle[0]);
		l3 = trarray.length;
		l4 = TRS1.size();
		tar = new float[l4 * 9];
		c = -1;
		for (int i = 0; i < l4; i++) {
			final Triangle t = TRS1.get(i);
			tar[++c] = (float) t.a.x;
			tar[++c] = (float) t.a.y;
			tar[++c] = (float) t.a.z;
			tar[++c] = (float) t.b.x;
			tar[++c] = (float) t.b.y;
			tar[++c] = (float) t.b.z;
			tar[++c] = (float) t.c.x;
			tar[++c] = (float) t.c.y;
			tar[++c] = (float) t.c.z;
		}

		for (int i = 0; i < l3; i++) {

			if (this.intersects_Triangle_Shape(trarray[i], tar)) {
				return true;
			}
		}

		if (s2 == l2) {
			final int l5 = new1.length;
			final float[] hull = new float[l5 * 3];
			c = -1;
			for (int i = 0; i < l5; i++) {
				hull[++c] = new1[i].x;
				hull[++c] = new1[i].y;
				hull[++c] = new1[i].z;
			}
			if (this.isCompletelyInside(second, hull)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * The test if a triangle object intersects a geometry array The passed
	 * triangle gets devided into 3 rays with a limited length. Those rays will
	 * be tested against the geometry array using the intersection utils method
	 * 
	 * @param triangle
	 * @param array
	 * @return
	 */
	private boolean intersects_Triangle_Shape(final Triangle triangle,
			final float[] array) {

		final Point3d[] A = { triangle.a, triangle.b, triangle.c };
		final Point3d[] B = { triangle.b, triangle.c, triangle.a };

		for (int i = 0; i < 3; i++) {

			final Vector3d direction = new Vector3d(B[i]);
			direction.sub(A[i]);

			if (this.utils.rayTriangleArray(A[i], direction, (float) direction
					.length(), array, array.length / 9, new Point3d(), true))

				return true;
		}

		return false;
	}

	/**
	 * The test that checks wether to first object's center is located inside
	 * the second object's hull. "Shoots" rays in all 6 orthogonal directions.
	 * If all(!) of these rays collide will the hull of the second object, this
	 * test succeeds
	 * 
	 * @param inner
	 * @param TRARouter
	 * @return
	 */
	private boolean isCompletelyInside(final Molecule_3D inner,
			final float[] TRARouter) {

		final Point3f upperInner = inner.getUpper(false);

		final Point3f lowerInner = inner.getLower(false);

		final float WXinner = Math.abs(upperInner.x - lowerInner.x);

		final float WYinner = Math.abs(upperInner.y - lowerInner.y);

		final float WZinner = Math.abs(upperInner.z - lowerInner.z);

		final Point3d centerInner = new Point3d(upperInner.x - (WXinner / 2),
				upperInner.y - (WYinner / 2), upperInner.z - (WZinner / 2));

		boolean allRaysHit = true;

		final int length = IntersectManager.possibleDirections.length;

		for (int i = 0; i < length; i++) {

			if (this.utils.rayTriangleArray(centerInner,
					IntersectManager.possibleDirections[i], 0, TRARouter,
					TRARouter.length / 9, new Point3d(), true) == false) {
				allRaysHit = false;
				break;
			}
		}

		return allRaysHit;
	}

	public boolean isCompletelyInside(final Point3d inner,
			final float[] TRARouter) {

		boolean allRaysHit = true;

		final int length = IntersectManager.possibleDirections.length;

		for (int i = 0; i < length; i++) {

			if (this.utils.rayTriangleArray(inner,
					IntersectManager.possibleDirections[i], 0, TRARouter,
					TRARouter.length / 9, new Point3d(), true) == false) {
				allRaysHit = false;
				break;
			}
		}

		return allRaysHit;
	}

	/**
	 * TODO
	 * 
	 * @param mol
	 * @param molecule_3D
	 * @return
	 */
	private boolean intersectsAtoms(final Molecule_3D mol1,
			final Molecule_3D mol2) {

		Atom_3D[] a1 = mol1.getAlignedAtoms();
		Atom_3D[] a2 = mol2.getAlignedAtoms();

		final Point3d upper1 = new Point3d(mol1.getUpper(true));
		final Point3d lower1 = new Point3d(mol1.getLower(true));

		final Point3d upper2 = new Point3d(mol2.getUpper(true));
		final Point3d lower2 = new Point3d(mol2.getLower(true));

		final Point3d toAdd = new Point3d(
				DataPool.J3DSettings.minatomdistance,
				DataPool.J3DSettings.minatomdistance,
				DataPool.J3DSettings.minatomdistance);

		upper1.add(toAdd);
		upper2.add(toAdd);
		lower1.sub(toAdd);
		lower2.sub(toAdd);

		final BoundingBox bb1 = new BoundingBox(lower1, upper1);
		final BoundingBox bb2 = new BoundingBox(lower2, upper2);

		int l1 = a1.length;
		int l2 = a2.length;

		final Vector<Atom_3D> v1 = new Vector<Atom_3D>();
		final Vector<Atom_3D> v2 = new Vector<Atom_3D>();
		for (int i = 0; i < l1; i++) {
			if (bb2.intersect(new Point3d(a1[i].position)))
				v1.add(a1[i]);
		}
		for (int i = 0; i < l2; i++) {
			if (bb1.intersect(new Point3d(a2[i].position)))
				v2.add(a2[i]);
		}

		a1 = v1.toArray(new Atom_3D[0]);
		a2 = v2.toArray(new Atom_3D[0]);

		l1 = a1.length;
		l2 = a2.length;

		for (int i = 0; i < l1; i++) {

			for (int j = 0; j < l2; j++) {
				final float dist = a1[i].position.distance(a2[j].position);
				if (dist <= (a1[i].radiusCov + a2[j].radiusCov + DataPool.J3DSettings.minatomdistance)) {
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Returns the extend bounding box if the atom collision is
	 * activated and the minatomdistance has been changed in
	 * the 3D Settings Window
	 * 
	 * @param box
	 * @return
	 */
	private BoundingBox getExtendedBoundingBox(final BoundingBox box) {

		// the minatomdistance is only added to the bounding box
		// if the atom collision is activated
		if (!DataPool.J3DSettings.atomCollisionCheck || DataPool.J3DSettings.minatomdistance == 0)
			return box;
		final Point3d upper = new Point3d();
		box.getUpper(upper);
		final Point3d lower = new Point3d();
		box.getLower(lower);

		final BoundingBox bb = new BoundingBox();

		final float addDist = (DataPool.J3DSettings.minatomdistance / 2f)
				* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;

		bb.setUpper(upper.x + addDist, upper.y + addDist, upper.z + addDist);
		bb.setLower(lower.x - addDist, lower.y - addDist, lower.z - addDist);

		return bb;
	}

	/**
	 * A tuple of points that represent a triangle
	 * 
	 * @author Tim Dingersen dackelnose@gmx.net
	 * 
	 */
	private class Triangle {
		public Point3d a;

		public Point3d b;

		public Point3d c;

		public BoundingBox boundingbox;

		public Triangle(final Point3f a, final Point3f b, final Point3f c,
				final Point3d lower, final Point3d upper) {
			this.a = new Point3d(a);
			this.b = new Point3d(b);
			this.c = new Point3d(c);
			this.boundingbox = new BoundingBox(lower, upper);
		}
	}

}
