/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.functionality3d;

import java.awt.Point;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Vector;

import javax.media.j3d.Canvas3D;
import javax.media.j3d.PointLight;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Transform3D;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3d;

import org.cellmicrocosmos.cm2.constants.CM2_CONSTANTS;
import org.cellmicrocosmos.cm2.dataobjects.Lipid_3D;
import org.cellmicrocosmos.cm2.dataobjects.MembraneModel;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.dataobjects.Protein_3D;
import org.cellmicrocosmos.cm2.engine.DataPool;
import org.cellmicrocosmos.cm2.interfaces.MembraneViewListener;


import com.sun.j3d.utils.picking.PickCanvas;
import com.sun.j3d.utils.picking.PickResult;
import com.sun.j3d.utils.picking.PickTool;

/**
 * Manages all picking and movement operations in the 3d view
 * 
 * @author Tim Dingersen dackelnose@gmx.net
 * 
 */
public class PickNDragManager implements MouseListener, MouseMotionListener,
		KeyListener {

	private final Vector<Molecule_3D> selected = new Vector<Molecule_3D>();

	private MembraneViewListener listener;

	private final Canvas3D canvas;

	private final CameraMode camera;

	private final IntersectManager intersector;

	private final PickCanvas pickcanvas;

	private boolean moving = false;

	private boolean isActive = true;

	private final PointLight light1;

	private final PointLight light2;

	private final PointLight light3;

	private boolean draggingAllowed = true;

	private final Hud hud;

	private Point3f oldPosition;

	private boolean maypick = true;
	private boolean mayDrag = true;

	/**
	 * Constructs the PND manager and registers all required references
	 * 
	 * @param canvas
	 * @param intersector
	 * @param cameraMode
	 * @param hash4All
	 * @param myPick
	 * @param light1
	 * @param light2
	 * @param light3
	 */
	public PickNDragManager(final Canvas3D canvas,
			final IntersectManager intersector, final CameraMode cameraMode,
			final PickCanvas myPick, final PointLight light1,
			final PointLight light2, final PointLight light3, final Hud hud) {

		this.canvas = canvas;
		this.camera = cameraMode;
		this.intersector = intersector;

		this.pickcanvas = myPick;
		this.canvas.addMouseListener(this);
		this.canvas.addMouseMotionListener(this);
		this.canvas.addKeyListener(this);
		this.hud = hud;

		this.pickcanvas.setTolerance(0.5f);
		this.pickcanvas.setMode(PickTool.GEOMETRY);

		this.light1 = light1;
		this.light2 = light2;
		this.light3 = light3;

	}

	/**
	 * Returns the currently selected proteins
	 * 
	 * @return
	 */
	public Vector<Molecule_3D> getSelected() {
		return this.selected;
	}

	/**
	 * Selects all proteins instances that have the passed code
	 * 
	 * @param code
	 * @param domainID
	 */
	public void selectByCode(final String code, final int domainID) {
		final MembraneModel model = DataPool.getMembraneModel();
		this.selected.clear();
		if (code != null) {
			for (int i = 0; i < model.getProteinCount(); i++) {
				if (model.getProteinAt(i).getPDBcode().equals(code)
						&& model.getProteinAt(i).getDomainID() == domainID) {
					this.selected.add(model.getProteinAt(i));
				}

			}
		}

		if (this.selected.size() > 0) {
			this.listener.pickMolecules(this.selected
					.toArray(new Protein_3D[0]));
		} else {
			this.listener.pickMolecules(null);
		}

	}

	/**
	 * Handles the protein selection. Called each time a mouseclick on the
	 * canvas is recognized. If multiselection is true, then all instaces of
	 * same type get selected
	 * 
	 * @param arg0
	 * 
	 * 
	 */
	private void handleObjectSelection(final PickResult MyPickResult,
			final MouseEvent arg0, final boolean multi) {
		final Molecule_3D Picked;
		final Point3f p = this.calculateLocation(arg0.getX(), arg0.getY(),
				DataPool.getCurrentLayerHeight()
						* CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE);
		// if (!new Rectangle(0, 0, DataPool.getMembraneModel().getSize().width,
		// DataPool.getMembraneModel().getSize().height)
		// .contains(new Point((int) p.x, (int) p.z)))
		// return;
		if (MyPickResult == null
				|| (((Shape3D) MyPickResult.getObject()).getUserData() instanceof Lipid_3D && (DataPool.J3DSettings.lipidPicking == false))) {

			p.scale(1 / CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE);
			final int domainID = DataPool.getMembraneModel()
					.getMicroDomainAtPoint(new Point((int) p.x, (int) p.z),
							DataPool.getCurrentLayerID());

			this.listener.highlightMicroDomain(domainID);
			this.listener.pickMolecules(null);

			// if (listener.isProperties_Membrane_Atrribus_Showing() == false)

			// SwingUtilities.invokeLater(new Thread() {
			// public void run() {
			// canvas.getParent().requestFocus();
			// }
			// });

			return;
		}

		if (((Shape3D) MyPickResult.getObject()).getUserData() instanceof Molecule_3D == false)
			return;

		Picked = (Molecule_3D) ((Shape3D) MyPickResult.getObject())
				.getUserData();

		this.selected.clear();
		this.selected.add(Picked);
		// if (multi) {
		// Protein_3D[] others = DataPool.getMembraneModel()
		// .getProteinsByPDBCode(Picked.getPDBcode());
		// for (int i = 0; i < others.length; i++) {
		// if (others[i] != Picked
		// && others[i].getDomainID() == Picked.getDomainID())
		// this.selected.add(others[i]);
		// }
		//
		// }

		SwingUtilities.invokeLater(new Thread() {
			@Override
			public void run() {
				PickNDragManager.this.listener
						.pickMolecules(PickNDragManager.this.selected
								.toArray(new Molecule_3D[0]));
				PickNDragManager.this.listener.highlightMicroDomain(Picked
						.getDomainID());

			}
		});

	}

	public void addMembraneViewListener(final MembraneViewListener listener) {
		this.listener = listener;

	}

	public void mouseClicked(final MouseEvent arg0) {
		if (this.isActive == false)
			return;
		PickResult MyPickResult;

		this.pickcanvas.setShapeLocation(arg0);

		try {
		MyPickResult = this.pickcanvas.pickClosest();
		
		
		if (MyPickResult != null)
			return;

		if (arg0.getButton() == MouseEvent.BUTTON1) {

			this.selected.clear();

			this.handleObjectSelection(null, arg0, false);
		}
		
		
		
		}catch(Throwable e) {
			
		}
		

	}

	public void mouseEntered(final MouseEvent arg0) {

	}

	public void mouseExited(final MouseEvent arg0) {

	}

	public void mouseReleased(final MouseEvent arg0) {
		if (this.isActive == false)
			return;

		this.mayDrag = true;

		this.camera.setEnabled(true);

		final int bt = this.hud.getButton(arg0.getPoint());
		if (bt != -1) {
			this.camera.forwardButton(bt, this.selected.firstElement());
			return;
		}
		if (this.isActive == false)
			return;

		PickResult MyPickResult;

		this.pickcanvas.setShapeLocation(arg0);

		MyPickResult = this.pickcanvas.pickClosest();

		if (MyPickResult == null) {
			this.moving = false;
			return;
		}

		if (this.maypick) {
			if (arg0.getButton() == MouseEvent.BUTTON1 && this.moving == false) {

				this.selected.clear();
				if (arg0.getClickCount() > 1)
					this.handleObjectSelection(MyPickResult, arg0, true);
				else
					this.handleObjectSelection(MyPickResult, arg0, false);
			}

			if ((arg0.getButton() == MouseEvent.BUTTON3 || (arg0.getModifiers() == CM2_CONSTANTS.GUI_SETTINGS_AND_CONTROLS.RIGHT_BUTTON_MODIFIER))) {

				if (MyPickResult == null)
					return;
				if ((this.selected.contains(MyPickResult.getObject()) == false)) {
					this.selected.clear();
					this.handleObjectSelection(MyPickResult, arg0, false);
				}
				if (this.selected.isEmpty())
					return;

				final JPopupMenu pop = this.listener.createCanvasPopUpMenu();
				pop.show(this.canvas, arg0.getX(), arg0.getY());

			}

		}

		if (this.moving == false)
			return;
		if (arg0.getButton() == MouseEvent.BUTTON1)
			this.finishMoving();
		this.moving = false;
	}

	/**
	 * Converts the mouselocation on canvas to a location in the universe.
	 * Precisely, this will be the location where the picking ray intersects the
	 * plane at the passed y-height
	 * 
	 * @param mouseX
	 * @param mouseY
	 * @param Yobject
	 * @return
	 */
	public Point3f calculateLocation(final int mouseX, final int mouseY,
			final float Yobject) {
		final Point3d eye_pos = new Point3d();
		final Point3d mouse_pos = new Point3d();

		this.canvas.getCenterEyeInImagePlate(eye_pos);
		this.canvas.getPixelLocationInImagePlate(mouseX, mouseY, mouse_pos);

		final Transform3D motion = new Transform3D();
		this.canvas.getImagePlateToVworld(motion);
		motion.transform(eye_pos);
		motion.transform(mouse_pos);

		final Vector3d direction = new Vector3d(mouse_pos);
		direction.sub(eye_pos);

		final float camX = (float) eye_pos.x;
		final float camY = (float) eye_pos.y;
		final float camZ = (float) eye_pos.z;

		final float dirX = (float) direction.x;
		final float dirY = (float) direction.y;
		final float dirZ = (float) direction.z;

		final float y = Yobject * CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;

		final float r = (y - camY) / dirY;
		final float finalX = camX + (r * dirX);
		final float finalZ = camZ + (r * dirZ);

		return new Point3f(finalX, y, finalZ);
	}

	public void mousePressed(final MouseEvent arg0) {

		final int bt = this.hud.getButton(arg0.getPoint());
		if (bt != -1) {
			this.camera.forwardButton(bt, this.hud.currentMoleculeAligning);

		}

		
		this.maypick = true;
		 if (this.isActive == false)
		 return;

		PickResult MyPickResult;

		this.pickcanvas.setShapeLocation(arg0);

		try {
		MyPickResult = this.pickcanvas.pickClosest();
		}catch (Throwable e) {
			System.out.println("Error caught at MousePressed, PNDManager");
			return;
		}
		if (MyPickResult == null || this.selected.isEmpty())
			this.mayDrag = false;

		else if (((Shape3D) MyPickResult.getObject()).getUserData() != this.selected
				.firstElement())
			this.mayDrag = false;
		else
			this.mayDrag = true;

		// camera.setEnabled(false);

	}

	public void mouseMoved(final MouseEvent arg0) {
		this.hud.forwardMouseEvent(arg0);
	}

	public void mouseDragged(final MouseEvent arg0) {
		this.maypick = false;
		this.updateLights();

		if (this.isActive == false)
			return;
		if (this.draggingAllowed == false || this.mayDrag == false)
			return;
		// if(!moving)return;

		if (!(arg0.getModifiers() == InputEvent.BUTTON1_MASK))
			return;

		if (this.selected.isEmpty())
			return;

		if (!this.moving)
			this.oldPosition = new Point3f(this.selected.firstElement()
					.getPosition());

		PickResult MyPickResult;

		this.pickcanvas.setShapeLocation(arg0);

		MyPickResult = this.pickcanvas.pickClosest();

		if (MyPickResult == null)
			return;

		if (((Shape3D) MyPickResult.getObject()).getUserData() == this.selected
				.firstElement())
			this.moving = true;

		this.camera.setEnabled(false);

		// if(!moving)this.oldPosition=new
		// Point3f(this.selected.firstElement().getPosition());
		// this.moving = true;
		// if (selected.size() > 1) {
		// Molecule_3D first = selected.firstElement();
		// selected.clear();
		// selected.add(first);
		// listener.pickMolecules(selected.toArray(new Protein_3D[0]));
		// }

		if (this.moving)
			this.moveIt(arg0);
	}

	/**
	 * Recalculates the pointlight positions regarding to the current camera
	 * position
	 * 
	 */
	public void updateLights() {

		if ((DataPool.J3DSettings.light1 || DataPool.J3DSettings.light2) == false)
			return;

		final Point3f eye_pos = CameraMode.getCurrentEyePosition();
		this.light1.setPosition(eye_pos);

		final Point3d pos2 = new Point3d(eye_pos.x, -eye_pos.y, eye_pos.z);
		final Point3d pos3 = new Point3d(-eye_pos.x, 0, eye_pos.z);

		this.light2.setPosition(new Point3f(pos2));
		this.light3.setPosition(new Point3f(pos3));
	}

	public void keyPressed(final KeyEvent arg0) {
		if (this.isActive == false)
			return;
		if (arg0.getKeyCode() == KeyEvent.VK_ESCAPE)
			this.cancelDragging();

	}

	public void keyReleased(final KeyEvent arg0) {

	}

	public void keyTyped(final KeyEvent arg0) {

	}

	/**
	 * Cancels the current dragging operation
	 * 
	 */
	private void cancelDragging() {
		if (this.moving == false)
			return;

		this.selected.firstElement().setPosition(new Point3f(this.oldPosition));
		this.selected.firstElement().move();
		this.selected.firstElement().setPositionIsInAValidSpace(true);

		this.moving = false;

	}

	/**
	 * Moves the first element of the current selection on the plane to the
	 * position the mouse is pointing to
	 * 
	 * @param arg0
	 */
	private void moveIt(final MouseEvent arg0) {
		final Point3f newPoint = this.calculateLocation(arg0.getX(), arg0
				.getY(), this.selected.firstElement().getYTranslation());

		this.selected.firstElement().setPosition(
				newPoint.x / CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE,
				newPoint.z / CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE);
		this.selected.firstElement().move();
		final boolean invalid = this.intersector.intercectsBorderOrAnyMolecule(
				this.selected.firstElement(), true);

		this.selected.firstElement().setPositionIsInAValidSpace(!invalid);

	}

	/**
	 * Finsished the current moving operation
	 * 
	 */
	private void finishMoving() {

		if (this.selected.isEmpty())
			return;

		// if (selected.firstElement().isPositionisinavalidspace() == false) {
		// this.cancelDragging();
		// return;
		// }
		final int domainID = DataPool.getMembraneModel().getMicroDomainAtPoint(
				new Point((int) this.selected.firstElement().getX(),
						(int) this.selected.firstElement().getZ()),
				DataPool.getCurrentLayerID());

		if (domainID < 0) {
			JOptionPane
					.showMessageDialog(
							this.canvas,
							"The center point of the molecule must\nbe placed inside membrane bounds!",
							"Illega Position", JOptionPane.ERROR_MESSAGE);
			this.cancelDragging();
		} else if (this.selected.firstElement() instanceof Lipid_3D
				&& domainID != this.selected.firstElement().getDomainID()) {
			JOptionPane.showMessageDialog(this.canvas,
					"The lipid can only be moved\ninside its associated area!",
					"Illega Position", JOptionPane.ERROR_MESSAGE);
			this.cancelDragging();
		}

		this.selected.firstElement().setPositionIsInAValidSpace(true);
		this.listener.updateProperties();
		
		DataPool.setMembraneHasBeenEditedAgain(true);

		

		DataPool.getMembraneModel().organizeDomains();

		this.listener.refreshViews(true, false, false, false, false, true);

		this.listener.pickMolecules(new Molecule_3D[] { this.selected
				.firstElement() });
		this.listener.highlightMicroDomain(this.selected.firstElement()
				.getDomainID());

	}

	/**
	 * Moves the passed Object to the passed mouse location on canvas (this time
	 * while Drag and drop is active)
	 * 
	 * @param current
	 * @param loc
	 */
	public void moveDNDObject(final Molecule_3D current, final Point loc) {

		final Point3f newPoint = this.calculateLocation((int) loc.getX(),
				(int) loc.getY(), current.getYTranslation());

		current.setPosition(newPoint.x / CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE,
				newPoint.z / CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE);

		final boolean invalid = this.intersector.intercectsBorderOrAnyMolecule(
				current, true);

		current.setPositionIsInAValidSpace(!invalid);
		current.move();

	}

	/**
	 * En-or disables the picking
	 * 
	 * @param b
	 */
	public void setActive(final boolean b) {
		this.isActive = b;

	}

	/**
	 * En-or disables dragging
	 * 
	 * @param b
	 */
	public void allowDragging(final boolean b) {
		this.draggingAllowed = b;

	}

	public void setPickedMolecules(final Molecule_3D[] proPicked) {
		this.selected.clear();
		if (proPicked == null || proPicked.length < 1) {

			return;
		}

		for (int i = 0; i < proPicked.length; i++) {
			this.selected.add(proPicked[i]);
		}

	}

}
