/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.gui.elements;

import java.awt.Font;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import javax.vecmath.Point3f;

import org.cellmicrocosmos.cm2.dataobjects.Lipid_3D;
import org.cellmicrocosmos.cm2.dataobjects.MembraneModel;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.dataobjects.Protein_3D;
import org.cellmicrocosmos.cm2.engine.DataPool;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.labels.ItemLabelPosition;
import org.jfree.chart.labels.StandardCategoryItemLabelGenerator;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.category.BarRenderer;
import org.jfree.data.category.DefaultCategoryDataset;


/**
 * The statistics displayed in the gui. Shows 
 *  the percental weight (lipids+proteins)
 *  the absolute weight (lipids+prtoeins)
 *  the percental distribution (lipids)
 *  the absolute distribution (lipids)
 * of each different type in the membrane in relation to each other.
 * Thereby the according microdomain is taken into account.
 * 
 * @author El_PolloDiablo@gmx.de, Tim Dingersen, Sebastian Rubert, Bjoern Sommer
 *         bjoern@CELLmicrocosmos.org
 * 
 */

@SuppressWarnings("serial")
public class StatisticsView extends JPanel {

	// the displayed Atomic Mass Diagramm-Tabs
	private final JTabbedPane massTabPane = new JTabbedPane();

	// the displayed Atomic Absolute Mass Diagramm-Tabs
	private final JTabbedPane massAbsTabPane = new JTabbedPane();
	
	// the displayed Atomic Ratio Diagramm-Tabs
	private final JTabbedPane numberTabPane = new JTabbedPane();

	// the displayed Atomic Absolute Number Diagramm-Tabs
	private final JTabbedPane numberAbsTabPane = new JTabbedPane();
	
	// the displayed Register-Tabs (change mass<->number)
	JTabbedPane tab = new JTabbedPane();

	// datasets of BarCharts
	private final DefaultCategoryDataset barDatasetICMass, barDatasetECMass, barDatasetICMassAbs, barDatasetECMassAbs,
									barDatasetICNumb, barDatasetECNumb, barDatasetICNumbAbs, barDatasetECNumbAbs;

	private DefaultCategoryDataset barDatasetAllMass, barDatasetAllMassAbs, 
									barDatasetAllNumb, barDatasetAllNumbAbs;

	// plots of BarCharts
	private CategoryPlot bPlotInMass, bPlotExMass, bPlotAllMass, bPlotInMassAbs, bPlotExMassAbs, bPlotAllMassAbs, 
							bPlotInNumb, bPlotExNumb, bPlotAllNumb, bPlotInNumbAbs, bPlotExNumbAbs, bPlotAllNumbAbs;

	// the 4 differentCharts
	private JFreeChart bChartInMass, bChartExMass, bChartAllMass, bChartInMassAbs, bChartExMassAbs, bChartAllMassAbs,
							bChartInNumb, bChartExNumb, bChartAllNumb, bChartInNumbAbs, bChartExNumbAbs, bChartAllNumbAbs;

	/**
	 * 
	 * The initialising constructor for this class with some test data and a
	 * previous TabbedPane constructed
	 * 
	 */
	public StatisticsView() {

		// hide all windows
		this.numberTabPane.setVisible(false);
		this.numberAbsTabPane.setVisible(false);
		this.massTabPane.setVisible(false);
		this.massAbsTabPane.setVisible(false);

		this.tab.setVisible(false);

		// Layout-Settings
		this.setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		this.barDatasetICMass = new DefaultCategoryDataset();
		this.barDatasetECMass = new DefaultCategoryDataset();
		this.barDatasetAllMass = new DefaultCategoryDataset();
		this.barDatasetICMassAbs = new DefaultCategoryDataset();
		this.barDatasetECMassAbs = new DefaultCategoryDataset();
		this.barDatasetAllMassAbs = new DefaultCategoryDataset();
		this.barDatasetICNumb = new DefaultCategoryDataset();
		this.barDatasetECNumb = new DefaultCategoryDataset();
		this.barDatasetAllNumb = new DefaultCategoryDataset();
		this.barDatasetICNumbAbs = new DefaultCategoryDataset();
		this.barDatasetECNumbAbs = new DefaultCategoryDataset();
		this.barDatasetAllNumbAbs = new DefaultCategoryDataset();

		// Components of TabbedPane for Atomic Mass
		this.massTabPane.addTab("EC+IC Molecules", this.createBarChartAllMass());
		this.massTabPane.addTab("EC Molecules", this.createBarChartExMass());
		this.massTabPane.addTab("IC Molecules", this.createBarChartInMass());

		// Components of TabbedPane for Absolute Atomic Mass
		this.massAbsTabPane.addTab("EC+IC Molecules", this.createBarChartAllMassAbs());
		this.massAbsTabPane.addTab("EC Molecules", this.createBarChartExMassAbs());
		this.massAbsTabPane.addTab("IC Molecules", this.createBarChartInMassAbs());

		// Components of TabbedPane for Lipid Ratio
		this.numberTabPane.addTab("EC+IC Lipids", this.createBarChartAllNumb());
		this.numberTabPane.addTab("EC Lipids", this.createBarChartExNumb());
		this.numberTabPane.addTab("IC Lipids", this.createBarChartInNumb());

		// Components of TabbedPane for Absolute Lipid Number
		this.numberAbsTabPane.addTab("EC+IC Lipids", this.createBarChartAllNumbAbs());
		this.numberAbsTabPane.addTab("EC Lipids", this.createBarChartExNumbAbs());
		this.numberAbsTabPane.addTab("IC Lipids", this.createBarChartInNumbAbs());

		BarRenderer rendererNumb = null;
		for (int i=0;i<12;i++) {
			switch (i) {
				case 0 : rendererNumb = (BarRenderer) this.bPlotExMass.getRenderer();break;
				case 1 : rendererNumb = (BarRenderer) this.bPlotInMass.getRenderer();break;
				case 2 : rendererNumb = (BarRenderer) this.bPlotAllMass.getRenderer();break;
				case 3 : rendererNumb = (BarRenderer) this.bPlotExMassAbs.getRenderer();break;
				case 4 : rendererNumb = (BarRenderer) this.bPlotInMassAbs.getRenderer();break;
				case 5 : rendererNumb = (BarRenderer) this.bPlotAllMassAbs.getRenderer();break;
				case 6 : rendererNumb = (BarRenderer) this.bPlotExNumb.getRenderer();break;
				case 7 : rendererNumb = (BarRenderer) this.bPlotInNumb.getRenderer();break;
				case 8 : rendererNumb = (BarRenderer) this.bPlotAllNumb.getRenderer();break;
				case 9 : rendererNumb = (BarRenderer) this.bPlotExNumbAbs.getRenderer();break;
				case 10: rendererNumb = (BarRenderer) this.bPlotInNumbAbs.getRenderer();break;
				case 11: rendererNumb = (BarRenderer) this.bPlotAllNumbAbs.getRenderer();break;

			}
			rendererNumb.setBaseItemLabelsVisible(true);

			rendererNumb
					.setBaseItemLabelGenerator(new StandardCategoryItemLabelGenerator());
			rendererNumb.setBasePositiveItemLabelPosition(new ItemLabelPosition(
					org.jfree.chart.labels.ItemLabelAnchor.OUTSIDE1,
					org.jfree.ui.TextAnchor.BASELINE_RIGHT));
		}

		this.add(this.tab);

		this.tab.setTabPlacement(SwingConstants.LEFT);

		this.tab.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);

		this.tab.setFont(this.tab.getFont().deriveFont(Font.BOLD));

		final JPanel p1 = new JPanel() {
			public void doLayout() {
				StatisticsView.this.numberTabPane.setPreferredSize(this
						.getSize());
				super.doLayout();
			}
		};
		p1.add(this.numberTabPane);
		this.tab.addTab("#", p1);

		final JPanel p2 = new JPanel() {
			public void doLayout() {
				StatisticsView.this.numberAbsTabPane.setPreferredSize(this
						.getSize());
				super.doLayout();
			}
		};
		p2.add(this.numberAbsTabPane);
		this.tab.addTab("|#|", p2);

		final JPanel p3 = new JPanel() {
			public void doLayout() {
				StatisticsView.this.massTabPane
						.setPreferredSize(this.getSize());
				super.doLayout();
			}
		};
		p3.add(this.massTabPane);
		this.tab.addTab("u", p3);

		final JPanel p4 = new JPanel() {
			public void doLayout() {
				StatisticsView.this.massAbsTabPane
						.setPreferredSize(this.getSize());
				super.doLayout();
			}
		};
		p4.add(this.massAbsTabPane);
		this.tab.addTab("|u|", p4);
		
		// clear the background colors of the statistic views
		this.updateColors();
		
		this.tab.validate();
	}

	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart for the
	 * Intracellular-View
	 */
	private ChartPanel createBarChartInMass() {
		this.bChartInMass = ChartFactory.createBarChart3D("Atomic Mass (u) IC Ratio",
				"", "", this.barDatasetICMass, PlotOrientation.VERTICAL, true,
				true, false);
		this.bPlotInMass = (CategoryPlot) this.bChartInMass.getPlot();
		this.bPlotInMass.setNoDataMessage("!!!No Data!!!");
		this.bPlotInMass.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Molecule u%");
		this.bPlotInMass.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartInMass);

		this.bPlotInMass.setBackgroundAlpha(0.0f);
		return pChart;
	}
	
	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart for the
	 * Intracellular-View
	 */
	private ChartPanel createBarChartInMassAbs() {
		this.bChartInMassAbs = ChartFactory.createBarChart3D("Atomic Mass (u) IC Absolute",
				"", "", this.barDatasetICMassAbs, PlotOrientation.VERTICAL, true,
				true, false);
		this.bPlotInMassAbs = (CategoryPlot) this.bChartInMassAbs.getPlot();
		this.bPlotInMassAbs.setNoDataMessage("!!!No Data!!!");
		this.bPlotInMassAbs.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Molecule u");
		this.bPlotInMassAbs.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartInMassAbs);

		this.bPlotInMassAbs.setBackgroundAlpha(0.0f);
		return pChart;
	}

	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart with Number of
	 * Lipids Ratio for the Intracellular-View
	 */
	private ChartPanel createBarChartInNumb() {

		this.bChartInNumb = ChartFactory.createBarChart3D(
				"Number of IC Lipids Ratio", "", "", this.barDatasetICNumb,
				PlotOrientation.VERTICAL, true, true, false);

		this.bPlotInNumb = (CategoryPlot) this.bChartInNumb.getPlot();
		this.bPlotInNumb.setNoDataMessage("!!!No Data!!!");
		this.bPlotInNumb.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Overall Lipid %");
		this.bPlotInNumb.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartInNumb);

		this.bPlotInNumb.setBackgroundAlpha(0.0f);

		return pChart;
	}

	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart with Number of
	 * Lipids for the Intracellular-View
	 */
	private ChartPanel createBarChartInNumbAbs() {

		this.bChartInNumbAbs = ChartFactory.createBarChart3D(
				"Number of IC Lipids Absolute", "", "", this.barDatasetICNumbAbs,
				PlotOrientation.VERTICAL, true, true, false);

		this.bPlotInNumbAbs = (CategoryPlot) this.bChartInNumbAbs.getPlot();
		this.bPlotInNumbAbs.setNoDataMessage("!!!No Data!!!");
		this.bPlotInNumbAbs.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Overall Lipid %");
		this.bPlotInNumbAbs.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartInNumbAbs);

		this.bPlotInNumbAbs.setBackgroundAlpha(0.0f);

		return pChart;
	}

	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart for the
	 * Extracellular-View
	 */
	private ChartPanel createBarChartExMass() {
		this.bChartExMass = ChartFactory.createBarChart3D("Atomic Mass (u) EC Ratio",
				"", "", this.barDatasetECMass, PlotOrientation.VERTICAL, true,
				true, false);
		this.bPlotExMass = (CategoryPlot) this.bChartExMass.getPlot();
		this.bPlotExMass.setNoDataMessage("!!!No Data!!!");
		this.bPlotExMass.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Molecule u%");
		this.bPlotExMass.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartExMass);

		this.bPlotExMass.setBackgroundAlpha(0.0f);
		return pChart;
	}
	
	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart for the
	 * Extracellular-View
	 */
	private ChartPanel createBarChartExMassAbs() {
		this.bChartExMassAbs = ChartFactory.createBarChart3D("Atomic Mass (u) EC Absolute",
				"", "", this.barDatasetECMassAbs, PlotOrientation.VERTICAL, true,
				true, false);
		this.bPlotExMassAbs = (CategoryPlot) this.bChartExMassAbs.getPlot();
		this.bPlotExMassAbs.setNoDataMessage("!!!No Data!!!");
		this.bPlotExMassAbs.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Molecule u");
		this.bPlotExMassAbs.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartExMassAbs);

		this.bPlotExMassAbs.setBackgroundAlpha(0.0f);
		return pChart;
	}

	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart with Number of
	 * Lipids Ratio for the Extracellular-View
	 */
	private ChartPanel createBarChartExNumb() {
		this.bChartExNumb = ChartFactory.createBarChart3D(
				"Number of EC Lipids Ratio", "", "Number of Lipids (%)",
				this.barDatasetECNumb, PlotOrientation.VERTICAL, true, true,
				false);

		this.bPlotExNumb = (CategoryPlot) this.bChartExNumb.getPlot();
		this.bPlotExNumb.setNoDataMessage("!!!No Data!!!");
		this.bPlotExNumb.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Overall Lipid %");
		this.bPlotExNumb.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartExNumb);

		this.bPlotExNumb.setBackgroundAlpha(0.0f);
		return pChart;
	}

	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart with Number of
	 * Lipids for the Extracellular-View
	 */
	private ChartPanel createBarChartExNumbAbs() {
		this.bChartExNumbAbs = ChartFactory.createBarChart3D(
				"Number of EC Lipids Absolute", "", "Number of Lipids",
				this.barDatasetECNumbAbs, PlotOrientation.VERTICAL, true, true,
				false);

		this.bPlotExNumbAbs = (CategoryPlot) this.bChartExNumbAbs.getPlot();
		this.bPlotExNumbAbs.setNoDataMessage("!!!No Data!!!");
		this.bPlotExNumbAbs.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Overall Lipid #");
		this.bPlotExNumbAbs.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartExNumbAbs);

		this.bPlotExNumbAbs.setBackgroundAlpha(0.0f);
		return pChart;
	}

	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart for both
	 * views
	 */
	private ChartPanel createBarChartAllMass() {
		this.bChartAllMass = ChartFactory.createBarChart3D("Atomic Mass (u) EC+IC Ratio",
				"", "", this.barDatasetAllMass, PlotOrientation.VERTICAL, true,
				true, false);
		this.bPlotAllMass = (CategoryPlot) this.bChartAllMass.getPlot();
		this.bPlotAllMass.setNoDataMessage("!!!No Data!!!");
		this.bPlotAllMass.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Molecule u%");
		this.bPlotAllMass.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartAllMass);

		this.bPlotAllMass.setBackgroundAlpha(0.0f);
		return pChart;
	}
	
	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart for both
	 * views
	 */
	private ChartPanel createBarChartAllMassAbs() {
		this.bChartAllMassAbs = ChartFactory.createBarChart3D("Atomic Mass (u) EC+IC Absolute",
				"", "", this.barDatasetAllMassAbs, PlotOrientation.VERTICAL, true,
				true, false);
		this.bPlotAllMassAbs = (CategoryPlot) this.bChartAllMassAbs.getPlot();
		this.bPlotAllMassAbs.setNoDataMessage("!!!No Data!!!");
		this.bPlotAllMassAbs.setRangeGridlinesVisible(false);

		final ValueAxis values = new NumberAxis("Molecule u");
		this.bPlotAllMassAbs.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartAllMassAbs);

		this.bPlotAllMassAbs.setBackgroundAlpha(0.0f);
		return pChart;
	}
	
	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart with Number of
	 * Lipids Ratio on both sides
	 */
	private ChartPanel createBarChartAllNumb() {

		this.bChartAllNumb = ChartFactory.createBarChart3D(
				"Number of EC+IC Lipids Ratio", "", "", this.barDatasetAllNumb,
				PlotOrientation.VERTICAL, true, true, false);

		this.bPlotAllNumb = (CategoryPlot) this.bChartAllNumb.getPlot();
		this.bPlotAllNumb.setNoDataMessage("!!!No Data!!!");
		this.bPlotAllNumb.setRangeGridlinesVisible(false);

		this.bPlotAllNumb.getRangeAxis().setVisible(false);

		final ValueAxis values = new NumberAxis("Overall Lipid %");
		this.bPlotAllNumb.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartAllNumb);

		this.bPlotAllNumb.setBackgroundAlpha(0.0f);

		return pChart;
	}
	
	/**
	 * 
	 * simple method to return a ChartPanel containing a BarChart with Number of
	 * Lipids on both sides
	 */
	private ChartPanel createBarChartAllNumbAbs() {

		this.bChartAllNumbAbs = ChartFactory.createBarChart3D(
				"Number of EC+IC Lipids Absolute", "", "", this.barDatasetAllNumbAbs,
				PlotOrientation.VERTICAL, true, true, false);

		this.bPlotAllNumbAbs = (CategoryPlot) this.bChartAllNumbAbs.getPlot();
		this.bPlotAllNumbAbs.setNoDataMessage("!!!No Data!!!");
		this.bPlotAllNumbAbs.setRangeGridlinesVisible(false);

		this.bPlotAllNumbAbs.getRangeAxis().setVisible(false);

		final ValueAxis values = new NumberAxis("Overall Lipid #");
		this.bPlotAllNumbAbs.setRangeAxis(values);
		final ChartPanel pChart = new ChartPanel(this.bChartAllNumbAbs);

		this.bPlotAllNumbAbs.setBackgroundAlpha(0.0f);

		return pChart;
	}
	
	/**
	 * 
	 * @param model
	 * @param domainID
	 */
	public void updateMembraneStatistics(final MembraneModel model,
			final int domainID) {

		if (model.getProteinCount() == 0 && model.getLipidCountExtra() == 0
				&& model.getLipidCountIntra() == 0) {
			this.massTabPane.setVisible(false);
			this.massAbsTabPane.setVisible(false);
			this.numberTabPane.setVisible(false);
			this.numberAbsTabPane.setVisible(false);
			return;
		}

		// update the datasets
		this.updateDatasetNew(0, model, domainID);
		this.updateDatasetNew(1, model, domainID);
		this.updateDatasetNew(2, model, domainID);
		
		this.bPlotInMass.setDataset(this.barDatasetICMass);
		this.bPlotExMass.setDataset(this.barDatasetECMass);
		this.bPlotAllMass.setDataset(this.barDatasetAllMass);
		
		this.bPlotInMassAbs.setDataset(this.barDatasetICMassAbs);
		this.bPlotExMassAbs.setDataset(this.barDatasetECMassAbs);
		this.bPlotAllMassAbs.setDataset(this.barDatasetAllMassAbs);
		
		this.bPlotInNumb.setDataset(this.barDatasetICNumb);
		this.bPlotExNumb.setDataset(this.barDatasetECNumb);
		this.bPlotAllNumb.setDataset(this.barDatasetAllNumb);
		
		this.bPlotInNumbAbs.setDataset(this.barDatasetICNumbAbs);
		this.bPlotExNumbAbs.setDataset(this.barDatasetECNumbAbs);
		this.bPlotAllNumbAbs.setDataset(this.barDatasetAllNumbAbs);
		
		this.massTabPane.setVisible(true);
		this.massAbsTabPane.setVisible(true);
		this.numberTabPane.setVisible(true);
		this.numberAbsTabPane.setVisible(true);

		this.tab.setVisible(true);

	}

	/**
	 * Update the dataset of the chosen side:
	 * 	side = 0: extra
	 *  side = 1: intra
	 *  side = 2: both
	 * @param side
	 * @param model
	 * @param domainID
	 */
	private void updateDatasetNew(final int side, final MembraneModel model,
			final int domainID) {

		final Hashtable<String, Molecule_3D> currenttypes = new Hashtable<String, Molecule_3D>();

		final int layerID = DataPool.getCurrentLayerID();

		String[] pTypes = null;
		Vector<Protein_3D> vProt;
		pTypes = model.getAllProteinTypesInDomain(domainID, layerID);
		vProt = this.getWantedProteins(side, domainID, layerID);
		
		Lipid_3D[] allLipids = null;
		switch (side) {
			case 0 : allLipids = model.getAllLipidsExtraInDomain(domainID, layerID);break;
			case 1 : allLipids = model.getAllLipidsIntraInDomain(domainID, layerID);break;
			case 2 : {
				Lipid_3D[]  allLipidsExtra = model.getAllLipidsExtraInDomain(domainID, layerID);
				Lipid_3D[]  allLipidsIntra = model.getAllLipidsIntraInDomain(domainID, layerID);
				allLipids = new Lipid_3D[allLipidsExtra.length+allLipidsIntra.length];
				for (int i=0;i<allLipidsExtra.length;i++) {
					allLipids[i] = allLipidsExtra[i];
				}
				for (int i=0;i<allLipidsIntra.length;i++) {
					allLipids[allLipidsExtra.length+i] = allLipidsIntra[i];
				}
				break;
			}
		}

		String[] lTypes = null;
		switch (side) {
			case 0 : lTypes = model.getAllLipidTypesExtraInDomain(domainID,layerID);break;
			case 1 : lTypes = model.getAllLipidTypesIntraInDomain(domainID,layerID);break;
			case 2 : {
				String[] lTypesExtra = model.getAllLipidTypesExtraInDomain(domainID,layerID);
				String[] lTypesIntra = model.getAllLipidTypesIntraInDomain(domainID,layerID);
				HashSet<String> lTypesHash = new HashSet<String>();
				for (int i=0;i<lTypesExtra.length;i++) {
					lTypesHash.add(lTypesExtra[i]);
				}
				for (int i=0;i<lTypesIntra.length;i++) {
					lTypesHash.add(lTypesIntra[i]);
				}
				lTypes = new String[lTypesHash.size()];
				lTypesHash.toArray(lTypes);
			}
			break;
		}

		BarRenderer rendererMass = null;
		switch (side) {
			case 0 : rendererMass = (BarRenderer) this.bPlotExMass.getRenderer();break;
			case 1 : rendererMass = (BarRenderer) this.bPlotInMass.getRenderer();break;
			case 2 : rendererMass = (BarRenderer) this.bPlotAllMass.getRenderer();break;
		}
				
		BarRenderer rendererMassAbs = null;
		switch (side) {
			case 0 : rendererMassAbs = (BarRenderer) this.bPlotExMassAbs.getRenderer();break;
			case 1 : rendererMassAbs = (BarRenderer) this.bPlotInMassAbs.getRenderer();break;
			case 2 : rendererMassAbs = (BarRenderer) this.bPlotAllMassAbs.getRenderer();break;
		}

		BarRenderer rendererNumb = null;
		switch (side) {
			case 0 : rendererNumb = (BarRenderer) this.bPlotExNumb.getRenderer();break;
			case 1 : rendererNumb = (BarRenderer) this.bPlotInNumb.getRenderer();break;
			case 2 : rendererNumb = (BarRenderer) this.bPlotAllNumb.getRenderer();break;
		}

		BarRenderer rendererNumbAbs = null;
		switch (side) {
			case 0 : rendererNumbAbs = (BarRenderer) this.bPlotExNumbAbs.getRenderer();break;
			case 1 : rendererNumbAbs = (BarRenderer) this.bPlotInNumbAbs.getRenderer();break;
			case 2 : rendererNumbAbs = (BarRenderer) this.bPlotAllNumbAbs.getRenderer();break;
		}
		
		DefaultCategoryDataset categoryDataSetMass = null;
		switch (side) {
			case 0 : categoryDataSetMass = this.barDatasetECMass;break;
			case 1 : categoryDataSetMass = this.barDatasetICMass;break;
			case 2 : categoryDataSetMass = this.barDatasetAllMass;break;
		}
		
		DefaultCategoryDataset categoryDataSetMassAbs = null;
		switch (side) {
			case 0 : categoryDataSetMassAbs = this.barDatasetECMassAbs;break;
			case 1 : categoryDataSetMassAbs = this.barDatasetICMassAbs;break;
			case 2 : categoryDataSetMassAbs = this.barDatasetAllMassAbs;break;
		}
		
		// compute the overall lipid and protein mass
		double overallLipidAndProteinMass = 0;
		for (int i = 0; i < pTypes.length; i++) {
			for (int j = 0; j < vProt.size(); j++) {
				final Protein_3D p = vProt.get(j);
				if (p.getPDBcode().equals(pTypes[i])) {
					overallLipidAndProteinMass += vProt.get(j).getAtomicMass();
				}
			}
		}
		for (int i = 0; i < lTypes.length; i++) {
			for (int j = 0; j < allLipids.length; j++) {
				if (allLipids[j].getPDBcode().equals(lTypes[i])) {
					overallLipidAndProteinMass += allLipids[j].getAtomicMass();
				}
			}
		}
		
		
		for (int i = 0; i < pTypes.length; i++) {
			double pSize = 0;

			for (int j = 0; j < vProt.size(); j++) {
				final Protein_3D p = vProt.get(j);
				if (p.getPDBcode().equals(pTypes[i])) {

					pSize = pSize + vProt.get(j).getAtomicMass();
					currenttypes.put(pTypes[i], p);
				}
			}

			categoryDataSetMass.setValue((pSize/overallLipidAndProteinMass) * 100, 
					pTypes[i], "Ratio");
			categoryDataSetMassAbs.setValue(pSize, pTypes[i], "Absolute");

		}

		final float overallLipidNumber = allLipids.length;

		DefaultCategoryDataset categoryDataSetNumb = null;
		switch (side) {
			case 0 : categoryDataSetNumb = this.barDatasetECNumb;break;
			case 1 : categoryDataSetNumb = this.barDatasetICNumb;break;
			case 2 : categoryDataSetNumb = this.barDatasetAllNumb;break;
		}
		
		DefaultCategoryDataset categoryDataSetNumbAbs = null;
		switch (side) {
			case 0 : categoryDataSetNumbAbs = this.barDatasetECNumbAbs;break;
			case 1 : categoryDataSetNumbAbs = this.barDatasetICNumbAbs;break;
			case 2 : categoryDataSetNumbAbs = this.barDatasetAllNumbAbs;break;
		}
		
		for (int i = 0; i < lTypes.length; i++) {
			double lSize = 0;
			float lNumberOutput = 0;

			for (int j = 0; j < allLipids.length; j++) {
				if (allLipids[j].getPDBcode().equals(lTypes[i])) {

					lSize = lSize + allLipids[j].getAtomicMass();
					currenttypes.put(lTypes[i], allLipids[j]);
					lNumberOutput++;
				}
			}

			categoryDataSetMass.setValue((lSize/overallLipidAndProteinMass) * 100, 
					lTypes[i], "Ratio");
			categoryDataSetMassAbs.setValue(lSize, lTypes[i], "Absolute");
			categoryDataSetNumb.setValue((lNumberOutput / overallLipidNumber) * 100,
					lTypes[i], "Ratio");
			categoryDataSetNumbAbs.setValue(lNumberOutput,
					lTypes[i], "Absolute");

		}

		for (int k = categoryDataSetMass.getRowCount() - 1; k > -1; k--) {
			if (!currenttypes.containsKey(categoryDataSetMass.getRowKey(k)))
				categoryDataSetMass.removeRow(k);
		}
		for (int k = categoryDataSetMassAbs.getRowCount() - 1; k > -1; k--) {
			if (!currenttypes.containsKey(categoryDataSetMassAbs.getRowKey(k)))
				categoryDataSetMassAbs.removeRow(k);
		}
		for (int k = categoryDataSetNumb.getRowCount() - 1; k > -1; k--) {
			if (!currenttypes.containsKey(categoryDataSetNumb.getRowKey(k)))
				categoryDataSetNumb.removeRow(k);
		}
		for (int k = categoryDataSetNumbAbs.getRowCount() - 1; k > -1; k--) {
			if (!currenttypes.containsKey(categoryDataSetNumbAbs.getRowKey(k)))
				categoryDataSetNumbAbs.removeRow(k);
		}

		final Enumeration<String> en = currenttypes.keys();
		while (en.hasMoreElements()) {
			final String key = en.nextElement();

			rendererMass.setSeriesPaint(categoryDataSetMass.getRowIndex(key), currenttypes.get(
					key).getColor());
			rendererMassAbs.setSeriesPaint(categoryDataSetMassAbs.getRowIndex(key), currenttypes.get(
					key).getColor());
			if (currenttypes.get(key) instanceof Lipid_3D) {
				rendererNumb.setSeriesPaint(categoryDataSetNumb.getRowIndex(key), currenttypes
						.get(key).getColor());
				rendererNumbAbs.setSeriesPaint(categoryDataSetNumb.getRowIndex(key), currenttypes
						.get(key).getColor());
			}
				
		}

	}

	/**
	 * Get the proteins of the wanted side:
	 *  side = 0: extra
	 *  side = 1: intra
	 *  side = 2: both
	 * @param side
	 * @param domainID
	 * @param layerID
	 * @return Vector<Protein_3D>
	 */
	private Vector<Protein_3D> getWantedProteins(final int side,
			final int domainID, final int layerID) {
		final Vector<Protein_3D> out = new Vector<Protein_3D>();
		final MembraneModel model = DataPool.getMembraneModel();

		final Protein_3D[] prots = model.getAllProteinsInDomain(domainID,
				layerID);
		if (side == 0) {

			for (int i = 0; i < prots.length; i++) {
				final Protein_3D prot = prots[i];

				final Point3f upper = prot.getUpper(false);

				if (upper.y > 0)
					out.addElement(prot);
			}
		} else if (side == 1) {

			for (int i = 0; i < prots.length; i++) {
				final Protein_3D prot = prots[i];

				final Point3f lower = prot.getLower(false);

				if (lower.y < 0)
					out.addElement(prot);
			}
		} else if (side == 2) {

			for (int i = 0; i < prots.length; i++) {
				final Protein_3D prot = prots[i];
				out.addElement(prot);
			}
		}
		return out;
	}

	/**
	 * Switches the tabs when the displayed membrane side changes
	 * 
	 * @param mode
	 */
	public void changeLayer(final int mode) {

//		final int massPrevindex = this.massTabPane.getSelectedIndex();
//		int massNewindex = massPrevindex;
//		if (mode == 0) {
//			if (massPrevindex == 1 || massPrevindex == 3)
//				massNewindex = massPrevindex - 1;
//		} else if (mode == 1) {
//			if (massPrevindex == 0 || massPrevindex == 2)
//				massNewindex = massPrevindex + 1;
//		}
//		this.massTabPane.setSelectedIndex(massNewindex);

		final int massPrevindex = this.massTabPane.getSelectedIndex();
		int massNewindex = massPrevindex;
		if (mode == 0) {
			if (massPrevindex == 1)
				massNewindex = massPrevindex - 1;
		} else if (mode == 1) {
			if (massPrevindex == 0)
				massNewindex = massPrevindex + 1;
		}
		this.massTabPane.setSelectedIndex(massNewindex);
		
		final int massAbsPrevindex = this.massAbsTabPane.getSelectedIndex();
		int massAbsNewindex = massAbsPrevindex;
		if (mode == 0) {
			if (massAbsPrevindex == 1)
				massAbsNewindex = massAbsPrevindex - 1;
		} else if (mode == 1) {
			if (massAbsPrevindex == 0)
				massAbsNewindex = massAbsPrevindex + 1;
		}
		this.massAbsTabPane.setSelectedIndex(massAbsNewindex);
		
		final int numbPrevindex = this.numberTabPane.getSelectedIndex();
		int numbNewindex = numbPrevindex;
		if (mode == 0) {
			if (numbPrevindex == 1)
				numbNewindex = numbPrevindex - 1;
		} else if (mode == 1) {
			if (numbPrevindex == 0)
				numbNewindex = numbPrevindex + 1;
		}
		this.numberTabPane.setSelectedIndex(numbNewindex);
		
		final int numbAbsPrevindex = this.numberAbsTabPane.getSelectedIndex();
		int numbAbsNewindex = numbAbsPrevindex;
		if (mode == 0) {
			if (numbAbsPrevindex == 1)
				numbAbsNewindex = numbAbsPrevindex - 1;
		} else if (mode == 1) {
			if (numbAbsPrevindex == 0)
				numbAbsNewindex = numbAbsPrevindex + 1;
		}
		this.numberAbsTabPane.setSelectedIndex(numbAbsNewindex);

	}

	public void updateColors() {

//		this.pChartIn.setBackgroundPaint(null);
//		this.pChartEx.setBackgroundPaint(null);
//		this.pChartAll.setBackgroundPaint(null);
		this.bChartInMass.setBackgroundPaint(null);
		this.bChartExMass.setBackgroundPaint(null);
		this.bChartAllMass.setBackgroundPaint(null);
		this.bChartInMassAbs.setBackgroundPaint(null);
		this.bChartExMassAbs.setBackgroundPaint(null);
		this.bChartAllMassAbs.setBackgroundPaint(null);
		this.bChartInNumb.setBackgroundPaint(null);
		this.bChartExNumb.setBackgroundPaint(null);
		this.bChartAllNumb.setBackgroundPaint(null);
		this.bChartInNumbAbs.setBackgroundPaint(null);
		this.bChartExNumbAbs.setBackgroundPaint(null);
		this.bChartAllNumbAbs.setBackgroundPaint(null);
	}

}
