/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.io;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.List;

import org.jdom.Comment;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

/**
 * The class that read and writes the Database contents
 * 
 * @author Tim Dingersen dackelnose@gmx.net
 * 
 */
public class DatabaseIO {

	public final static String ROOTELEMENTNAME_DB = "CM2_DATABASE_CONTENTS";

	// the format for the writer
	final Format format = Format.getPrettyFormat();

	/**
	 * Read the data from a file in a document that gets returned
	 * 
	 * @param file
	 * @return
	 * @throws IOException
	 * @throws JDOMException
	 */
	public Document loadFromFile(final File file) throws IOException,
			JDOMException {

		final Document doc = new SAXBuilder().build(file);

		if (doc.getRootElement() == null)
			throw new JDOMException("No root element found!");
		// invalidates the xml if no root element with the specified name can be
		// found
		else if (doc.getRootElement().getName().equals(
				DatabaseIO.ROOTELEMENTNAME_DB) == false)
			throw new JDOMException("The root element \""
					+ DatabaseIO.ROOTELEMENTNAME_DB + "\" cannot be resolved!");
		else
			return doc;
	}

	/**
	 * Write the data in the passed Document into a file
	 * 
	 * @param file
	 * @param doc
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	public void writeToFile(final File file, final Document doc)
			throws IOException {

		final String name = file.getName().substring(0,
				file.getName().length() - 4);

		final File dbpath = new File(file.getParent());

		if (dbpath.exists() == false)
			dbpath.mkdirs();

		final Comment comment = new Comment(this.getComment(name));
		doc.getContent().add(0, comment);

		final XMLOutputter outputter = new XMLOutputter();
		outputter.setFormat(this.format);
		final FileOutputStream output = new FileOutputStream(file);
		outputter.output(doc, output);
		output.close();

	}

	/**
	 * Returns the Comment String that should always be written into the XML
	 * file. The passed name will appear in this comment, too
	 * 
	 * @param nameofdb
	 * @return
	 */
	private String getComment(final String nameofdb) {
		final String comment = "This Document stores the entries of the Database named\n\n\""
				+ nameofdb
				+ "\"\n\nEach ENTRY-tag represents an single entry while its childs provide\n"
				+ "the entry's Properties.\n\n"
				+ "Beware: uncareful editing of this file may result in loss of Data\n";
		return comment;
	}

	/**
	 * Creates an Empty Document, used when a new Database is created from the
	 * Engine
	 * 
	 * @return
	 */
	public static Document getAnEmptyDocument() {
		final Document doc = new Document();
		final Element rootelement = new Element(DatabaseIO.ROOTELEMENTNAME_DB);
		doc.setRootElement(rootelement);
		return doc;
	}

	/**
	 * Read the data from a file and return a preview
	 * 
	 * @param file
	 * @return
	 * @throws IOException
	 * @throws JDOMException
	 */
	public static int[] readPreview(final File file) throws IOException,
			JDOMException {

		final Document doc = new SAXBuilder().build(file);

		if (doc.getRootElement() == null)
			throw new JDOMException("No root element found!");

		else if (doc.getRootElement().getName().equals(
				DatabaseIO.ROOTELEMENTNAME_DB) == false)
			throw new JDOMException("The root element \""
					+ DatabaseIO.ROOTELEMENTNAME_DB + "\" cannot be resolved!");

		final List mainelements = doc.getRootElement().getChildren();
		int Pros = 0;
		int Lips = 0;
		// Reading through each Entry
		for (int i = 0; i < mainelements.size(); i++) {

			final List subelements = ((Element) mainelements.get(i))
					.getChildren();

			final Hashtable<String, String> hash = new Hashtable<String, String>();

			// Values are put into a hashtable first
			for (int j = 0; j < subelements.size(); j++) {

				final Element subelement = ((Element) subelements.get(j));

				if (subelement.getName() != null
						&& subelement.getValue() != null)
					hash.put(subelement.getName(), subelement.getValue());

			}

			// ignores Entries without the tag "GROUP"
			if (hash.get("GROUP") == null)
				continue;

			if (hash.get("GROUP")
					.equals(MembraneXmlReader.TAGNAMES_MEMBRANE[0]))
				Pros++;
			else if (hash.get("GROUP").equals(
					MembraneXmlReader.TAGNAMES_MEMBRANE[1].substring(0, 6)))
				Lips++;
			else
				continue;

		}

		return new int[] { Pros, Lips };
	}

}
