/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.io;

import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.geom.Ellipse2D;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Enumeration;

import javax.vecmath.Matrix3f;
import javax.vecmath.Point3f;

import org.cellmicrocosmos.cm2.constants.CM2_CONSTANTS;
import org.cellmicrocosmos.cm2.dataobjects.Atom_3D;
import org.cellmicrocosmos.cm2.dataobjects.Layer;
import org.cellmicrocosmos.cm2.dataobjects.Lipid_3D;
import org.cellmicrocosmos.cm2.dataobjects.MembraneModel;
import org.cellmicrocosmos.cm2.dataobjects.MicroDomain_3D;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.dataobjects.Protein_3D;
import org.cellmicrocosmos.cm2.engine.DataPool;
import org.jdom.Comment;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;


/**
 * The Writer for the MembraneModel
 * 
 * @author Tim Dingersen dackelnose@gmx.net
 * 
 */
public class MembraneXmlWriter {

	// the format for the writer
	private final Format format = Format.getPrettyFormat();

	String workspace = null;

	/**
	 * 
	 * @param workspaceLocation
	 *            - The Directory of your workspace in order to write out
	 *            relative paths if the file are inside of it. Can be null if
	 *            you don't have a workspace or if you want to use absolute
	 *            paths
	 */
	public MembraneXmlWriter(final File workspaceLocation) {

		if (workspaceLocation != null)
			this.workspace = workspaceLocation.getAbsolutePath();

	}

	/**
	 * Stores the Data of the passed MembraneModel in a JDOM Document which gets
	 * written into the passed file.
	 * 
	 * @param file
	 * @param writeFilePaths
	 *            - true, if the filepaths should be written, false only for
	 *            writing a temporary model for the pdbwriter
	 * @param membraneModel
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	public void writeMembrane(final File file, final MembraneModel model) throws IOException {

		final Document doc = new Document();

		final Element rootElement = new Element(
				MembraneXmlReader.ROOTELEMENTNAME_MEMBRANE);
		doc.setRootElement(rootElement);

		final Comment comment = new Comment(this.getComment(model.getHeader()));
		doc.getContent().add(0, comment);

		// Tag for the numbers of Entries is created
		final Element counts = new Element(
				MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
		counts.setAttribute("PROTEINS", "" + model.getProteinCount());
		counts.setAttribute("LIPIDS", "" + model.getTotalMoleculeCount(1));
		rootElement.addContent(counts);

		// creating tags for General Membrane Data
		if (model.getLastGeneratedMembraneFile() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("LastGeneratedMembraneFile", model
					.getLastGeneratedMembraneFile().getPath());
			rootElement.addContent(ele);
		}
		if (model.getHeader() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Header", model.getHeader());
			rootElement.addContent(ele);
		}
		if (model.getCode() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("PdbCode", model.getCode());
			rootElement.addContent(ele);
		}
		if (model.getTitle() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Title", model.getTitle());
			rootElement.addContent(ele);
		}
		if (model.getSOURCE() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Species", model.getSOURCE());
			rootElement.addContent(ele);
		}
		if (model.getAuthor() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Author", model.getAuthor());
			rootElement.addContent(ele);
		}
		if (model.getDate() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Date", model.getDate());
			rootElement.addContent(ele);
		}
		if (model.getRemark() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Comment", model.getRemark());
			rootElement.addContent(ele);
		}

		if (model.getKEYWDS() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Keywds", model.getKEYWDS());
			rootElement.addContent(ele);
		}

		if (model.getEXPDTA() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Expdta", model.getEXPDTA());
			rootElement.addContent(ele);
		}

		if (model.getJRNL() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Jrnl", model.getJRNL());
			rootElement.addContent(ele);
		}

		if (model.getAlgorithmName() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("AlgorithmName", model.getAlgorithmName());
			rootElement.addContent(ele);
		}

		if (model.getAlgorithmDescription() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("AlgorithmDescription", model
					.getAlgorithmDescription());
			rootElement.addContent(ele);
		}

		if (model.getAlgorithmStringRepresentation() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("AlgorithmStringRepresentation", model
					.getAlgorithmStringRepresentation());
			rootElement.addContent(ele);
		}
		if (model.getSize() != null) {
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("Width", "" + model.getSize().width);
			ele.setAttribute("Height", "" + model.getSize().height);
			rootElement.addContent(ele);
		}
		{
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("DecimalPlaces", "" + model.getDecimalPlaces());
			rootElement.addContent(ele);
		}
		{
			final Element ele = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[5]);
			ele.setAttribute("ComputeAbsoluteValues", "" + model.getComputeAbsoluteValues());
			rootElement.addContent(ele);
		}
		

		final Element domains = new Element(
				MembraneXmlReader.TAGNAMES_MEMBRANE[9]);

		rootElement.addContent(domains);

		for (int i = 0; i < model.countMicroDomains(); i++) {
			final MicroDomain_3D micro = model.getMicroDomain(i);

			final Element ele = new Element("Domain");

			ele.setAttribute("Id", "" + micro.id);
			ele.setAttribute("LayerId", "" + micro.layerId);

			if (micro.geometry == MicroDomain_3D.RECTANGULAR) {
				ele.setAttribute("Geom", "Rectangular");
				ele.setAttribute("x", ((Rectangle) micro.geoObject).x + "");
				ele.setAttribute("y", ((Rectangle) micro.geoObject).y + "");
				ele.setAttribute("width", ((Rectangle) micro.geoObject).width
						+ "");
				ele.setAttribute("height", ((Rectangle) micro.geoObject).height
						+ "");
			}

			else if (micro.geometry == MicroDomain_3D.ELLIPSE) {
				ele.setAttribute("Geom", "Ellipse");
				ele.setAttribute("x",
						((int) ((Ellipse2D.Double) micro.geoObject).x) + "");
				ele.setAttribute("y",
						((int) ((Ellipse2D.Double) micro.geoObject).y) + "");
				ele
						.setAttribute(
								"width",
								((int) ((Ellipse2D.Double) micro.geoObject).width)
										+ "");
				ele.setAttribute("height",
						((int) ((Ellipse2D.Double) micro.geoObject).height)
								+ "");
			}

			else {
				ele.setAttribute("Geom", "Polygon");
				final Polygon poly = (Polygon) micro.geoObject;
				String coords = "";
				for (int j = 0; j < poly.npoints; j++) {
					coords += poly.xpoints[j] + " " + poly.ypoints[j] + " ";
				}
				ele.setAttribute("Coords", coords);
			}

			domains.addContent(ele);

		}

		final Element layers = new Element(
				MembraneXmlReader.TAGNAMES_MEMBRANE[10]);

		rootElement.addContent(layers);

		for (int i = 0; i < model.getLayerCount(); i++) {
			final Layer layer = model.getLayer(i);

			final Element ele = new Element("Layer");

			ele.setAttribute("Id", "" + layer.getLayerId());
			ele.setAttribute("Y_Val", "" + layer.getYHeight());

			layers.addContent(ele);

		}

		/*
		 * Entries for each Molecule are getting made. First for the Proteins,
		 * then for Lipids
		 */
		for (int group = 0; group < 5; group++) {

			final Element proteins = new Element(
					MembraneXmlReader.TAGNAMES_MEMBRANE[group]);

			rootElement.addContent(proteins);

			int count;
			if (group == 0)
				count = model.getProteinCount();
			else if (group == 1)
				count = model.getLipidCountExtra();
			else if (group == 2)
				count = model.getLipidCountIntra();
			else if (group == 3)
				count = model.getExampleLipidsCountExtra();
			else
				count = model.getExampleLipidsCountIntra();

			for (int i = 0; i < count; i++) {

				Molecule_3D molecule;
				if (group == 0)
					molecule = (Protein_3D) model.getProteinAt(i);
				else if (group == 1)
					molecule = (Lipid_3D) model.getLipidAtExtra(i);
				else if (group == 2)
					molecule = (Lipid_3D) model.getLipidAtIntra(i);
				else if (group == 3)
					molecule = (Lipid_3D) model.getExampleLipidAtExtra(i);
				else
					molecule = (Lipid_3D) model.getExampleLipidAtIntra(i);

				final Element instance = new Element(
						MembraneXmlReader.TAGNAMES_MEMBRANE[6]);
				proteins.addContent(instance);

				if (group == 0) {
					if (((Protein_3D) molecule).getID() != null) {
						final Element ele = new Element(
								MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
						ele.setAttribute("Id", ((Protein_3D) molecule).getID());
						instance.addContent(ele);
					}
				}
				if (molecule.getPDBFile() != null) {

					final String path = molecule.getPDBFile().getAbsolutePath();

					final Element ele = new Element(
							MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
					ele.setAttribute("PdbFile", path);
					instance.addContent(ele);
				}
				if (molecule.getCurrentProfileName() != null) {

					final String prof = molecule.getCurrentProfileName();

					final Element ele = new Element(
							MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
					ele.setAttribute("Profile", prof);
					instance.addContent(ele);
				}

				
//				MyBounds bb = new MyBounds(molecule.allAtomsSorted);

				Point3f newPos=new Point3f(molecule.getPosition());
//				newPos.add(bb.center);
				
				final Element tr = new Element(
						MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
				tr.setAttribute("Translation", "" + newPos.y);
				instance.addContent(tr);

				
				
				
				if (molecule.getPosition() != null) {
					
					
					
//					for(int a=0;a<molecule.allAtomsSorted.length;a++) {
//						newPoints[a]=Cm2_Toolkit.transformToNewPoint();
//					}
					
					final Element ele = new Element(
							MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
					ele.setAttribute("Pos_X", "" + newPos.x);
					ele.setAttribute("Pos_Z", "" + newPos.z);
					instance.addContent(ele);
				}

				final Element rot = new Element(
						MembraneXmlReader.TAGNAMES_MEMBRANE[7]);

				final Matrix3f m = molecule.getRotationMatrix();
				final String mstring = m.m00 + " " + m.m01 + " " + m.m02 + " "
						+ m.m10 + " " + m.m11 + " " + m.m12 + " " + m.m20 + " "
						+ m.m21 + " " + m.m22;
				rot.setAttribute("Matrix3f", mstring);
				instance.addContent(rot);

				if (molecule.getColor() != null) {
					final Element ele = new Element(
							MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
					ele
							.setAttribute("Color", ""
									+ molecule.getColor().getRGB());
					instance.addContent(ele);
				}
				
				final Element elemU = new Element(
						MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
				elemU
						.setAttribute("UseAtomColors", ""
								+ molecule.isUsingAtomColors());
				instance.addContent(elemU);
				
				if (molecule.getModelSet() != -1) {
					final Element ele = new Element(
							MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
					ele.setAttribute("Model", "" + molecule.getModelSet());
					instance.addContent(ele);
				}
			
				final Element elem = new Element(
						MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
				elem.setAttribute("DomainID", "" + molecule.getDomainID());
				instance.addContent(elem);

				final Element elem2 = new Element(
						MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
				elem2.setAttribute("LayerID", "" + molecule.getLayerID());
				instance.addContent(elem2);

				if (group > 0) {
					final Element ele = new Element(
							MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
					ele.setAttribute("Quantity", ""
							+ ((Lipid_3D) molecule).getPercent());
					instance.addContent(ele);
				}

//				final Atom[] altered = DataPool
//						.getAtomsWithAlteredPositions(molecule);
				

				
				final Atom_3D[] orig = molecule.allAtomsSorted;

				for (int a = 0; a < orig.length; a++) {

//					for (int b = 0; b < altered.length; b++) {
//						if (orig[a] == altered[b]) {

							final Element elem3 = new Element(
									MembraneXmlReader.TAGNAMES_MEMBRANE[7]);
							elem3.setAttribute("Altered_Atom", a + " "
									+ orig[a].position.x + " "
									+ orig[a].position.y + " "
									+ orig[a].position.z);
							instance.addContent(elem3);

//							break;
//						}

//					}

				}

			}
		}

		final Enumeration en = DataPool.rawPdbData.keys();

		while (en.hasMoreElements()) {
			final String key = (String) en.nextElement();
			final String[] data = DataPool.rawPdbData.get(key);

			final Element rawdata = new Element("RAWDATA");
			rawdata.addContent(key);
			rawdata.addContent(CM2_CONSTANTS.IO.RAWDATA_SEPERATOR);
			for (int i = 0; i < data.length; i++) {

				rawdata.addContent(data[i]);
				rawdata.addContent(CM2_CONSTANTS.IO.RAWDATA_SEPERATOR);
			}

			rootElement.addContent(rawdata);

		}

		// Documents gets written here
		final XMLOutputter outputter = new XMLOutputter();
		outputter.setFormat(this.format);
		final FileOutputStream output = new FileOutputStream(file);
		outputter.output(doc, output);
		output.close();
	}

	/**
	 * Returns the Comment String that should always be written into the XML
	 * file. The passed name will appear in this comment, too
	 * 
	 * @param name
	 * @return
	 */
	private String getComment(final String name) {
		final String comment = "This Document describes the Membrane named\n\n\""
				+ name
				+ "\"\n\nBeware: uncareful editing of this file may result in loss of Data\n\n"
				+ "Tag Contents:\n\n"
				+ MembraneXmlReader.TAGNAMES_MEMBRANE[0]
				+ " = All Protein Instances in the Membrane\n"
				+ MembraneXmlReader.TAGNAMES_MEMBRANE[1]
				+ " = All Lipid Instances on Extracellular side\n"
				+ MembraneXmlReader.TAGNAMES_MEMBRANE[2]
				+ " = All Lipid Instances on Inracellular side\n"
				+ MembraneXmlReader.TAGNAMES_MEMBRANE[3]
				+ " = Each Reference Lipid on Extracellular side\n"
				+ MembraneXmlReader.TAGNAMES_MEMBRANE[4]
				+ " = Each Reference Lipid on Intracellular side\n";
		return comment;
	}
}
