/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.io;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.SwingConstants;
import javax.swing.border.TitledBorder;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileView;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import org.cellmicrocosmos.cm2.dataobjects.Atom_3D;
import org.cellmicrocosmos.cm2.dataobjects.MembraneModel;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.dataobjects.MyBounds;
import org.cellmicrocosmos.cm2.dataobjects.Particle;
import org.cellmicrocosmos.cm2.engine.DataPool;
import org.cellmicrocosmos.cm2.gui.MyFCAccessory;
import org.cellmicrocosmos.cm2.gui.dialogs.StatusDialog;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.CoordinateTransformationHints;
import org.cellmicrocosmos.cm2.misc.Cm2_IconDepot;

/**
 * This class is used to parse a once saved PDB file against the original Cm2 file.
 * It is also possible to recreate a PDB file saved with CmME directly, if the 
 * Reverse Parsing information was stored in the PDB file. In this case, the whole
 * Cm2-XML-file is saved in the PDB file in the PDB comments section.
 * 
 * @author Tim Dingersen dackelnose@gmx.net, Bjoern Sommer bjoern@CELLmicrocosmos.org
 *
 */
public class ReversePDBParser {

	File tempFile;
	File pdbfile;
	public StatusDialog dia;
	private boolean showStatusDialog;

	public boolean redoAtomsShiftedToOtherSide = true;

	public ReversePDBParser(File tempDir, boolean showStatusDialog) {

		this.showStatusDialog = showStatusDialog;
		if (tempDir == null)
			tempDir = new File(System.getProperty("user.home"));

		this.tempFile = new File(tempDir + "tmp.xml");
		if (this.tempFile.exists() == false)
			this.tempFile = new File(tempDir + File.separator + "tmp.xml");

	}

	public MembraneModel restoreMembraneModelFromPdbFile(final File pdbFile) throws Exception {
		return restoreMembraneModelFromPdbFile(pdbFile,null, null);
	}
	
	public MembraneModel restoreMembraneModelFromPdbFile(final File pdbFile, final File cm2File, final CoordinateTransformationHints overwriteHints)
			throws Exception {

		this.pdbfile = pdbFile;
		try {
			this.restoreXmlFile(pdbFile);
			return this.load();
		} catch (Exception e) {
			if (!e.getMessage().startsWith("The file does"))
				throw e;

			File actualCm2File = null;
			
			int val = JOptionPane.YES_OPTION;
			
			final JFileChooser fc = new JFileChooser();
			
			fc.setMultiSelectionEnabled(false);
			fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
			fc.setAcceptAllFileFilterUsed(false);

			final FileFilter filter = new FileFilter() {

				@Override
				public boolean accept(final File arg0) {
					if (arg0.isDirectory())
						return true;

					if ((!arg0.getName().toLowerCase().endsWith(".cm2")))
						return false;

					return true;
				}

				@Override
				public String getDescription() {
					return "Membrane Description (.cm2)";
				}

			};

			fc.setFileFilter(filter);

			// customizing a fileview
			final FileView fv = new FileView() {

				@Override
				public Icon getIcon(final File arg0) {

					if (arg0.isDirectory())
						return null;

					return Cm2_IconDepot.getSpecifiedIcon("FC_Membrane");

				}
			};

			fc.setFileView(fv);

			fc.setAccessory(new MyFCAccessory(fc, 0, "Membrane Info"));

			fc.setCurrentDirectory(pdbFile.getParentFile());

			fc.setDialogTitle("Choose the corresponding membrane file");
			
			if (cm2File==null) {
			
				val = JOptionPane
						.showConfirmDialog(
								null,
								"The file does not contain the required information!\n\nThis can have two reasons:\n"
										+ "-The option \"Write additional information for reverse parsing\" has been disabled or\n"
										+ "-The relevant part of the REMARK section has been cleared by other applications\n\n"
										+ "In case the corresponding and unchanged .Cm2-file is still available, the Membrane Editor\n"
										+ "can regain the missing data.\n"
										+ "Do you want to proceed?",
								"Missing Information", JOptionPane.YES_NO_OPTION,
								JOptionPane.ERROR_MESSAGE);
	
				if (val == JOptionPane.NO_OPTION)
					return null;
	
				
				if (fc.showOpenDialog(null) == JFileChooser.CANCEL_OPTION)
	
					return null;
				
				actualCm2File = fc.getSelectedFile();
			
			} else {
				
				actualCm2File = cm2File;
				
			}
				

			if (showStatusDialog) {
				dia = new StatusDialog(true);
				dia.setProgress("Reading files...", 0, true);

			}
			MembraneModel originalModel;
			MembraneXmlReader reader = new MembraneXmlReader();
			reader.readMembrane(actualCm2File);
			if (showStatusDialog)
				dia.setProgress("Reading files...", 15, true);
			originalModel = reader.getResultingMembraneModel();
			
			if (overwriteHints==null) {
				originalModel.setSortByType(DataPool.CoordinateTransformationSettings.sortByType);
			} else {
				originalModel.setSortByType(overwriteHints.sortByType);
//				originalModel.setComputeAbsoluteValues(overwriteHints.computeAbsoluteValues);
			}
			
			// Molecule_3D[]mm=originalModel.getAllMoleculesInMembrane();
			// for(int i=0;i<mm.length;i++) {
			// mm[i].sortAtoms();
			// }
			//			
			// for(int i=0;i<originalModel.getExampleLipidsCountExtra();i++) {
			// originalModel.getExampleLipidAtExtra(i).sortAtoms();
			// }
			// for(int i=0;i<originalModel.getExampleLipidsCountIntra();i++) {
			// originalModel.getExampleLipidAtIntra(i).sortAtoms();
			// }

			PdbReader_3D pdbreader = new PdbReader_3D();

			String[] modifiedContent = pdbreader.readModel(pdbreader
					.readFile(pdbFile), 0)[0];

			if (showStatusDialog)
				dia.setProgress("Merging...", 30, true);
			Vector<Atom_3D> modifiedAtoms = new Vector<Atom_3D>();
			for (int i = 0; i < modifiedContent.length; i++) {
				try {
					Atom_3D a = new Atom_3D(modifiedContent[i]);

					String name = a.name.toUpperCase();
					
					// ignore all water atoms HW1, HW2 and OW and 
					// water coarse grained particles W and ions NA+, 
					if (name.startsWith("HW") || name.startsWith("OW") || 
						name.equals("W") || 
						name.equals("NA+") || name.equals("CL-"))
						continue;

					modifiedAtoms.addElement(a);
				} catch (Exception e2) {
					continue;
				}
			}

			Vector<Molecule_3D[]> possibleMolecules = new Vector<Molecule_3D[]>();

			if (originalModel.countAtoms(-1, -1, -1, false) == modifiedAtoms 
					.size()) {
				possibleMolecules.addElement(originalModel
						.getAllMoleculesInMembrane(-1, -1));
			}

			if (possibleMolecules.isEmpty()) {

				for (int l = -1; l < originalModel.getLayerCount(); l++) {
					int layerID;
					if (l == -1)
						layerID = 0;
					else
						layerID = originalModel.getLayer(l).getLayerId();

					if (originalModel.countAtoms(-1, -1, layerID, false) == modifiedAtoms
							.size()) {

						possibleMolecules.addElement(originalModel
								.getAllMoleculesInMembrane(-1, layerID));
					}

					for (int d = -1; d < originalModel.countMicroDomains(); d++) {
						int domainID;
						if (d == -1)
							domainID = 0;
						else
							domainID = originalModel.getMicroDomain(d).id;
						if (d != -1
								&& originalModel.getMicroDomain(d).layerId != layerID)
							continue;

						if (originalModel.countAtoms(-1, domainID, layerID,
								false) == modifiedAtoms.size()) {

							possibleMolecules.addElement(originalModel
									.getAllMoleculesInMembrane(domainID,
											layerID));
						}

					}

				}
			}

			if (showStatusDialog)
				dia.setProgress("Merging...", 40, true);

			Vector<Molecule_3D[]> definiteMolecules = new Vector<Molecule_3D[]>();

			boolean goon = true;
			for (int a = 0; a < possibleMolecules.size(); a++) {
				Molecule_3D[] possible = possibleMolecules.get(a);
				int c = 0;
				
				for (int i = 0; i < possible.length; i++) {
					Atom_3D[] originalAtoms = possible[i].atoms;
					
					for (int at = 0; at < originalAtoms.length; at++) {
						
//						System.out.println(" Compare for Molecule "+a+"|"+i+
//								": "+originalAtoms[at].name.trim().toUpperCase()+
//								" "+modifiedAtoms.get(c).name.trim().toUpperCase());
						if (!originalAtoms[at].name.trim().equalsIgnoreCase(
							modifiedAtoms.get(c).name.trim())) {
							goon = false;
							break;
						}

						c++;

					}
					if (goon == false)
						break;
				}

				if (goon == true)
					definiteMolecules.addElement(possible);

			}

			// show the ReDo dialog, if the PDB could not be parsed against the original data
			if (definiteMolecules.size() != 1) {
				if (showStatusDialog)
					dia.dispose();
				
				val = JOptionPane
						.showOptionDialog(
								null,
								"Cannot merge the data properly!\n\n"
								+ "Maybe you have chosen the wrong .Cm2-file or \n"
								+ "the "+(DataPool.CoordinateTransformationSettings.sortByType?"Sort-By-Type":"Sort-By-Side")+" setting for reparsing the membrane is wrong?\n"
								+ "Do you want to try again and change to "+(!DataPool.CoordinateTransformationSettings.sortByType?"Sort-By-Type":"Sort-By-Side")+" setting?",
								"BackwardParsing ERROR!",
								JOptionPane.YES_NO_OPTION,
								JOptionPane.ERROR_MESSAGE,
								null,
								new String[] {"Try Again with "+(!DataPool.CoordinateTransformationSettings.sortByType?"Sort-By-Type":"Sort-By-Side"),"Try Again", "Cancel" },
								null);

				if (val < 2) {
					if (val == 0)
						DataPool.CoordinateTransformationSettings.sortByType =
							!DataPool.CoordinateTransformationSettings.sortByType;
					return this.restoreMembraneModelFromPdbFile(pdbFile);
				} else {
					return null;
				}

			}

			CoordinateTransformationHints hints = 
					this.doBackwardsTransformation(modifiedAtoms, pdbFile,originalModel, overwriteHints);
			if (hints == null) {
				if (showStatusDialog)
					dia.dispose();
				return null;
			}

			if (showStatusDialog)
				dia.setProgress("Merging...", 50, true);

			Molecule_3D[] mols = definiteMolecules.firstElement();

			float step = 50f / (float) mols.length;
			float status = 50;

			int C = 0;

			for (int i = 0; i < mols.length; i++) {

				final Molecule_3D mol = mols[i];

				for (int j = 0; j < mol.atoms.length; j++) {

					// System.out.println(mol.atoms[j].name+"   "+modifiedAtoms.get(C).name+"  "+(mol.atoms[j]
					// instanceof Particle));

					if (mol.atoms[j] instanceof Particle == false) {
						mol.applyAtomPosition(j, modifiedAtoms.get(C).position);
						C++;

					} else {

						mol.applyAtomPosition(j, modifiedAtoms.get(C).position);

						Atom_3D[] included = ((Particle) mol.atoms[j]).includedAtoms
								.toArray(new Atom_3D[0]);

						MyBounds mb = new MyBounds(included);

						Vector3f trans = new Vector3f(mol.atoms[j].position);
						trans.sub(mb.center);

						for (int k = 0; k < included.length; k++) {
							included[k].position.add(trans);
						}

						C++;

					}

				}

				mol.updateGeometry();

				status += step;
				if (showStatusDialog)
					dia.setProgress("Merging...", (int) status, true);
			}

			if (hints.shiftY == CoordinateTransformationHints.SHIFTY_AUTO_MEDIAN) {
				float medianValue = 0;
				int moleculeCount = 0;
				for (int i = 0; i < mols.length; i++) {
					medianValue = medianValue + mols[i].getYTranslation();
					moleculeCount ++;
				}
				float computedMedian = (medianValue!=0)?new Float(-medianValue/moleculeCount):0;
				System.out.println("computedMedian "+computedMedian);
				for (int i = 0; i < mols.length; i++) {
					mols[i].setYTranslation(mols[i].getYTranslation()+computedMedian);
				}
			}
			
			// int c = 0;
			// for (int i = 0; i < mols.length; i++) {
			//
			//
			//
			// for (int a = 0; a < mols[i].getAlignedAtoms().length; a++) {
			//
			// //
			// System.out.println(mols[i].getAlignedAtoms()[a].name+" "+mols[i].getAlignedAtoms()[a].position+" | "+modifiedAtoms.get(c).name+"  "+modifiedAtoms.get(c).position);
			// mols[i].applyAtomPosition(a, modifiedAtoms.get(c).position);
			//
			//					
			// c++;
			// }
			//				
			//				
			//				
			//
			// mols[i].updateGeometry();
			//
			// status += step;
			// if (showStatusDialog)
			// dia.setProgress("Merging...", (int) status, true);
			// }

			if (this.redoAtomsShiftedToOtherSide)
				this.redoAtomsShiftedToOtherSide(mols);

			if (showStatusDialog)
				dia.dispose();
			val = JOptionPane
					.showConfirmDialog(
							null,
							"The coordinates have been merged successfully!\n\n"
									+ "Do you wish to update the pdb file now?\n(Backup will be created)",
							"Restore Completed!", JOptionPane.YES_NO_OPTION,
							JOptionPane.QUESTION_MESSAGE);

			if (val == JOptionPane.YES_OPTION) {

				fc.setCurrentDirectory(pdbFile.getParentFile());

				String backup = pdbFile.getAbsolutePath();
				int cc = 1;
				while (true) {
					backup = pdbFile.getAbsolutePath() + File.separator + "Backup"
							+ (cc < 2 ? "" : cc) + " of " + pdbFile.getName();
					if (!new File(backup).exists())
						break;
					cc++;
				}

				fc.setSelectedFile(new File(backup));

				FileFilter filter2 = new FileFilter() {

					@Override
					public boolean accept(final File arg0) {
						if (arg0.getName().toLowerCase().endsWith(".pdb")
								|| arg0.isDirectory())
							return true;
						else
							return false;

					}

					@Override
					public String getDescription() {
						return ".pdb";
					}

				};

				FileView fv2 = new FileView() {

					@Override
					public Icon getIcon(final File arg0) {

						if (arg0.isDirectory())
							return null;

						return Cm2_IconDepot.getSpecifiedIcon("FC_Pdb_Ok");
					}
				};

				fc.setFileView(fv2);

				fc.setAccessory(new MyFCAccessory(fc, 2, "Pdb Info"));

				fc.setFileFilter(filter2);

				fc.setDialogTitle("Choose a backup file");

				if (fc.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {

					File backupfile = fc.getSelectedFile();
					File toChange = new File(pdbFile.getAbsolutePath());
					if (!pdbFile.renameTo(backupfile)) {
						JOptionPane.showMessageDialog(null,
								"Cannot rename the file \n" + pdbFile.getName()
										+ "\ninto\n" + backupfile.getName(),
								"ERROR", JOptionPane.ERROR_MESSAGE);
						return originalModel;
					}

					boolean reverseParsingBefore = DataPool.LimitationSettings.reverseParsing;

					MembraneModel oldModel = DataPool.getMembraneModel();
					try {
						DataPool.LimitationSettings.reverseParsing = true;
						DataPool.replaceMembraneModel(originalModel);
						PDBWriter writer = new PDBWriter(toChange, null);
						writer.writeMembrane(-1, -1, hints);
						DataPool.LimitationSettings.reverseParsing = reverseParsingBefore;
						DataPool.replaceMembraneModel(oldModel);
					} catch (Exception e2) {
						DataPool.LimitationSettings.reverseParsing = reverseParsingBefore;
						DataPool.replaceMembraneModel(oldModel);
						throw e2;
					}
				}

			}

			return originalModel;

		}

	}

	private CoordinateTransformationHints doBackwardsTransformation(
			Vector<Atom_3D> modifiedAtoms, File file, MembraneModel originalModel, CoordinateTransformationHints overwriteHints) {

		CoordinateTransformationHints hints = (overwriteHints==null)?this.getHints(file):overwriteHints;

		if (hints == null) {
			MyTransformationDialog dialog = new MyTransformationDialog();
			hints = dialog.getResult();
			if (hints == null)
				return null;
		}

		if (hints.swapYZ) {
			for (int i = 0; i < modifiedAtoms.size(); i++) {
				Atom_3D a = modifiedAtoms.get(i);

				a.position = new Point3f(a.position.x, a.position.z,
						a.position.y);
			}
		}

		if (hints.shiftXZ == 2) {

			int subX = 0;
			int subZ = 0;
			subX += originalModel.getSize().width / 2;
			subZ += originalModel.getSize().height / 2;

			for (int i = 0; i < modifiedAtoms.size(); i++) {

				Atom_3D a = modifiedAtoms.get(i);

				a.position = new Point3f(a.position.x + subX, a.position.y,
						a.position.z + subZ);
				
			}

		}

		float maxx = -1000000, maxy = -1000000, maxz = -1000000, minx = 1000000, miny = 1000000, minz = 1000000;

		Molecule_3D[] mols = originalModel.getAllMoleculesInMembrane(hints.sortByType);

		for (int i = 0; i < mols.length; i++) {

			Atom_3D[] atoms = mols[i].getAtoms(true, false);
			for (int j = 0; j < atoms.length; j++) {
				
				final Point3f p = atoms[j].position;
				minx = Math.min(minx, p.x);
				miny = Math.min(miny, p.y);
				minz = Math.min(minz, p.z);

				maxx = Math.max(maxx, p.x);
				maxy = Math.max(maxy, p.y);
				maxz = Math.max(maxz, p.z);

			}
		}

		if (hints.shiftXZ == 1 || hints.shiftY > 0) {

			final float cx = hints.shiftXZ == 1 ? (minx + ((maxx - minx) / 2))
					: 0;
			float cy = hints.shiftY == 1 ? (miny + ((maxy - miny) / 2)) : 0;
			final float cz = hints.shiftXZ == 1 ? (minz + ((maxz - minz) / 2))
					: 0;

			if (hints.shiftY == 2)
				
				cy = miny;
				
			for (int i = 0; i < modifiedAtoms.size(); i++) {

				Atom_3D a = modifiedAtoms.get(i);

				a.position = new Point3f(a.position.x + cx, a.position.y + cy,
						a.position.z + cz);

			}

		}
		
		return hints;
	}

	private void restoreXmlFile(final File file) throws Exception {

		final Vector<String> input = new Vector<String>();

		final BufferedReader br = new BufferedReader(new FileReader(file));

		boolean started = false;
		boolean finished = false;
		while (true) {
			final String s = br.readLine();

			if (s == null)
				break;

			if (started == false
					&& s.trim().contains(
							"COPY OF THE MEMBRANE MODEL BEGINS HERE")) {
				started = true;
				br.readLine();
				continue;
			}
			if (s.trim().contains("COPY OF THE MEMBRANE MODEL ENDS HERE")) {
				finished = true;
				break;
			}

			if (started) {
				input.addElement(s.replaceFirst("REMARK", ""));
			}

		}
		br.close();

		if ((started && finished) == false) {
			throw new Exception(
					"The file does not contain the required information");
		}

		final BufferedWriter bw = new BufferedWriter(new FileWriter(
				this.tempFile));

		for (int i = 0; i < input.size(); i++) {

			bw.write(input.get(i).trim());
			bw.newLine();

		}

		this.tempFile.deleteOnExit();
		bw.close();
		input.clear();
		System.gc();

	}

	private MembraneModel load() throws Exception {

		if (showStatusDialog) {
			dia = new StatusDialog(true);
			dia.setProgress("PDB reverse parsing (this may take a while)...", 0, true);

		}
		try {

			final MembraneXmlReader reader = new MembraneXmlReader();

			reader.readMembrane(this.tempFile);
			final MembraneModel newMembrane = reader
					.getResultingMembraneModel();

			final PdbReader_3D pdbr = new PdbReader_3D();
			final String[] contents = pdbr.readFile(this.pdbfile);

			if (showStatusDialog)
				dia.setProgress("Reading...", 25, true);
			final String[] model = pdbr.readModel(contents, 0)[0];

			final Vector<Atom_3D> atoms = new Vector<Atom_3D>();
			for (int i = 0; i < model.length; i++) {
				try {

					atoms.addElement(new Atom_3D(model[i]));
				} catch (final Exception e) {

					continue;
				}
			}

			if (newMembrane.countAtoms(-1, -1, -1, false) != atoms.size())
				throw new Exception(
						"Cannot restore due to mismatching atom counts!");

			// Molecule_3D[]mm=newMembrane.getAllMoleculesInMembrane();
			// for(int i=0;i<mm.length;i++) {
			// mm[i].sortAtoms();
			// }
			//			
			// for(int i=0;i<newMembrane.getExampleLipidsCountExtra();i++) {
			// newMembrane.getExampleLipidAtExtra(i).sortAtoms();
			// }
			// for(int i=0;i<newMembrane.getExampleLipidsCountIntra();i++) {
			// newMembrane.getExampleLipidAtIntra(i).sortAtoms();
			// }

			if (this.doBackwardsTransformation(atoms, pdbfile, newMembrane, null) == null) {

				if (showStatusDialog)
					dia.dispose();
				return null;

			}

			if (showStatusDialog)
				dia.setProgress("Reading...", 50, true);
			
			CoordinateTransformationHints hints = this.getHints(pdbfile);
			
			final Molecule_3D[] mols = newMembrane.getAllMoleculesInMembrane(hints.sortByType);
			
			float step = 50f / (float) mols.length;
			float status = 50;

			int C = 0;

			for (int i = 0; i < mols.length; i++) {

				final Molecule_3D mol = mols[i];

				for (int j = 0; j < mol.atoms.length; j++) {

					// System.out.println(mol.atoms[j].name+"   "+atoms.get(C).name+"  "+(mol.atoms[j]
					// instanceof Particle));

					if (mol.atoms[j] instanceof Particle == false) {
						mol.applyAtomPosition(j, atoms.get(C).position);
						C++;

					} else {

						mol.applyAtomPosition(j, atoms.get(C).position);

						Atom_3D[] included = ((Particle) mol.atoms[j]).includedAtoms
								.toArray(new Atom_3D[0]);

						MyBounds mb = new MyBounds(included);

						Vector3f trans = new Vector3f(mol.atoms[j].position);
						trans.sub(mb.center);

						for (int k = 0; k < included.length; k++) {
							included[k].position.add(trans);
						}

						C++;

					}

				}

				mol.updateGeometry();

				status += step;
				if (showStatusDialog)
					dia.setProgress("Updating Atoms...", (int) status, true);
			}

			//			
			// int c = 0;
			// for (int i = 0; i < mols.length; i++) {
			//				
			// final Molecule_3D mol = mols[i];
			//				
			// mol.sortAtoms();
			// final int atomc = mol.atoms.length;
			//				
			// Atom[]aa=new Atom[mol.atoms.length];
			//
			// for (int j = c; j < c + atomc; j++) {
			//					
			// aa[j]=atoms.get(j);
			//						 
			//						
			//
			// }
			// c += atomc;
			//				
			// Arrays.sort(aa);
			//				
			// for (int j = 0; j < aa.length ; j++) {
			//					
			// final Atom a = aa[j];
			//						 
			// System.out.println(mol.atoms[j]+"  "+a);
			// mol.applyAtomPosition(j, a.getPosition());
			//						
			//					
			//
			// }
			//				
			//
			// mol.updateGeometry();
			//
			// status += step;
			// if (showStatusDialog)
			// dia.setProgress("Updating Atoms...", (int) status, true);
			// }

			if (this.redoAtomsShiftedToOtherSide)
				this.redoAtomsShiftedToOtherSide(mols);

			if (showStatusDialog)
				dia.dispose();
			return newMembrane;
		} catch (final Exception e) {

			e.printStackTrace();
			if (showStatusDialog)
				dia.dispose();
			throw e;
		}

	}

	public CoordinateTransformationHints getHints(File file) {
		CoordinateTransformationHints hints = new CoordinateTransformationHints();
		PdbReader_3D pdbreader = new PdbReader_3D();

		try {
			String[] contents = pdbreader.readFile(file);

			boolean b1 = false;
			boolean b2 = false;
			boolean b3 = false;
			boolean b4 = false;
			for (int i = 0; i < contents.length; i++) {
				String s = contents[i];

				if (s.trim().startsWith("REMARK    SHIFTXZ=")) {

					hints.shiftXZ = Integer.parseInt(s.replaceAll(
							"REMARK    SHIFTXZ=", "").trim());
					b1 = true;
				}
				if (s.trim().startsWith("REMARK    SHIFTY=")) {
					hints.shiftY = Integer.parseInt(s.replaceAll(
							"REMARK    SHIFTY=", "").trim());
					b2 = true;
				}
				if (s.trim().startsWith("REMARK    SWAPYZ=")) {
					hints.swapYZ = Boolean.parseBoolean(s.replaceAll(
							"REMARK    SWAPYZ=", "").trim());
					b3 = true;
				}
				if (s.trim().startsWith("REMARK    SORTBYTYPE=")) {
					hints.sortByType = Boolean.parseBoolean(s.replaceAll(
							"REMARK    SORTBYTYPE=", "").trim());
					b4 = true;
				}
				if (s.trim().startsWith("REMARK    COMPUTEABSOLUTEVALUES=")) {
					hints.computeAbsoluteValues = Boolean.parseBoolean(s.replaceAll(
							"REMARK    COMPUTEABSOLUTEVALUES=", "").trim());
				}

			}

			if (!(b1 && b2 && b3 && b4))
				return null;

			return hints;
		} catch (Exception e) {

			return null;
		}

	}

	class MyTransformationDialog extends JDialog implements ActionListener {

		private static final long serialVersionUID = 1L;

		JButton ok = new JButton("Ok");

		JButton cancel = new JButton("Abort");

		CoordinateTransformationHints ret = null;

		TransformationsPanel tpanel = new TransformationsPanel();

		public MyTransformationDialog() {

			this.ok.addActionListener(this);
			this.cancel.addActionListener(this);

			final JPanel main = new JPanel();
			final JPanel bar = new JPanel();

			main.setLayout(new BoxLayout(main, BoxLayout.X_AXIS));

			main.add(this.tpanel);

			final JPanel header = new JPanel();
			header.setLayout(new BoxLayout(header, BoxLayout.X_AXIS));

			final JLabel hlabel = new JLabel(
					"<html>No hints on former coordinate transformation settings<br>could be found in the file!<br><br>"
							+ "In order to restore the membrane properly, you need<br>to remember the correct settings and to enter<br>"
							+ "them below</html>", SwingConstants.LEADING);

			header.add(Box.createHorizontalStrut(8));
			header.add(hlabel);
			header.add(Box.createHorizontalStrut(8));
			header.add(Box.createVerticalStrut(100));

			this.getContentPane().add(header, BorderLayout.NORTH);
			this.getContentPane().add(main, BorderLayout.CENTER);
			this.getContentPane().add(bar, BorderLayout.SOUTH);

			bar.setLayout(new BoxLayout(bar, BoxLayout.X_AXIS));
			bar.add(Box.createHorizontalGlue());
			bar.add(this.ok);
			bar.add(Box.createHorizontalStrut(10));
			bar.add(this.cancel);
			bar.add(Box.createHorizontalGlue());

			this.setTitle("Coordinate Transformations");

			this.pack();
			this.setLocationRelativeTo(null);
			this.setResizable(false);
			this.ok.requestFocusInWindow();
			this.setAlwaysOnTop(true);
			this.setModal(true);
			this.setVisible(true);
			
			ret.sortByType = DataPool.CoordinateTransformationSettings.sortByType;

		}

		public CoordinateTransformationHints getResult() {

			return ret;
		}

		private class TransformationsPanel extends JPanel {

			private static final long serialVersionUID = 1L;
			JRadioButton radio1 = new JRadioButton("None");
			JRadioButton radio2 = new JRadioButton("To physical center");
			JRadioButton radio3 = new JRadioButton(
					"To center by defined bounds");
			JRadioButton radio4 = new JRadioButton("None");
			JRadioButton radio5 = new JRadioButton("To physical center");
			JRadioButton radio6 = new JRadioButton(
					"To the bottom (only positive y-values)");

			JCheckBox check = new JCheckBox("Swap Y/Z axis  (Gromacs, e.g.)");

			JPanel cpPanel = new JPanel();

			JPanel otherPanel = new JPanel();

			public TransformationsPanel() {

				this.setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
				this.add(this.cpPanel);
				this.add(Box.createVerticalStrut(8));
				this.add(this.otherPanel);
				this.add(Box.createVerticalGlue());

				this.cpPanel.setLayout(new BoxLayout(this.cpPanel,
						BoxLayout.Y_AXIS));
				this.otherPanel.setLayout(new BoxLayout(this.otherPanel,
						BoxLayout.Y_AXIS));
				final ButtonGroup bg = new ButtonGroup();
				bg.add(this.radio1);
				bg.add(this.radio2);
				bg.add(this.radio3);
				final ButtonGroup bg2 = new ButtonGroup();
				bg2.add(this.radio4);
				bg2.add(this.radio5);
				bg2.add(this.radio6);

				radio1.setSelected(true);
				radio4.setSelected(true);

				this.cpPanel.setBorder(BorderFactory.createTitledBorder(null,
						"Origin of Coordinates",
						TitledBorder.DEFAULT_JUSTIFICATION,
						TitledBorder.DEFAULT_POSITION, this.getFont()
								.deriveFont(Font.BOLD,
										this.getFont().getSize2D() + 2f)));
				this.otherPanel.setBorder(BorderFactory.createTitledBorder(
						null, "Other Options",
						TitledBorder.DEFAULT_JUSTIFICATION,
						TitledBorder.DEFAULT_POSITION, this.getFont()
								.deriveFont(Font.BOLD,
										this.getFont().getSize2D() + 2f)));

				final JPanel subj1 = new JPanel();
				final JPanel subj2 = new JPanel();

				subj1.setLayout(new BoxLayout(subj1, BoxLayout.Y_AXIS));
				subj2.setLayout(new BoxLayout(subj2, BoxLayout.Y_AXIS));

				this.cpPanel.add(subj1);
				this.cpPanel.add(subj2);

				subj1.setBorder(BorderFactory
						.createTitledBorder("Planar Shifting"));
				subj2.setBorder(BorderFactory
						.createTitledBorder("Height Shifting"));

				final JLabel sl1 = new JLabel("<html>"
						+ "The shifting of the center on the XZ plane</html>");

				final JLabel sl2 = new JLabel("<html>"
						+ "The shifting of the center on the Y Axis</html>");

				final Color fg = this.getForeground().brighter();
				this.radio1.setForeground(fg);
				this.radio2.setForeground(fg);
				this.radio3.setForeground(fg);
				this.radio4.setForeground(fg);
				this.radio5.setForeground(fg);
				this.radio6.setForeground(fg);
				this.check.setForeground(fg);

				subj1.add(sl1);

				subj1.add(this.radio1);
				subj1.add(this.radio2);
				subj1.add(this.radio3);

				subj2.add(sl2);

				subj2.add(this.radio4);
				subj2.add(this.radio5);
				subj2.add(this.radio6);

				final JPanel subj3 = new JPanel();

				subj3.setLayout(new BoxLayout(subj3, BoxLayout.Y_AXIS));

				this.otherPanel.add(Box.createVerticalStrut(5));
				this.otherPanel.add(subj3);
				this.otherPanel.add(Box.createVerticalStrut(5));

				subj3.add(this.check);
				subj3.add(Box.createHorizontalGlue());

			}

		}

		public void actionPerformed(final ActionEvent arg0) {

			if (arg0.getSource() == this.ok) {

				ret = new CoordinateTransformationHints();

				if (tpanel.radio1.isSelected())
					ret.shiftXZ = CoordinateTransformationHints.SHIFTXZ_NONE;
				else if (tpanel.radio2.isSelected())
					ret.shiftXZ = CoordinateTransformationHints.SHIFTXZ_PHYSICAL_CENTER;
				else
					ret.shiftXZ = CoordinateTransformationHints.SHIFTXZ_BOUNDS_CENTER;

				if (tpanel.radio4.isSelected())
					ret.shiftY = CoordinateTransformationHints.SHIFTY_NONE;
				else if (tpanel.radio5.isSelected())
					ret.shiftY = CoordinateTransformationHints.SHIFTY_PHYSICAL_CENTER;
				else
					ret.shiftY = CoordinateTransformationHints.SHIFTY_BOTTOM;

				ret.swapYZ = tpanel.check.isSelected();

				this.dispose();
			}
			if (arg0.getSource() == this.cancel) {
				ret = null;
				this.dispose();
			}

		}

	}

	private void redoAtomsShiftedToOtherSide(Molecule_3D[] mols) {
		Vector<Point3f> allPoints = new Vector<Point3f>();
		for (int i = 0; i < mols.length; i++) {
			Atom_3D[] atoms = mols[i].getAtoms(true, false);
			for (int j = 0; j < atoms.length; j++) {
				allPoints.addElement(atoms[j].position);
			}
		}

		MyBounds mainBounds = new MyBounds(allPoints.toArray(new Point3f[0]));

		for (int m = 0; m < mols.length; m++) {

			Molecule_3D mol = mols[m];
			if (mol.positionAfterReading == null)
				continue;

			for (int r = 0; r < 2; r++) {

				Atom_3D[] atoms = mol.getAtoms(true, true);

				int w = 10;
				int h = 10;
				int oc = 0;
				while (true) {
					Rectangle rect = new Rectangle(
							(int) mol.positionAfterReading.x - (w / 2),
							(int) mol.positionAfterReading.z - (h / 2), w, h);

					oc++;
					for (int i = 0; i < atoms.length; i++) {

						if (atoms[i] == null)
							continue;

						if (rect.contains(new Point((int) atoms[i].position.x,
								(int) atoms[i].position.z))) {
							atoms[i] = null;
							oc = 0;
						}
					}

					w += 10;
					h += 10;
					if (oc > 4)
						break;
				}

				boolean completelymoved = true;
				for (int i = 0; i < atoms.length; i++) {
					if (atoms[i] == null) {
						completelymoved = false;
						break;
					}
				}
				if (completelymoved)
					continue;

				for (int i = 0; i < atoms.length; i++) {

					if (atoms[i] == null)
						continue;

					float NS = mol.positionAfterReading.z - atoms[i].position.z;
					float WE = mol.positionAfterReading.x - atoms[i].position.x;

					if (Math.abs(NS) >= Math.abs(WE)) {

						if (NS > 0) {
							// from North
							Point3f newPos = new Point3f(atoms[i].position.x,
									atoms[i].position.y, atoms[i].position.z
											+ mainBounds.widthZ);
							mol.applyAtomPosition(i, newPos);
						} else {
							// from South
							Point3f newPos = new Point3f(atoms[i].position.x,
									atoms[i].position.y, atoms[i].position.z
											- mainBounds.widthZ);
							mol.applyAtomPosition(i, newPos);
						}
					} else {
						if (WE > 0) {
							// from West
							Point3f newPos = new Point3f(atoms[i].position.x
									+ mainBounds.widthX, atoms[i].position.y,
									atoms[i].position.z);
							mol.applyAtomPosition(i, newPos);
						} else {
							// from East
							Point3f newPos = new Point3f(atoms[i].position.x
									- mainBounds.widthX, atoms[i].position.y,
									atoms[i].position.z);
							mol.applyAtomPosition(i, newPos);
						}
					}

				}

			}
			mol.updateGeometry();

		}

	}
}
