/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.io.ppe.network.http;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;

import org.cellmicrocosmos.cm2.io.ppe.toolbox.ConfParser;
import org.cellmicrocosmos.cm2.io.ppe.toolbox.PPE_CONSTANTS.CONF_PARSER;
import org.cellmicrocosmos.cm2.io.ppe.toolbox.PPE_CONSTANTS.CONNECTION;
import org.cellmicrocosmos.cm2.io.ppe.toolbox.PPE_CONSTANTS.IO;
import org.cellmicrocosmos.cm2.misc.Cm2_Toolkit;


/**
 * The PdbTMConnection establishes a connection to the
 * Protein Data Bank of Transmembrane Proteins.With the
 * help of some methods a local copy or an InputStream
 * can be generated to the pdbtmall.xml. With the
 * getCountOf-methods data bank statistics can be displayed.
 * 
 * @author Christian Gamroth
 * 
 */
public class PdbTMConnection extends Connection {

	/**
	 * Constructs a new PdbTMConnection.
	 */
	public PdbTMConnection() {
				
		// initialize connection
		try {
			final ConfParser confparser = new ConfParser();
			this.setList(confparser.parseForPortConfig(CONF_PARSER.PDBTM_KEY));
			confparser.freeSpace();
			this.setUrl(new URL(this.getList().get(0), this.getList().get(1), Integer
					.parseInt(this.getList().get(2)), this.getList().get(3)));
		} catch (final MalformedURLException e) {
			System.err
					.println(">> MalformedURLException at PDBTMConnection(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		} catch (final Exception e) {
			System.err
					.println(">> Exception at PDBTMConnection(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		}
	}

	/**
	 * Obtains the number of all entries in the Protein Data Bank.
	 * 
	 * @return a String of the number of all entries in the Protein Data Bank
	 */
	public String getCountOfPdbAll() {
		try {
			final URL tempurl = new URL(this.getUrl(), CONNECTION.SUBDIR_PDB_ALL);
			return this.getBody(tempurl);
		} catch (final MalformedURLException e) {
			System.err
					.println(">> MalformedURLException at getCountOfPdbAll(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		} catch (final Exception e) {
			System.err
					.println(">> Exception at getCountOfPdbAll(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		}
		return null;
	}

	/**
	 * Obtains the number of all entries in the Protein Data Bank of Transmembrane
	 * Proteins.
	 * 
	 * @return a String of the number of all entries in the Protein Data Bank of
	 * Transmembrane Proteins
	 */
	public String getCountOfPdbTM() {
		try {
			final URL tempurl = new URL(this.getUrl(), CONNECTION.SUBDIR_PDBTM_ALL);
			return this.getBody(tempurl);
		} catch (final MalformedURLException e) {
			System.err
					.println(">> MalformedURLException at getCountOfPdbTM(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		} catch (final Exception e) {
			System.err
					.println(">> Exception at getCountOfPdbTM(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		}
		return null;
	}

	/**
	 * Obtains the number of all entries with alpha helices in the Protein Data
	 * Bank of Transmembrane Proteins.
	 * 
	 * @return a String of all entries with alpha helices in the Protein Data Bank
	 * of Transmembrane Proteins
	 */
	public String getCountOfPdbTMalpha() {
		try {
			final URL tempurl = new URL(this.getUrl(), CONNECTION.SUBDIR_PDBTM_ALPHA);
			return this.getBody(tempurl);
		} catch (final MalformedURLException e) {
			System.err
					.println(">> MalformedURLException at getCountOfPdbTMalpha(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		} catch (final Exception e) {
			System.err
					.println(">> Exception at getCountOfPdbTMalpha(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		}
		return null;
	}

	/**
	 * Obtains the number of all entries with beta barrel in the Protein Data Bank
	 * of Transmembrane Proteins.
	 * 
	 * @return a String of all entries with beta barrel in the Protein Data Bank of
	 * Transmembrane Proteins
	 */
	public String getCountOfPdbTMbeta() {
		try {
			final URL tempurl = new URL(this.getUrl(), CONNECTION.SUBDIR_PDBTM_BETA);
			return this.getBody(tempurl);
		} catch (final MalformedURLException e) {
			System.err
					.println(">> MalformedURLException at getCountOfPdbTMbeta(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		} catch (final Exception e) {
			System.err
					.println(">> Exception at getCountOfPdbTMbeta(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		}
		return null;
	}

	/**
	 * Retrieves the content of a given URL.
	 * 
	 * @param tempurl The URL for looking at.
	 * @return a String contains the content of the given URL
	 */
	private String getBody(final URL tempurl) {
		try {
			final HttpURLConnection con = (HttpURLConnection) tempurl
					.openConnection();
			con.setDoInput(true);
			con.setRequestProperty(CONNECTION.REQUEST_KEY_AUTHORIZATION,
					this.getList().get(4));
			con.connect();
			final Reader instream = new InputStreamReader(con.getInputStream());
			final BufferedReader input = new BufferedReader(instream);
			final String body = input.readLine();
			instream.close();
			input.close();
			con.disconnect();
			return body;
		} catch (final IOException e) {
			System.err.println(">> IOException at getBody(URL): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		} catch (final Exception e) {
			System.err.println(">> Exception at getBody(URL): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		}
		return null;
	}

	/**
	 * Generates an InputStream to the pdbtmall.xml.
	 * 
	 * @return an InputStream to the pdbtmall.xml
	 */
	public InputStream getInputStreamFromXML() {
		try {
			final URL tempurl = new URL(this.getUrl(),
					CONNECTION.SUBDIR_PDBTM_XMLFILE);
			final HttpURLConnection con = (HttpURLConnection) tempurl
					.openConnection();
			con.setDoInput(true);
			con.setRequestProperty(CONNECTION.REQUEST_KEY_AUTHORIZATION,
					this.getList().get(4));
			con.connect();
			return con.getInputStream();
		} catch (final MalformedURLException e) {
			System.err
					.println(">> MalformedURLException at getInputStreamFromXML(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		} catch (final IOException e) {
			System.err.println(">> IOException at getInputStreamFromXML(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		} catch (final Exception e) {
			System.err.println(">> Exception at getInputStreamFromXML(): ");
			System.err.println(e);
			Cm2_Toolkit.complainInternet();
		}
		return null;
	}

	/**
	 * Generates a local copy to the pdbtmall.xml.
	 */
	public void storeXMLOfPdbTM() {
		String tempLine = null;
		try {
			final URL tempurl = new URL(this.getUrl(),
					CONNECTION.SUBDIR_PDBTM_XMLFILE);
			final HttpURLConnection con = (HttpURLConnection) tempurl
					.openConnection();
			con.setDoInput(true);
			con.setRequestProperty(CONNECTION.REQUEST_KEY_AUTHORIZATION,
					this.getList().get(4));
			con.connect();
			final Reader instream = new InputStreamReader(con.getInputStream());
			final BufferedReader input = new BufferedReader(instream);
			final File xmlFile = new File(IO.SETTINGS_DIR, IO.XML_FILENAME_PDBTM);
			// TODO TempFile handeln.
			if (xmlFile.exists()) {
				xmlFile.delete();
				xmlFile.createNewFile();
			}
			final PrintWriter writer = new PrintWriter(new FileWriter(xmlFile));
			while ((tempLine = input.readLine()) != null) {
				writer.println(tempLine);
			}
			xmlFile.setReadOnly();
//			System.out.println(">> File erfolgreich erstellt!");
			instream.close();
			input.close();
			writer.close();
			con.disconnect();
		} catch (final MalformedURLException e) {
			System.err
					.println(">> MalformedURLException at storeXMLOfPdbTM(): ");
			System.err.println(e);
			Cm2_Toolkit.complainXML(IO.XML_FILENAME_PDBTM);
		} catch (final IOException e) {
			System.err.println(">> IOException at storeXMLOfPdbTM(): ");
			System.err.println(e);
			Cm2_Toolkit.complainXML(IO.XML_FILENAME_PDBTM);
		} catch (final Exception e) {
			System.err.println(">> Exception at storeXMLOfPdbTM(): ");
			System.err.println(e);
			Cm2_Toolkit.complainXML(IO.XML_FILENAME_PDBTM);
		}
	}

}
