/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.io.ppe.toolbox;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.JOptionPane;
import javax.vecmath.Matrix3f;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.cellmicrocosmos.cm2.constants.CM2_CONSTANTS;
import org.cellmicrocosmos.cm2.dataobjects.Atom_3D;
import org.cellmicrocosmos.cm2.dataobjects.Molecule_3D;
import org.cellmicrocosmos.cm2.dataobjects.Protein_3D;
import org.cellmicrocosmos.cm2.io.ppe.toolbox.PPE_CONSTANTS.IO;
import org.cellmicrocosmos.cm2.io.ppe.toolbox.PPE_CONSTANTS.MIFILE;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.MembranePackingAlgorithm;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.Protein;
import org.cellmicrocosmos.cm2.misc.Cm2_Toolkit;
import org.cellmicrocosmos.cm2.misc.Cm2_Toolkit.ToolkitMembranePackingAlgorithm;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;


/**
 * The MIFHandler reads in a MIFile and generates a Hashmap
 * with the required transformation matrixes. The class inherits
 * from the DefaultHandler and implements some of his methods.
 * 
 * @author Christian Gamroth, Bjoern Sommer bjoern@CELLmicrocosmos.org
 *
 */
public class MIFHandler extends DefaultHandler {

	private SAXParserFactory factory = null;
	private Map<String, Matrix3f> matrixmap_pdbtm;
	private Map<String, Vector3f> vectormap_pdbtm;
	private Map<String, Boolean> boolmap_opm;
	private String hashkey = null;
	private File mifile_pdbtm;
	private File mifile_opm;
	private SAXParser parser;
	private Matrix3f rotationMatrix;
	private Vector3f transformVector;

	/**
	 * Constructs a new MIFHandler. 
	 */
	public MIFHandler() {
		try {
			
			factory = SAXParserFactory.newInstance();

			// PDBTM
			this.mifile_pdbtm = new File(IO.SETTINGS_DIR, IO.MIF_FILENAME_PDBTM);
			this.matrixmap_pdbtm = new HashMap<String, Matrix3f>();
			this.vectormap_pdbtm = new HashMap<String, Vector3f>();
			this.parser = factory.newSAXParser();
			this.makePDBTMHashMap();
			
			// OPM
			this.mifile_opm = new File(IO.SETTINGS_DIR, IO.MIF_FILENAME_OPM);
			this.boolmap_opm = new HashMap<String, Boolean>();
			this.makeOPMHashMap();
			
		} catch (final ParserConfigurationException e) {
			System.out
					.println(">> ParserConfigurationException at MIFHandler(): ");
			System.out.println(e);
		} catch (final SAXException e) {
			System.out.println(">> SAXException at MIFHandler(): ");
			System.out.println(e);
		}
		
		// activate this to compare PDBTM and OPM
//		this.comparePDBTMtoOPM();
		
	}

	/**
	 * Generates a new Hashmap with transformation matrixes.
	 */
	private void makePDBTMHashMap() {
		try {
			if(this.mifile_pdbtm!=null && this.mifile_pdbtm.exists())
				this.parser.parse(this.mifile_pdbtm, this);
			else
				JOptionPane.showMessageDialog(
						null,
						"The pdbtm.mif file is missing.\nPlease go to the PDB Download Menu and press Update PDBTM.",
						"ERROR", JOptionPane.ERROR_MESSAGE);
		} catch (final SAXException e) {
			System.out.println(">> SAXException at makeHashMap().");
			if (mifile_pdbtm.exists()) {
				mifile_pdbtm.deleteOnExit();
				System.out.println(">> Corrupt pdbtm.mif file was deleted. Just Download it again using the PDB Update Window. ");
			}
			e.printStackTrace();
		} catch (final IOException e) {
			System.out.println(">> IOException at makeHashMap(): ");
			e.printStackTrace();
		} 
	}

	/**
	 * Generates a new Hashmap with boolean values.
	 */
	private void makeOPMHashMap() {
		try {
//			this.parser = this.factory.newSAXParser();
			if(this.mifile_opm!=null && this.mifile_opm.exists())
				this.parser.parse(this.mifile_opm, this);
			else
				JOptionPane.showMessageDialog(
						null,
						"The opm.mif file is missing.\nPlease go to the PDB Download Menu and press Update OPM.",
						"ERROR", JOptionPane.ERROR_MESSAGE);
		} catch (final SAXException e) {
			System.out.println(">> SAXException at makeHashMap().");
			if (mifile_opm.exists()) {
				mifile_opm.deleteOnExit();
				System.out.println(">> Corrupt opm.mif file was deleted. Just Download it again using the PDB Update Window. ");
			}
			e.printStackTrace();
		} catch (final IOException e) {
			System.out.println(">> IOException at makeHashMap(): ");
			e.printStackTrace();
		}
	}
	
	/**
	 * 
	 */
	@Override
	public void startElement(final String uri, final String localName,
			final String qName, final Attributes attributes) {
		if (qName.equals(MIFILE.NEW_ENTRY)) {
			this.rotationMatrix = new Matrix3f();
			this.transformVector = new Vector3f();
		}
		if (qName.equals(MIFILE.PDBTM_ENTRY)) {
			this.hashkey = attributes.getValue(0).toLowerCase();
		}
		if (qName.equals(MIFILE.OPM_ENTRY)) {
			this.hashkey = attributes.getValue(0).toLowerCase();
			this.rotationMatrix = null;
		}
		if (qName.equals(MIFILE.NORMAL_ENTRY)) {
			/* place for doing something with normal
			   Vector3f vector = new Vector3f(Float.parseFloat(attributes.getValue(0))
					, Float.parseFloat(attributes.getValue(1))
					, Float.parseFloat(attributes.getValue(2))); */
		}
		if (qName.equals(MIFILE.ROWX_ENTRY)) {
			this.rotationMatrix.setRow(0, Float.parseFloat(attributes.getValue(0)),
					Float.parseFloat(attributes.getValue(1)),
					Float.parseFloat(attributes.getValue(2)));
			this.transformVector.x = Float.parseFloat(attributes.getValue(3));
		}
		// has to swap YZ
		if (qName.equals(MIFILE.ROWY_ENTRY)) {
			this.rotationMatrix.setRow(2, Float.parseFloat(attributes.getValue(0)),
					Float.parseFloat(attributes.getValue(1)),
					Float.parseFloat(attributes.getValue(2)));
			this.transformVector.z = Float.parseFloat(attributes.getValue(3));
		}
		// has to swap YZ
		if (qName.equals(MIFILE.ROWZ_ENTRY)) {
			this.rotationMatrix.setRow(1, -Float.parseFloat(attributes.getValue(0)),
					-Float.parseFloat(attributes.getValue(1)),
					-Float.parseFloat(attributes.getValue(2)));
			this.transformVector.y = -Float.parseFloat(attributes.getValue(3));
		}
	}

	/**
	 * 
	 */
	@Override
	public void endElement(final String uri, final String localName,
			final String qName) {
		if (qName.equals(MIFILE.NEW_ENTRY)) {
			if (this.rotationMatrix != null) {
				this.matrixmap_pdbtm.put(this.hashkey, this.rotationMatrix);
				if (this.transformVector != null)
					this.vectormap_pdbtm.put(this.hashkey, this.transformVector);
			} else {
				this.boolmap_opm.put(this.hashkey, true);
			}
		}
	}

	/**
	 * Checks whether the given PDB-Id in the MIFile exists.
	 * 
	 * @param pdbID The PDB-Id of the protein for looking at.
	 * @return <code>true</code> if the PDB-Id in the MIFile exists
	 */
	public boolean isElementOfMIF(final String pdbID) {
		return ( this.matrixmap_pdbtm.containsKey(pdbID.toLowerCase()) || this.boolmap_opm.containsKey(pdbID.toLowerCase()) );
	}
	

	/**
	 * Checks whether the given PDB-Id in the PDBTM entry of the MIFile exists.
	 * 
	 * @param pdbID The PDB-Id of the protein for looking at.
	 * @return <code>true</code> if the PDB-Id in the MIFile exists
	 */
	public boolean isElementOfPDBTM(final String pdbID) {
		return this.matrixmap_pdbtm.containsKey(pdbID.toLowerCase());
	}
	

	/**
	 * Checks whether the given PDB-Id in the OPM entry of the MIFile exists.
	 * 
	 * @param pdbID The PDB-Id of the protein for looking at.
	 * @return <code>true</code> if the PDB-Id in the MIFile exists
	 */
	public boolean isElementOfOPM(final String pdbID) {
		return this.boolmap_opm.containsKey(pdbID.toLowerCase());
	}

	/**
	 * Obtains the transform matrix of a given PDB-Id.
	 *  
	 * @param pdbID The PDB-Id of the protein.
	 * @return a Matrix3f representing the transformation matrix
	 */
	public Matrix3f getTransformMatrixOf(final String pdbID) {
		if (this.isElementOfPDBTM(pdbID)) {
			return this.matrixmap_pdbtm.get(pdbID.toLowerCase());
		} else {
			return null;
		}
	}
	
	/**
	 * Obtains the transformation vector of a given PDB-Id.
	 *  
	 * @param pdbID The PDB-Id of the protein.
	 * @return a Matrix3f representing the transformation matrix
	 */
	public Vector3f getTransformVectorOf(final String pdbID) {
		if (this.isElementOfPDBTM(pdbID)) {
			return this.vectormap_pdbtm.get(pdbID.toLowerCase());
		} else {
			return null;
		}
	}
	
	/**
	 * This methods adjust the rotation and the position of the given
	 * Protein to the actual membrane which it is placed into.
	 * 
	 * @param newProt The Protein_3D to align
	 * @author Bjoern Sommer bjoern@CELLmicrocosmos.org
	 */
	public void applyOPMAlignmentTo(Molecule_3D newProt) {
		Atom_3D[] allAtomsOfCurrentMolecule = newProt.opmAtoms;
		
		// try to get three atoms of each layer in order to compute the plane
		//  only atomOX layer: extracellular protein
		//  only atomNX layer: intracellular protein
		//  both atom layers : transmembrane protein
		Atom_3D atomN1 = null; 
		Atom_3D atomO1 = null; 
		int atomN1Index = -1;
		int atomO1Index = -1;
		for (int i=0;i<allAtomsOfCurrentMolecule.length;i++) {
			
			if(allAtomsOfCurrentMolecule[i].name.equals("N")) { 
				if (atomN1==null) {
					atomN1 = allAtomsOfCurrentMolecule[i];
					atomN1Index = i;
				} else {
					break;
				}
			} else if(allAtomsOfCurrentMolecule[i].name.equals("O")) {
				if (atomO1==null) {
					atomO1 = allAtomsOfCurrentMolecule[i];
					atomO1Index = i;
				} else {
					break;
				}
			}
			
		}
	
		// change the rotation matrix of the protein by applying the standard OPM rotations
		Matrix3f current = Cm2_Toolkit.createAUnitMatrix();
		Cm2_Toolkit.rotate_Z_Axis_Matrix(current, new Float(Math.toDegrees(Math.PI)));
		// adjust the rotation around X axis for extracellular proteins
		if (atomN1==null && atomO1!=null) {
			Cm2_Toolkit.rotate_X_Axis_Matrix(current, new Float(Math.toDegrees(Math.PI))+270f);
		// adjust the rotation around X axis for intracellular and transmembrane proteins
		} else {
			Cm2_Toolkit.rotate_X_Axis_Matrix(current, new Float(Math.toDegrees(Math.PI))+90f);
		}
		Cm2_Toolkit.rotate_Y_Axis_Matrix(current, new Float(Math.toDegrees(Math.PI)));

		// assign the rotation and the y=0 translation to the protein
		newProt.setYTranslation(0);
		newProt.setRotationMatrix(current);
		newProt.move();
		newProt.recalculatePoints();
		
		Atom_3D[] alignedOPMs=newProt.getAlignedOPMAtoms();
		// move into the center of the two dummy layers
		if (atomN1!=null && atomO1!=null) {
			if(alignedOPMs[atomN1Index].position.y>alignedOPMs[atomO1Index].position.y) {
				// check if the molecule is rotated correctly
				Cm2_Toolkit.rotate_Z_Axis_Of_CoordinateSystem(current, 180f);
				// assign the roation and the y=0 translation to the protein
				newProt.setYTranslation(0);
				newProt.setRotationMatrix(current);
				newProt.move();
				newProt.recalculatePoints();
			}
			
			alignedOPMs=newProt.getAlignedOPMAtoms();
			Point3d boundingBoxUpper = new Point3d();
			newProt.getAlignedGeometricBounds().getUpper(boundingBoxUpper);
				
			newProt.setYTranslation( new Float(Math.abs(boundingBoxUpper.y)) - 
									 ((alignedOPMs[atomO1Index].getPosition().y + alignedOPMs[atomN1Index].getPosition().y)/2) );
			newProt.move();
			newProt.recalculatePoints();
			
		// move on the extracellular side downwards
		} else if (atomN1==null && atomO1!=null){
			
			// move the protein to the starting position at the top level
			Point3d boundingBoxLower = new Point3d();
			newProt.getAlignedGeometricBounds().getLower(boundingBoxLower);
			float boundingBoxOStartUnder0 = new Float(Math.abs(boundingBoxLower.y)) / CM2_CONSTANTS.J3D.J3D_ANGSTROM_SCALE;
			float dummyLayerOStartUnder0 = boundingBoxOStartUnder0+alignedOPMs[atomO1Index].getPosition().y;
			newProt.setYTranslation(36f+boundingBoxOStartUnder0);
			newProt.move();
			newProt.recalculatePoints();
			newProt.getAlignedGeometricBounds().getLower(boundingBoxLower);
			// if the membrane contains extracellular lipids try to align the protein, 
			// first in bigger, than in smaller steps (for higher accuracy)
			ToolkitMembranePackingAlgorithm proteinMoverAlgo = new ToolkitMembranePackingAlgorithm();
			if (proteinMoverAlgo.hasMembraneAnyExtracellularLipids() && proteinMoverAlgo.getAllLipids(MembranePackingAlgorithm.EXTRA).size()>0) {
				// bigger steps (-1)
				while (!proteinMoverAlgo.intersectsWithLipid(new Protein((Protein_3D) newProt,proteinMoverAlgo), MembranePackingAlgorithm.EXTRA) &&
						boundingBoxLower.y>0) {
					newProt.setYTranslation(newProt.getYTranslation()-1f);
					newProt.move();
					newProt.recalculatePoints();
					newProt.getAlignedGeometricBounds().getLower(boundingBoxLower);
				}
				// redo last step
				newProt.setYTranslation(newProt.getYTranslation()+1f);
				newProt.move();
				newProt.recalculatePoints();
				newProt.getAlignedGeometricBounds().getLower(boundingBoxLower);
				// smaller steps (-0.01)
				while (!proteinMoverAlgo.intersectsWithLipid(new Protein((Protein_3D) newProt,proteinMoverAlgo), MembranePackingAlgorithm.EXTRA) &&
						boundingBoxLower.y>0) {
					newProt.setYTranslation(newProt.getYTranslation()-0.1f);
					newProt.move();
					newProt.recalculatePoints();
					newProt.getAlignedGeometricBounds().getLower(boundingBoxLower);
				}
				// if the protein did not contact any lipid, set it back to the standard position
				// otherwise move it downwards so that the dummy layer matches the lipid intersection point on the top
				if (boundingBoxLower.y<=0) {
					newProt.setYTranslation(0);
					newProt.move();
					newProt.recalculatePoints();
					newProt.setYTranslation(36f+boundingBoxOStartUnder0-dummyLayerOStartUnder0);	
					newProt.move();
					newProt.recalculatePoints();
				} else {
					newProt.setYTranslation(newProt.getYTranslation()-
							dummyLayerOStartUnder0);
					newProt.move();
					newProt.recalculatePoints();
				}
			}
	
		// move on the intracellular side upwards
		} else if (atomN1!=null && atomO1==null){
	
			// move the protein to the starting position at the bottom level
			Point3f boundingBoxUpper = newProt.getUpper(true);
			float boundingBoxNStartOver0 = Math.abs(boundingBoxUpper.y);
//			float dummyLayerNStartOver0 = boundingBoxNStartOver0-newProt.getAlignedAtoms()[atomN1Index].getPosition().y;
			float dummyLayerNStartOver0 = boundingBoxNStartOver0-alignedOPMs[atomN1Index].getPosition().y;
//			float dummyLayerOStartUnder0 = boundingBoxOStartUnder0+alignedOPMs[atomO1Index].getPosition().y;
			newProt.setYTranslation(-36f-boundingBoxNStartOver0);
			newProt.move();
			newProt.recalculatePoints();
			boundingBoxUpper = newProt.getUpper(true);
			// if the membrane contains intracellular lipids try to align the protein, 
			// first in bigger, than in smaller steps (for higher accuracy)
			ToolkitMembranePackingAlgorithm proteinMoverAlgo = new ToolkitMembranePackingAlgorithm();
			if (proteinMoverAlgo.hasMembraneAnyIntracellularLipids() && proteinMoverAlgo.getAllLipids(MembranePackingAlgorithm.INTRA).size()>0) {
				// bigger steps (+1)
				while (!proteinMoverAlgo.intersectsWithLipid(new Protein((Protein_3D) newProt,proteinMoverAlgo), MembranePackingAlgorithm.INTRA) &&
						boundingBoxUpper.y<0) {
					newProt.setYTranslation(newProt.getYTranslation()+1f);
					newProt.move();
					newProt.recalculatePoints();
					boundingBoxUpper = newProt.getUpper(true);
				}
				// redo last step
				newProt.setYTranslation(newProt.getYTranslation()-1f);
				newProt.move();
				newProt.recalculatePoints();
				boundingBoxUpper = newProt.getUpper(true);
				// smaller steps (+0.01)
				while (!proteinMoverAlgo.intersectsWithLipid(new Protein((Protein_3D) newProt,proteinMoverAlgo), MembranePackingAlgorithm.INTRA) &&
						boundingBoxUpper.y<0) {
					newProt.setYTranslation(newProt.getYTranslation()+0.1f);
					newProt.move();
					newProt.recalculatePoints();
					boundingBoxUpper = newProt.getUpper(true);
				}
				// if the protein did not contact any lipid, set it back to the standard position
				// otherwise move it upwards so that the dummy layer matches the lipid intersection point on the bottom
				if (boundingBoxUpper.y>=0) {
					newProt.setYTranslation(0);
					newProt.move();
					newProt.recalculatePoints();	
					newProt.setYTranslation(-36f-boundingBoxNStartOver0+dummyLayerNStartOver0);	
					newProt.move();
					newProt.recalculatePoints();
				} else {
					newProt.setYTranslation(newProt.getYTranslation()+
							dummyLayerNStartOver0);
					newProt.move();
					newProt.recalculatePoints();
				}
			}	
			
		} 
	}

	/**
	 * Use this method to compare both database entries to each other.
	 * The result is given to the console.
	 */
	public void comparePDBTMtoOPM() {
		System.out.println("\nPPE comparison of PDBTM and OPM");
		System.out.println("-------------------------------");
		System.out.println("PDBTM contains "+this.matrixmap_pdbtm.size()+" entries.");
		System.out.println("OPM   contains "+this.boolmap_opm.size()+" entries.");
		HashMap<String,Boolean> entriesInOPMandPDBTM = new HashMap<String,Boolean>();
		Iterator<String> pdbtmKeySetIterator = this.matrixmap_pdbtm.keySet().iterator();
		while(pdbtmKeySetIterator.hasNext()) {
			String actualPDB = pdbtmKeySetIterator.next();
			if (this.boolmap_opm.containsKey(actualPDB))
				entriesInOPMandPDBTM.put(actualPDB, true);
		}
		Iterator<String> opmKeySetIterator = this.boolmap_opm.keySet().iterator();
		while(opmKeySetIterator.hasNext()) {
			String actualPDB = opmKeySetIterator.next();
			if (this.matrixmap_pdbtm.containsKey(actualPDB)) {
				if (!entriesInOPMandPDBTM.containsKey(actualPDB))
					entriesInOPMandPDBTM.put(actualPDB, true);
			}
		}
		System.out.println(entriesInOPMandPDBTM.size()+" PDB entries in all are found in PDBTM as well as OPM.");
	}
			
}
