/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.membranepackingalgorithm;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Area;

import javax.vecmath.Point3f;

import org.cellmicrocosmos.cm2.dataobjects.Lipid_3D;
import org.cellmicrocosmos.cm2.dataobjects.MembraneModel;
import org.cellmicrocosmos.cm2.dataobjects.MicroDomain_3D;
import org.cellmicrocosmos.cm2.engine.DataPool;


/**
 * This class represents a microdomain in the membrane. Objects of this class
 * can be accessed from the membranepackingalgorithm class and be used for better
 * access to molecule members when writing an algorithm that supports
 * microdomains
 * 
 * @author Tim Dingersen dackelnose@gmx.net, Bjoern Sommer bjoern@CELLmicrocosmos.org
 * 
 */
public class MicroDomain {

	private final MicroDomain_3D originalDomain;

	private final Rectangle bounds;

	private final Lipid_3D[] lipidsEC;

	private final Lipid_3D[] lipidsIC;

	private MembranePackingAlgorithm algo;

	/**
	 * Contructs a new MicroDomain object
	 * 
	 * ! NOT MEANT TO BE USED BY PLUGIN DEVELOPERS !
	 * 
	 */
	public MicroDomain(final MicroDomain_3D microDomain, MembranePackingAlgorithm algo) {
		this.originalDomain = microDomain;
		this.algo = algo;
		this.bounds = this.originalDomain.getBoundingBox();
		final MembraneModel model = DataPool.getMembraneModel();
		this.lipidsEC = model.getAllExampleLipidsExtraInDomain(
				this.originalDomain.id, this.originalDomain.layerId);
		this.lipidsIC = model.getAllExampleLipidsIntraInDomain(
				this.originalDomain.id, this.originalDomain.layerId);
	}

	/**
	 * Returns the bounding rectangle of the microdomain's defined area on the
	 * x/z plane
	 * 
	 * @return the bounds of this microdomain.
	 */
	public Rectangle getBounds() {
		return this.bounds;
	}

	/**
	 * Returns the area of the domain geometry.
	 * The geometry may be a rectangle, an ellipse or a free polygon shape. 
	 * @return area of this microdomain
	 */
	public double getArea() {
		return this.originalDomain.getArea();
	}
	
	/**
	 * Tests wether a molecule's center is located inside the defined area of
	 * this microdomain.
	 * 
	 * @param mol the
	 *            molecule that you want to test
	 * @return true if the molecule's position intersects with the domains
	 *         defined area, false otherwise
	 */
	public boolean isInsideArea(final Molecule mol) {
		final Point3f p = mol.getPosition();
		if (new Area(this.originalDomain.geoObject).contains(new Point(
				(int) p.x, (int) p.z)))
			return true;
		return false;
	}

	/**
	 * Creates a new Lipid that can be placed in the membrane. The type of lipid
	 * you will get from this method depends on the lipid template setup of the
	 * extracellular membrane side. The initial rotation will be the way it has
	 * been set up there too. The initial y-translation will be set in a way
	 * that the whole lipid is just above the y-plane
	 * 
	 * 
	 * @return the new lipid object or null if no template lipids exist on this
	 *         side
	 */
	public Lipid createNextExtracellularLipid() {

		return DataPool.getMembraneModel().createNextLipid(
				MembranePackingAlgorithm.EXTRA, this.lipidsEC, this.originalDomain.id,
				this.originalDomain.layerId, algo);
	}

	/**
	 * Creates a new Lipid that can be placed in the membrane. The type of lipid
	 * you will get from this method depends on the lipid template setup of the
	 * intracellular membrane side. The initial rotation will be the way it has
	 * been set up there too. The initial y-translation will be set in a way
	 * that the whole lipid is just underneath the y-plane
	 * 
	 * 
	 * @return the new lipid object object or null if no template lipids exist
	 *         on this side
	 */
	public Lipid createNextIntracellularLipid() {
		return DataPool.getMembraneModel().createNextLipid(
				MembranePackingAlgorithm.INTRA, this.lipidsIC, this.originalDomain.id,
				this.originalDomain.layerId, algo);

	}

	/**
	 * Tests if any extracellular lipids are available for in this microdomain
	 */
	public boolean hasAnyExtracellularLipids() {
		return this.lipidsEC.length > 0;
	}

	/**
	 * Tests if any intracellular lipids are available for in this microdomain
	 */
	public boolean hasAnyIntracellularLipids() {
		return this.lipidsIC.length > 0;
	}

	/**
	 * Returns the Domain_ID for this microdomain
	 */
	public int getDomainID() {
		return this.originalDomain.id;
	}

	/**
	 * Returns the Layer_ID for this microdomain
	 */
	public int getLayerID() {
		return this.originalDomain.layerId;
	}

	/**
	 * Returns a copy of the extracellular lipid template at the specified
	 * index.
	 * 
	 * @param index the
	 *            index of the lipid template
	 * @return a copy of the indexed lipid template
	 */
	public Lipid getSampleLipid_ExtraCellular(final int index) {
		return new Lipid((Lipid_3D) this.lipidsEC[index].copy(true, false),
				algo);
	}

	/**
	 * Returns a copy of the intracellular lipid template at the specified index
	 * 
	 * @param index the
	 *            index of the lipid template
	 * @return a copy of the indexed lipid template
	 */
	public Lipid getSampleLipid_IntraCellular(final int index) {
		return new Lipid((Lipid_3D) this.lipidsIC[index].copy(true, false),
				algo);
	}

	/**
	 * Returns a copy of all lipid templates of the extracellular side of this
	 * microdomain
	 */
	public Lipid[] getAllSampleLipids_ExtraCellular() {
		final Lipid[] ret = new Lipid[this.lipidsEC.length];
		for (int i = 0; i < this.lipidsEC.length; i++) {
			ret[i] = new Lipid((Lipid_3D) this.lipidsEC[i].copy(true, false),
					algo);
		}
		return ret;
	}

	/**
	 * Returns a copy of all lipid templates of the intracellular side of this
	 * microdomain
	 */
	public Lipid[] getAllSampleLipids_IntraCellular() {
		final Lipid[] ret = new Lipid[this.lipidsIC.length];
		for (int i = 0; i < this.lipidsIC.length; i++) {
			ret[i] = new Lipid((Lipid_3D) this.lipidsIC[i].copy(true, false),
					algo);
		}
		return ret;
	}

	/**
	 * Returns the number of lipid templates on the extracellular side of this
	 * microdomain
	 */
	public int countSampleLipids_ExtraCellular() {
		return this.lipidsEC.length;
	}

	/**
	 * Returns the number of lipid templates on the intracellular side of this
	 * microdomain
	 */
	public int countSampleLipids_IntraCellular() {
		return this.lipidsIC.length;
	}
}
