/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.membranepackingalgorithms;

import java.awt.Component;
import java.util.Vector;

import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.cellmicrocosmos.cm2.membranepackingalgorithm.Lipid;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.MembranePackingAlgorithm;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.MembranePackingAlgorithmGUI;


/**
 * This tool algorithm deletes selected intersecting lipids
 * from the bilayer.
 * 
 * @author Bjoern Sommer bjoern@CELLmicrocosmos.org
 *
 */
public class IntersectionDeleter extends MembranePackingAlgorithm {

	// indicates if lipids should be deleted if intersect
	boolean deleteExtraLipids;
	boolean deleteIntraLipids;

	// a flag that can be toggled to abort the algorithm
	boolean stop = false;
	
	int userSelectedDomain = -1;

	/*
	 * This method gets invoked before the fillWithLipids() process is started
	 * This is the point where you should present your GUI, like i'm doing here
	 */
	public boolean buildAndShowGui() {

		// check if the user selected a membrane
		if (this.getSelectedArea() != -1) {

			int val = JOptionPane
					.showConfirmDialog(
							null,
							"You have currently selected a particular area.\nDo you want to consider this area only?\n",
							"Selected Area", JOptionPane.YES_NO_OPTION,
							JOptionPane.QUESTION_MESSAGE);

			if (val == JOptionPane.YES_OPTION) {
				userSelectedDomain = this.getSelectedArea();
			}

		}
		
		// building up the options panel for the GUI
		final JPanel controlPanel = new JPanel();
		final JLabel infoLipids = new JLabel("Delete intersecting lipids: ");
		final JCheckBox checkBoxExtraLipids = new JCheckBox("extracellular");
		checkBoxExtraLipids.setSelected(true);
		final JCheckBox checkBoxIntraLipids = new JCheckBox("intracellular");
		checkBoxIntraLipids.setSelected(true);

		controlPanel.add(infoLipids);
		controlPanel.add(checkBoxExtraLipids);
		controlPanel.add(checkBoxIntraLipids);

		/*
		 * showing the GUI to the user. The returnvalue indicates if the user
		 * confirmed or cancelled the dialog so far
		 */
		final boolean returnValue = MembranePackingAlgorithmGUI.popUpGUI(controlPanel,
				"Intersection Deleter", "Options");

		// returning false if all should be cancelled
		if (returnValue == false)
			return false;

		this.deleteExtraLipids = checkBoxExtraLipids.isSelected();
		this.deleteIntraLipids = checkBoxIntraLipids.isSelected();

		/*
		 * the user settings are ok plus the user must have confirmed if the
		 * thread reaches this code, so true gets returned in order to proceed
		 */
		return true;

	}

	/*
	 * This method gets called when the algorithm gets started. Remember: The
	 * algorithm finishes when the thread leaves this method
	 */
	public void fillWithLipids() throws Exception {

		/*
		 * starting a loop that runs until the terminal condition is reached or
		 * the stop flag has been set on true
		 */
		while (this.stop == false) {

			this.setDisplayedProgress(0);
			this.displayInformation("Clean up on the extracellular side.");
			if (this.deleteExtraLipids) {
				final Vector<Lipid> lipsExtra = this.getAllLipidsInDomain(userSelectedDomain, MembranePackingAlgorithm.EXTRA);
				if (lipsExtra != null)
					for (final Lipid l : lipsExtra) {
						if (this.intersects(l)
								|| this.intersectsWithProteins(l))
							this.removeLipidsFromMembrane(new Lipid[] { l });
					}
			}

			this.setDisplayedProgress(50);
			this.displayInformation("Clean up on the intracellular side.");
			if (this.deleteIntraLipids) {
				final Vector<Lipid> lipsIntra = this.getAllLipidsInDomain(userSelectedDomain, MembranePackingAlgorithm.INTRA);
				if (lipsIntra != null)
					for (final Lipid l : lipsIntra) {
						if (this.intersects(l)
								|| this.intersectsWithProteins(l))
							this.removeLipidsFromMembrane(new Lipid[] { l });
					}
			}
			this.displayInformation("Finished.");
			this.setDisplayedProgress(100);

			this.stop = true;

			if (!(this.deleteExtraLipids && this.deleteIntraLipids))
				JOptionPane
						.showMessageDialog(
								null,
								"You did not clean up both membrane sides -\nso the result maybe still inconsistant!\n(in contrast to the 'Bilayer State' message)",
								"Inconsistance possible", JOptionPane.OK_OPTION);

		}

	}

	public String getFullInformation() {

		// this string is getting displayed in the textarea were the algorithms
		// are selected
		return	"<html>This algorithm deletes all intersecting lipids.<br>" +
				"You can choose if you want to clean up<br>" +
				"the extra- and/or intracellular side.<br><br>" +
				"It may be useful if you added new proteins to<br>" +
				"a membrane containing lipids and you do not want<br>" +
				"to recalculate the whole membrane again.<br><br>" +
				"Author: Bj&ouml;rn Sommer<br>" +
				"Version: 1.0</html>";
	}

	public String getName() {

		// the chosen name for this algorithm. It has to return a unique name in
		// order to get accepted
		return "Intersection Deleter Tool";
	}

	public Component getRuntimeControlPanel() {

		// we don't provide any runtime controls for this simple algorithm
		return null;
	}

	public String getShortInformation() {

		// you can leave out some information if you want
		return "Deletes intersecting lipids. (Version 1.01)";
	}

	public String getStringRepresentationOfSetupParameters() {

		// returning a String that describes the users setup
		return "Clean up extracellular lipids: " + this.deleteExtraLipids
				+ "\n" + "Clean up intracellular lipids: "
				+ this.deleteIntraLipids;
	}

	public boolean overrideExistingLipids() {

		// existing lipids may remain in this case
		return false;
	}

	public void stopAlgorithm() {

		// setting the stop flag on true, forcing the algorithm to stop at the
		// next loop
		this.stop = true;
	}


	public boolean supportsAbsoluteValues() {
		return true;
	}
	
	public boolean supportsMicrodomains() {
		// this algorithm provides a full microdomain support
		return true;
	}
	
}
