/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.membranepackingalgorithms;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.vecmath.Point3f;

import org.cellmicrocosmos.cm2.membranepackingalgorithm.Application_Constants;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.Lipid;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.MembranePackingAlgorithm;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.MembranePackingAlgorithmGUI;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.Protein;


/**
 * This tool algorihtm can be used to shift all membrane components
 * of one type (lipids of the inner and/or outer layer and/or proteins)
 * at once.
 * @author Bjoern Sommer bjoern@CELLmicrocosmos.org
 *
 */
public class MembraneShifter extends MembranePackingAlgorithm {

	// indicates if lipids should be deleted if intersect
	private boolean shiftExtraLipids;
	private boolean shiftIntraLipids;
	private boolean shiftProteins;
	
	private boolean checkMoleculeIntersections;
	private boolean checkBorderIntersections;
	
	private JSpinner membraneShiftValueSpinner = null;
	private JLabel membraneShiftValueSpinnerLabel = new JLabel("<html>&nbsp;Angstrom&nbsp;&nbsp;&nbsp;Presets:&nbsp;</html>");
	
	private JSpinner membraneXShiftValueSpinner = null;
	private JSpinner membraneZShiftValueSpinner = null;
	
	private int intersectionCounter = 0;
	
	private int userSelectedDomain = -1;
	
	// a flag that can be toggled to abort the algorithm
	boolean stop = false;
	
	/*
	 * This method gets invoked before the fillWithLipids() process is started
	 * This is the point where you should present your GUI, like i'm doing here
	 */
	public boolean buildAndShowGui() {
		
		// check if the user selected a membrane
		if (this.getSelectedArea() != -1) {

			int val = JOptionPane
					.showConfirmDialog(
							null,
							"You have currently selected a particular area.\nDo you want to consider this area only?\n",
							"Selected Area", JOptionPane.YES_NO_OPTION,
							JOptionPane.QUESTION_MESSAGE);

			if (val == JOptionPane.YES_OPTION) {
				userSelectedDomain = this.getSelectedArea();
			}

		}
		
		final Vector<Lipid> lipsExtra = this.getAllLipidsInDomain(userSelectedDomain, MembranePackingAlgorithm.EXTRA);
		final Vector<Lipid> lipsIntra = this.getAllLipidsInDomain(userSelectedDomain, MembranePackingAlgorithm.INTRA);
		final Vector<Protein> proteins = this.getAllProteinsInDomain(userSelectedDomain);
		
		final JCheckBox checkBoxExtraLipids = new JCheckBox("Extracellular Lipids");
		final JCheckBox checkBoxIntraLipids = new JCheckBox("Intracellular Lipids");
		final JCheckBox checkBoxProteins = new JCheckBox("Proteins");

		// building up the options panel for the GUI
		JPanel controlPanel = new JPanel();
		controlPanel.setLayout(new BoxLayout(controlPanel, BoxLayout.Y_AXIS));
		
		// the shifter panel
		JPanel panelSpinner = new JPanel();
		panelSpinner.setLayout(new BoxLayout(panelSpinner, BoxLayout.X_AXIS));
		panelSpinner.setBackground(Color.GRAY);
		panelSpinner.add(new JLabel("<html>&nbsp;Y-Shift&nbsp;</html>"));
		//  the spinner 
		membraneShiftValueSpinner = new JSpinner(new SpinnerNumberModel(0, -1*Application_Constants.MAX_Y_TRANSLATION, Application_Constants.MAX_Y_TRANSLATION, 1.0f));
		membraneShiftValueSpinner.setMaximumSize(new Dimension(60, 50));
		membraneShiftValueSpinner.setMinimumSize(new Dimension(60, 50));
		panelSpinner.add(membraneShiftValueSpinner);
		panelSpinner.add(membraneShiftValueSpinnerLabel);
		//  the MINIMUM button
		JButton buttonMinValue = new JButton("TOP");
		buttonMinValue.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				float minValue = Application_Constants.MAX_Y_TRANSLATION;
				if (checkBoxExtraLipids.isSelected() && lipsExtra != null) {
					for (final Lipid l : lipsExtra) {
						if (minValue > l.getYTranslation() || minValue == Application_Constants.MAX_Y_TRANSLATION)
							minValue = l.getYTranslation();
					}
				}
				if (checkBoxIntraLipids.isSelected() && lipsIntra != null) {
					for (final Lipid l : lipsIntra) {
						if (minValue > l.getYTranslation() || minValue == Application_Constants.MAX_Y_TRANSLATION)
							minValue = l.getYTranslation();
					}
				}
				if (checkBoxProteins.isSelected() && proteins != null) {
					for (final Protein p : proteins) {
						if (minValue > p.getYTranslation() || minValue == Application_Constants.MAX_Y_TRANSLATION)
							minValue = p.getYTranslation();
					}
				}
				membraneShiftValueSpinner.setValue(-minValue);
			}
		});
		panelSpinner.add(buttonMinValue);
		//  the MEDIAN button
		JButton buttonMedianValue = new JButton("<html><font color=\"5555FF\">CENTER</font><html>");
		buttonMedianValue.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				float medianValue = 0;
				int moleculeCount = 0;
				if (checkBoxExtraLipids.isSelected() && lipsExtra != null) {
					for (final Lipid l : lipsExtra) {
						medianValue = medianValue + l.getYTranslation();
						moleculeCount ++;
					}
				}
				if (checkBoxIntraLipids.isSelected() && lipsIntra != null) {
					for (final Lipid l : lipsIntra) {
						medianValue = medianValue + l.getYTranslation();
						moleculeCount ++;
					}
				}
				if (checkBoxProteins.isSelected() && proteins != null) {
					for (final Protein p : proteins) {
						medianValue = medianValue + p.getYTranslation();
						moleculeCount ++;
					}
				}
				if (medianValue!=0)
					membraneShiftValueSpinner.setValue(new Float(-medianValue/moleculeCount));
				else
					membraneShiftValueSpinner.setValue(0);
			}
		});
		panelSpinner.add(buttonMedianValue);
		//  the MAXIMUM button
		JButton buttonMaxValue = new JButton("BOTTOM");
		buttonMaxValue.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				float maxValue = -Application_Constants.MAX_Y_TRANSLATION;
				if (checkBoxExtraLipids.isSelected() && lipsExtra != null) {
					for (final Lipid l : lipsExtra) {
						if (maxValue < l.getYTranslation() || maxValue == -Application_Constants.MAX_Y_TRANSLATION)
							maxValue = l.getYTranslation();
					}
				}
				if (checkBoxIntraLipids.isSelected() && lipsIntra != null) {
					for (final Lipid l : lipsIntra) {
						if (maxValue < l.getYTranslation() || maxValue == -Application_Constants.MAX_Y_TRANSLATION)
							maxValue = l.getYTranslation();
					}
				}
				if (checkBoxProteins.isSelected() && proteins != null) {
					for (final Protein p : proteins) {
						if (maxValue < p.getYTranslation() || maxValue == -Application_Constants.MAX_Y_TRANSLATION)
							maxValue = p.getYTranslation();
					}
				}
				membraneShiftValueSpinner.setValue(-maxValue);
			}
		});
		panelSpinner.add(buttonMaxValue);
		
		// the molecule panel
		JPanel panelMolecules = new JPanel();
		panelMolecules.setLayout(new BoxLayout(panelMolecules, BoxLayout.X_AXIS));
		panelMolecules.add(new JLabel("<html>&nbsp;Apply to this molecules:&nbsp;</html>"));
		checkBoxExtraLipids.setSelected(true);
		panelMolecules.add(checkBoxExtraLipids);
		checkBoxIntraLipids.setSelected(true);
		panelMolecules.add(checkBoxIntraLipids);
		checkBoxProteins.setSelected(true);
		panelMolecules.add(checkBoxProteins);

		// the molecule intersection panel
		JPanel panelIntersections = new JPanel();
		panelIntersections.setLayout(new BoxLayout(panelIntersections, BoxLayout.X_AXIS));
		panelIntersections.add(new JLabel("<html>&nbsp;Do not move molecules with&nbsp;</html>"));
		JCheckBox checkBoxMoleculeIntersections = new JCheckBox("molecule intersections");
		checkBoxMoleculeIntersections.setSelected(false);
		panelIntersections.add(checkBoxMoleculeIntersections);
		JCheckBox checkBoxBorderIntersections = new JCheckBox("border intersections");
		checkBoxBorderIntersections.setSelected(false);
		panelIntersections.add(checkBoxBorderIntersections);
		
		// the shifter XZ panel
		JPanel panelXZSpinner = new JPanel();
		panelXZSpinner.setLayout(new BoxLayout(panelXZSpinner, BoxLayout.X_AXIS));
		//  the shifter X panel
		panelXZSpinner.setBackground(Color.GRAY);
		panelXZSpinner.add(new JLabel("<html>&nbsp;X-Shift&nbsp;</html>"));
		//  the X spinner 
		membraneXShiftValueSpinner = new JSpinner(new SpinnerNumberModel(0, -1*Application_Constants.MAX_Y_TRANSLATION, Application_Constants.MAX_Y_TRANSLATION, 1.0f));
		membraneXShiftValueSpinner.setMaximumSize(new Dimension(60, 50));
		membraneXShiftValueSpinner.setMinimumSize(new Dimension(60, 50));
		panelXZSpinner.add(membraneXShiftValueSpinner);
		//  the shifter Z panel
		panelXZSpinner.add(new JLabel("<html>&nbsp;Z-Shift&nbsp;</html>"));
		//  the Z spinner 
		membraneZShiftValueSpinner = new JSpinner(new SpinnerNumberModel(0, -1*Application_Constants.MAX_Y_TRANSLATION, Application_Constants.MAX_Y_TRANSLATION, 1.0f));
		membraneZShiftValueSpinner.setMaximumSize(new Dimension(60, 50));
		membraneZShiftValueSpinner.setMinimumSize(new Dimension(60, 50));
		panelXZSpinner.add(membraneZShiftValueSpinner);

		// build the main panel
		controlPanel.add(panelSpinner);
		controlPanel.add(Box.createVerticalStrut(10));
		controlPanel.add(panelMolecules);
		controlPanel.add(Box.createVerticalStrut(10));
		controlPanel.add(panelIntersections);
		controlPanel.add(Box.createVerticalStrut(10));
		controlPanel.add(panelXZSpinner);
		
		buttonMedianValue.doClick();

		/*
		 * showing the GUI to the user. The returnvalue indicates if the user
		 * confirmed or cancelled the dialog so far
		 */
		final boolean returnValue = MembranePackingAlgorithmGUI.popUpGUI(controlPanel,
				"Membrane Shifter", "Change the Y-Translation of Molecules");

		// returning false if all should be cancelled
		if (returnValue == false)
			return false;

		this.shiftExtraLipids = checkBoxExtraLipids.isSelected();
		this.shiftIntraLipids = checkBoxIntraLipids.isSelected();
		this.shiftProteins = checkBoxProteins.isSelected();
		
		this.checkMoleculeIntersections = checkBoxMoleculeIntersections.isSelected();
		this.checkBorderIntersections = checkBoxBorderIntersections.isSelected();

		/*
		 * the user settings are ok plus the user must have confirmed if the
		 * thread reaches this code, so true gets returned in order to proceed
		 */
		return true;

	}

	/*
	 * This method gets called when the algorithm gets started. Remember: The
	 * algorithm finishes when the thread leaves this method
	 */
	public void fillWithLipids() throws Exception {

		/*
		 * starting a loop that runs until the terminal condition is reached or
		 * the stop flag has been set on true
		 */
		while (this.stop == false) {
			
			this.setDisplayedProgress(0);
			this.displayInformation("Shift the extracellular side.");
			final Vector<Lipid> lipsExtra = this.getAllLipidsInDomain(userSelectedDomain, MembranePackingAlgorithm.EXTRA);
			if (this.shiftExtraLipids) {
				if (lipsExtra != null)
					for (final Lipid l : lipsExtra) {
						l.setPosition(new Point3f(	l.getPosition().x+(Float.parseFloat(membraneXShiftValueSpinner.getValue().toString())), 
													l.getYTranslation()+(Float.parseFloat(membraneShiftValueSpinner.getValue().toString())), 
													l.getPosition().z+(Float.parseFloat(membraneZShiftValueSpinner.getValue().toString()))));
						if (checkBorderIntersections?this.intersectsBorders(l):this.intersectsEarthOrSky(l) || 
							checkMoleculeIntersections && this.intersects(l)) {
							l.setPosition(new Point3f(	l.getPosition().x+(-Float.parseFloat(membraneXShiftValueSpinner.getValue().toString())), 
														l.getYTranslation()+(-Float.parseFloat(membraneShiftValueSpinner.getValue().toString())), 
														l.getPosition().z+(-Float.parseFloat(membraneZShiftValueSpinner.getValue().toString()))));
							intersectionCounter++;
						}
					}
			}

			this.setDisplayedProgress(30);
			this.displayInformation("Shift the intracellular side.");
			final Vector<Lipid> lipsIntra = this.getAllLipidsInDomain(userSelectedDomain, MembranePackingAlgorithm.INTRA);
			if (this.shiftIntraLipids) {
				if (lipsIntra != null)
					for (final Lipid l : lipsIntra) {
						l.setPosition(new Point3f(	l.getPosition().x+(Float.parseFloat(membraneXShiftValueSpinner.getValue().toString())), 
													l.getYTranslation()+(Float.parseFloat(membraneShiftValueSpinner.getValue().toString())), 
													l.getPosition().z+(Float.parseFloat(membraneZShiftValueSpinner.getValue().toString()))));
						if (checkBorderIntersections?this.intersectsBorders(l):this.intersectsEarthOrSky(l) || 
							checkMoleculeIntersections && this.intersects(l)) {
							l.setPosition(new Point3f(	l.getPosition().x+(-Float.parseFloat(membraneXShiftValueSpinner.getValue().toString())), 
														l.getYTranslation()+(-Float.parseFloat(membraneShiftValueSpinner.getValue().toString())), 
														l.getPosition().z+(-Float.parseFloat(membraneZShiftValueSpinner.getValue().toString()))));
							intersectionCounter++;
						}
					}
			}
			
			this.setDisplayedProgress(60);
			this.displayInformation("Shift proteins.");
			final Vector<Protein> proteins = this.getAllProteinsInDomain(userSelectedDomain);
			if (this.shiftProteins) {
				if (proteins != null)
					for (final Protein p : proteins) {
						p.setPosition(new Point3f(	p.getPosition().x+(Float.parseFloat(membraneXShiftValueSpinner.getValue().toString())), 
													p.getYTranslation()+(Float.parseFloat(membraneShiftValueSpinner.getValue().toString())), 
													p.getPosition().z+(Float.parseFloat(membraneZShiftValueSpinner.getValue().toString()))));
						if (checkBorderIntersections?this.intersectsBorders(p):this.intersectsEarthOrSky(p) || 
							checkMoleculeIntersections && this.intersects(p)) {
							p.setPosition(new Point3f(	p.getPosition().x+(-Float.parseFloat(membraneXShiftValueSpinner.getValue().toString())), 
														p.getYTranslation()+(-Float.parseFloat(membraneShiftValueSpinner.getValue().toString())), 
														p.getPosition().z+(-Float.parseFloat(membraneZShiftValueSpinner.getValue().toString()))));
							intersectionCounter++;
						}
					}
			}
			
			this.displayInformation("Finished.");
			this.setDisplayedProgress(100);

			this.stop = true;

//			if (! 	(this.shiftExtraLipids && lipsExtra != null) && 
//					(this.shiftIntraLipids && lipsIntra != null) && 
//					(this.shiftProteins  && proteins != null) 		)
//				JOptionPane
//						.showMessageDialog(
//								null,
//								"You did not move all molecules at once. So (additional) intersections may occure.",
//								"Inconsistance possible", JOptionPane.OK_OPTION);
			
			if (intersectionCounter>0)
				JOptionPane
						.showMessageDialog(
								null,
								intersectionCounter+" Molecules could not be shifted due intersections.\nThe limits of the Y-Shift are +/-"+Application_Constants.MAX_Y_TRANSLATION,
								"Intersections", JOptionPane.OK_OPTION);

		}

	}

	public String getFullInformation() {

		// this string is getting displayed in the textarea were the algorithms
		// are selected
		return 	"<html>This algorithm shifts molecules in the Y-direction.<br>" +
				"Intra-, Extracellular Lipids or Proteins may be chosen<br>" +
				"and the Algorithm may check for other molecule or<br>"+
				"border intersections. Intersecting molecules are<br>"+
				"not shifted, if this option is activated.<br><br>" +
				"There are three presets to shift the molecules to the<br>" +
				"center, top or bottom. The center value is <br>" +
				"automatically set for the Y-Shift. This values are<br>"+
				"based only on the molecule types selected before!<br><br>" +
				"Hint 1: Please be aware that it may make sense to<br>" +
				"deactivate the proteins before computing the center<br>" +
				"of a protein containing lipid layer.<br><br>" +
				"Hint 2: This tool especially may make sense if you<br>" +
				"re-import a membrane which has been simulated<br>" +
				"outside of the MembraneEditor. For this purpose the<br>"+
				"intersection checking modes should be deactivated,<br>" +
				"because usually many re-imported molecules intersect.<br><br>" +
				"Author: Bj&ouml;rn Sommer<br>" +
				"Version: 1.1</html>";
	}

	public String getName() {

		// the chosen name for this algorithm. It has to return a unique name in
		// order to get accepted
		return "Membrane Shifter Tool";
	}

	public Component getRuntimeControlPanel() {

		// we don't provide any runtime controls for this simple algorithm
		return null;
	}

	public String getShortInformation() {

		// you can leave out some information if you want
		return "Changes the Y-Translation of molecule types. (Version 1.1)";
	}

	public String getStringRepresentationOfSetupParameters() {

		// returning a String that describes the users setup
		return "Shift values: " 
				+ "\n" + " Y " + this.membraneShiftValueSpinner.getValue().toString()
				+ "\n" + " X " + this.membraneXShiftValueSpinner.getValue().toString()
				+ "\n" + " Z " + this.membraneZShiftValueSpinner.getValue().toString()
				+ "\n" + "Shift extracellular lipids: " + this.shiftExtraLipids
				+ "\n" + "Shift intracellular lipids: " + this.shiftIntraLipids
				+ "\n" + "Shift proteins: " + this.shiftProteins
				+ "\n" + "Check molecule intersections: " + this.checkMoleculeIntersections
				+ "\n" + "Check border intersections: " + this.checkBorderIntersections
				+ "\n" + "Selected Layer/Domain: "+userSelectedDomain+"/"+this.getCurrentLayerID();
	}

	public boolean overrideExistingLipids() {

		// existing lipids may remain in this case
		return false;
	}

	public void stopAlgorithm() {

		// setting the stop flag on true, forcing the algorithm to stop at the
		// next loop
		this.stop = true;
	}
	
	public boolean supportsAbsoluteValues() {
		// this algorithm does not depend on percental values
		return true;
	}

	public boolean supportsMicrodomains() {
		// this algorithm provides a full microdomain support
		return true;
	}
	
}
