/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.pluginmanager;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Hashtable;
import java.util.Vector;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

import javax.swing.JOptionPane;

import org.cellmicrocosmos.cm2.constants.CM2_CONSTANTS;
import org.cellmicrocosmos.cm2.engine.DataPool;
import org.cellmicrocosmos.cm2.io.FileIOHandler;
import org.cellmicrocosmos.cm2.membranepackingalgorithm.MembranePackingAlgorithm;
import org.cellmicrocosmos.cm2.misc.Cm2_Toolkit;


/**
 * 
 * This class manages the different MembranePackingAlgorithms.
 * 
 * @author Tim Dingersen dackelnose@gmx.net, Bjoern Sommer bjoern@CELLmicrocosmos.org
 * 
 */
public class PluginManager {

	public Vector<Plugin> plugins = new Vector<Plugin>();

	private int sleepCounter = 0; 
	
	public void init() {
		final String[] included = CM2_CONSTANTS.ALGORITHMS.INCLUDED_ALGORITHMS;
		for (int i = 0; i < included.length; i++) {
			
			try {
				
				final MembranePackingAlgorithm a = (MembranePackingAlgorithm) Class.forName(
						included[i]).newInstance();
				
				final Plugin pl = new Plugin();

				pl.setName(a.getName());
				pl.setInformation(a.getFullInformation());
				pl.setJavapath(included[i]);
				pl.setState(Plugin.STATE_READY);
				pl.setSupportsMDs(a.supportsMicrodomains());
				pl.setSupportsAbsoluteValues(a.supportsAbsoluteValues());
				pl.setOverridesLipids(a.overrideExistingLipids());
				pl.setType(Plugin.INCLUDED);

				this.plugins.add(pl);
			} catch (final Exception e) {
				continue;
			}
		}
		Vector<Plugin> userplugins;
		try {
			userplugins = PluginIO.read();
		} catch (final Exception e) {
			System.err.println(e.getMessage());
			return;
		}

		for (int i = 0; i < userplugins.size(); i++) {

			final Plugin pl = userplugins.get(i);

			final boolean b1 = pl.getState().equals(Plugin.STATE_READY)
					|| pl.getState().equals(Plugin.STATE_NO_ALGORITHMS)
					|| pl.getState().equals(Plugin.STATE_FILE_NOT_READABLE)
					|| pl.getState().equals(Plugin.STATE_FILE_NOT_FOUND)
					|| pl.getState().equals(Plugin.STATE_COMPILE_ERROR);

			if (!b1)
				continue;

			final boolean b2 = pl.getType().equals(Plugin.JAVA_FILE)
					|| pl.getType().equals(Plugin.JAR_FILE);

			if (!b2)
				continue;

			this.plugins.add(pl);
		}

	}

	public Plugin readToPlugin(final File orig) throws Exception {

		final Plugin newPlugin = new Plugin();
		newPlugin.setOriginalFile(orig);
		newPlugin.setLastModified(orig.lastModified());

		if (orig.getName().toLowerCase().endsWith(".java")) {

			newPlugin.setType(Plugin.JAVA_FILE);

			final File newjar = new File(FileIOHandler.getTempJar());
			
			// create the jar file from the java file
			try {
				Cm2_Toolkit.compressToJar(orig, newjar);
			} catch (final Exception e) {
				newPlugin.setState(Plugin.STATE_COMPILE_ERROR);
				return newPlugin;
			}
			
			// To make sure that the writing of the jar-file is finished,
			// this workaround is used. Otherwise LINUX systems may not find
			// the class when using Class.forName() in createMembranePackingAlgorithm()
			Thread.sleep(sleepCounter);
			System.gc();
			System.runFinalization();
			
			// build the MembranePackingAlgorithm and get the class
			try {
				final MembranePackingAlgorithm algo = this
						.createMembranePackingAlgorithm(newPlugin, orig);
				return this.fillPlugin(newPlugin, algo, orig);
			} catch (final Exception e) {
				newPlugin.setState(Plugin.STATE_COMPILE_ERROR);
				e.printStackTrace();
				if (!Cm2_Toolkit.amIRunningOnWebstart()) {
					sleepCounter+=1000;
					JOptionPane.showMessageDialog(null, "This error may be a synchronization problem between Java and your local file system.\n"+
														"In this case, re-opening Membrane Algorithms may fix the problem,\n"+
														"because the Thread Sleep counter was set now to "+sleepCounter+"ms.",
														"Java Synchronization Problem",
														JOptionPane.ERROR_MESSAGE);
				}
				return newPlugin;
			}
			
		// get the jar file, create the MembranePackingAlgorithm and get the class
		} else if (orig.getName().toLowerCase().endsWith(".jar")) {
			newPlugin.setType(Plugin.JAR_FILE);
			try {
				final MembranePackingAlgorithm algo = this
						.createMembranePackingAlgorithm(newPlugin, orig);
				return this.fillPlugin(newPlugin, algo, orig);
			} catch (final Exception e) {
				e.printStackTrace();
				newPlugin.setState(Plugin.STATE_NO_ALGORITHMS);
				return newPlugin;
			}

		} else
			throw new Exception("Invalid file extension");

	}

	public MembranePackingAlgorithm createMembranePackingAlgorithm(final Plugin newPlugin, final File originalFile)
			throws Exception {
		
//		Class<?> c = null;
		
		final File jarfile = (originalFile.getName().toLowerCase().endsWith(".java")) ?
				new File(FileIOHandler.getTempJar()) : originalFile;
		final URL u = jarfile.toURI().toURL();
		
		// ====================================================================================
		// the URLClassLoader needs to get the actual class loader assigned:
		// getClass().getClassLoader()
		// otherwise webstart will not find those classes.
		// if there will be any problems in future with the local version, the
		// loader should be initialized without the actual class loader
		// "URLClassLoader loader = new URLClassLoader(new URL[] { u });"
		// if Cm2_Toolkit.amIRunningOnWebstart() is false
		// sources:
		// (http://www.martinlippert.com/publications/JS-Classloading-in-Eclipse-final-web.pdf)
		// http://java.sun.com/j2se/1.5.0/docs/guide/javaws/developersguide/faq.html#211
		// http://lopica.sourceforge.net/faq.html#customcl
		//
		// another interesting source:
		// http://jimlife.wordpress.com/2007/12/19/java-adding-new-classpath-at-runtime/
		// ====================================================================================
		
		final URLClassLoader loader = new URLClassLoader(new URL[] { u }, this
				.getClass().getClassLoader());
		final Method addURL = URLClassLoader.class.getDeclaredMethod("addURL",
				new Class[] { URL.class });
		addURL.setAccessible(true);
		addURL.invoke(loader, new Object[] { u });

		// ====================================================================================
		// Class.forName() regularly will not find anything using webstart, but
		// because the URLClassLoader AND the actual class loader are added, it works
		// for both, locally and for webstart, see above.
		// ====================================================================================
		
//		try {
//			c = Class.forName(Cm2_Toolkit
//					.getSimpleName(originalFile), true, loader);
//		}catch(Exception e){
//			System.out.println("EXCEPTION originalFile : "+originalFile.getAbsolutePath());
//			System.out.println("EXCEPTION getSimpleName: "+Cm2_Toolkit.getSimpleName(originalFile));
//			System.out.println("EXCEPTION url "+u.toString());
//			System.out.println("EXCEPTION class "+this.getClass().getClassLoader().toString());
//			e.printStackTrace();
//		}
		
		// looking into jar manifest and find the 'MembranePackingAlgorithm-Class' tag
		JarFile realJarFile = null;
		String MPAClassName = null;
		try {
			realJarFile = new JarFile(jarfile.getAbsoluteFile());
			Manifest manifest = realJarFile.getManifest();
			Attributes attributes = manifest.getMainAttributes();
			MPAClassName = attributes.getValue("MembranePackingAlgorithm-Class");
		} catch(Exception e) {
			// handle exception here
		}
		// if the tag was found, the corresponding JarEntry is read out by InputStream
		Class<?> c = null;
		if(MPAClassName != null){
			// -> jar file is a new complex plugin 
			newPlugin.setExtendedPlugin(true);
			JarEntry entry = realJarFile.getJarEntry(MPAClassName + ".class");
			// URL[] urls = loader.getURLs();
			// for(int i = 0; i<urls.length; i++) {
			// 	System.out.println(urls[i].toString());
			// }
			byte[] classBytewise = new byte[(int) entry.getSize()];
			InputStream iStream = realJarFile.getInputStream(entry);
			iStream.read(classBytewise);
			Class<?> cls = Class.forName("java.lang.ClassLoader");
			final Method defineClass = cls.getDeclaredMethod("defineClass", new Class[] { String.class, byte[].class, int.class, int.class });
			defineClass.setAccessible(true);
			try{
				Object[] args = new Object[] { null, classBytewise, new Integer(0), new Integer(classBytewise.length) };
				c = (Class<?>) defineClass.invoke(loader, args);
			} finally {
				defineClass.setAccessible(false);
			}
		} else {
			// -> jar file is a old plugin
			c = Class.forName(Cm2_Toolkit.getSimpleName(originalFile), true, loader);
		}
		
//		final Class<?> c = Class.forName(Cm2_Toolkit
//				.getSimpleName(originalFile), true, loader);
		final MembranePackingAlgorithm algo = (MembranePackingAlgorithm) c.newInstance();
		
//		MembranePackingAlgorithm algo = null;
//		if (c!=null)
//			algo = (MembranePackingAlgorithm) c.newInstance();
		return algo;

	}

	private Plugin fillPlugin(final Plugin newPlugin,
			final MembranePackingAlgorithm algo, final File file) {

		newPlugin.setName(algo.getName());
		newPlugin.setSupportsMDs(algo.supportsMicrodomains());
		newPlugin.setSupportsAbsoluteValues(algo.supportsAbsoluteValues());
		newPlugin.setOverridesLipids(algo.overrideExistingLipids());
		newPlugin.setState(Plugin.STATE_READY);
		newPlugin.setInformation(algo.getFullInformation());
		newPlugin.setOriginalFile(file);
		newPlugin.setLastModified(file.lastModified());
		return newPlugin;
	}

	public boolean isAlreadyIn(final String name, final File file) {
		for (int i = 0; i < this.plugins.size(); i++) {
			if (name.equals(this.plugins.get(i).getName())
					|| this.equalFiles(this.plugins.get(i).getOriginalFile(),
							file))
				return true;
		}
		return false;
	}

	public void writeToDisc() {

		final Vector<Plugin> output = new Vector<Plugin>();
		for (int i = 0; i < this.plugins.size(); i++) {
			final Plugin pl = this.plugins.get(i);
			if (pl.type != null && !pl.type.equals(Plugin.INCLUDED))
				output.add(pl);
		}

		try {
			PluginIO.write(output);
		} catch (final Exception e) {

			System.err.println(e.getMessage());
		}

	}

	public boolean isAlreadyIn(final String name, final File file,
			final Plugin exceptforthis) {
		for (int i = 0; i < this.plugins.size(); i++) {
			if (this.plugins.get(i) != exceptforthis
					&& (name.equals(this.plugins.get(i).getName()) || this
							.equalFiles(this.plugins.get(i).getOriginalFile(),
									file)))
				return true;
		}
		return false;
	}

	private boolean equalFiles(final File file1, final File file2) {
		if (file1 != null
				&& file2 != null
				&& file1.getAbsolutePath().equalsIgnoreCase(
						file2.getAbsolutePath()))
			return true;
		else
			return false;
	}

	public void performUpdate() {

		final Vector<Plugin> finishedPlugins = new Vector<Plugin>();

		for (int i = 0; i < this.plugins.size(); i++) {

			final Plugin plugin = this.plugins.get(i);

			if (Plugin.INCLUDED.equals(plugin.type)) {
				finishedPlugins.add(plugin);
				continue;
			}

			final File file = plugin.getOriginalFile();

			if (file == null || file.exists() == false) {
				plugin.setState(Plugin.STATE_FILE_NOT_FOUND);
				finishedPlugins.add(plugin);
				continue;
			} else if (file.canRead() == false) {
				plugin.setState(Plugin.STATE_FILE_NOT_READABLE);
				finishedPlugins.add(plugin);
				continue;
			} else {
				plugin.setState(Plugin.STATE_READY);
				plugin.setLastModified(0);
			}

			final long realage = file.lastModified();
			if (realage == plugin.getLastModified()) {
				finishedPlugins.add(plugin);
				continue;
			}

			try {

				final Plugin newplugin = this.readToPlugin(plugin
						.getOriginalFile());

				if (Plugin.STATE_READY.equals(newplugin.state) == false)
					newplugin.setName(plugin.getName());

				finishedPlugins.add(newplugin);
			} catch (final Exception e) {
				e.printStackTrace();
			}

		}

		this.plugins.clear();

		final Hashtable<String, String> names = new Hashtable<String, String>();
		final Hashtable<String, String> files = new Hashtable<String, String>();

		for (int i = 0; i < finishedPlugins.size(); i++) {
			final Plugin plugin = finishedPlugins.get(i);

			final String name = plugin.getName();
			if (name == null || name.length() < 1)
				continue;

			if (Plugin.INCLUDED.equals(plugin.type)) {
				this.plugins.add(plugin);
				continue;
			}

			if (names.containsKey(plugin.getName())
					|| files.containsKey(plugin.getOriginalFile()
							.getAbsolutePath()))
				continue;

			names.put(plugin.getName(), "");
			files.put(plugin.getOriginalFile().getAbsolutePath(), "");
			this.plugins.add(plugin);

		}
	}

	public void writeAlgorithmSettings(Plugin plugin, MembranePackingAlgorithm algo)
			throws Exception {

		File destDir = new File(DataPool.workspace.getAbsolutePath()
				+ File.separator + CM2_CONSTANTS.IO.ALGORITHM_SETTINGS.getName()
				+ File.separator + plugin.getName());

		if (!destDir.exists() && !destDir.mkdirs())
			throw new Exception("Cannot create Directories inside Workspace");

		if (plugin.type.equals(Plugin.INCLUDED)) {

			File source = new File(FileIOHandler.home + "cm2settings"
					+ File.separator + plugin.getName());

			if (!source.canRead())
				return;

			File[] files = source.listFiles();

			for (int i = 0; i < files.length; i++) {

				File out = new File(destDir.getAbsolutePath() + File.separator
						+ files[i].getName());

				if (out.exists()) {
					if (out.lastModified() >= files[i].lastModified()) {
						continue;
					}
				}

				final BufferedInputStream bis = new BufferedInputStream(
						new FileInputStream(files[i]));

				final BufferedOutputStream bos = new BufferedOutputStream(
						new FileOutputStream(out));

				final byte[] bytes = new byte[1024];
				while (true) {
					final int bytesread = bis.read(bytes);
					if (bytesread < 0)
						break;
					bos.write(bytes, 0, bytesread);
				}

				bis.close();
				bos.close();
			}

		} else if (plugin.type.equals(Plugin.JAR_FILE)) {

			File source = plugin.getOriginalFile();

			if (!source.canRead())
				return;
			try {
				Cm2_Toolkit.extractFromJar(source.getAbsolutePath(), destDir
						.getAbsolutePath()
						+ "temp");
			} catch (Exception e) {
				e.printStackTrace();
			}

			File file1 = new File(destDir.getAbsolutePath() + "temp"
					+ File.separator + "META-INF");
			File file2 = new File(destDir.getAbsolutePath() + "temp"
					+ File.separator + ".project");
			File file3 = new File(destDir.getAbsolutePath() + "temp"
					+ File.separator + ".classpath");

			file1.delete();
			file2.delete();
			file3.delete();

			file1.deleteOnExit();
			file2.deleteOnExit();
			file3.deleteOnExit();

			File[] files = new File(destDir.getAbsolutePath() + "temp")
					.listFiles();

			for (int i = 0; i < files.length; i++) {

				if (files[i].getName().endsWith(".class")) {
					files[i].delete();
					files[i].deleteOnExit();
					continue;
				}

				File out = new File(destDir.getAbsolutePath() + File.separator
						+ files[i].getName());

				if (out.exists()) {
					if (out.lastModified() >= files[i].lastModified()) {
						continue;
					}
				}

				final BufferedInputStream bis = new BufferedInputStream(
						new FileInputStream(files[i]));

				final BufferedOutputStream bos = new BufferedOutputStream(
						new FileOutputStream(out));

				final byte[] bytes = new byte[1024];
				while (true) {
					final int bytesread = bis.read(bytes);
					if (bytesread < 0)
						break;
					bos.write(bytes, 0, bytesread);
				}

				bis.close();
				bos.close();
			}

			Cm2_Toolkit.deleteDir(new File(destDir.getAbsolutePath() + "temp"),
					true);

		} else
			return;

	}

}
