/*******************************************************************************
 * CELLmicrocosmos 2.2 MembraneEditor (version 2.2.2_0)
 *
 * Copyright (c) 2005, 2014 The CELLmicrocosmos 2 Development Team
 *                          (Bjrn Sommer et al., Bielefeld University)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package org.cellmicrocosmos.cm2.membranepackingalgorithm;

import javax.swing.JComponent;

/**
 * This class must be overwritten in oder to create an own Enegry Functin for
 * the Atom_level_minimizing Algorithm.
 * 
 * To enable it as a plugin, wrap its binaries into a .jar File. The .jar File
 * must have an equal name (without any File extensions) to the name chosen for
 * this class. Copy the jar-File into the plugin-directory of the
 * Atom_level_minimizing Algorithm's folder inside your user Workspace.
 * 
 * @author Tim Dingersen dackelnose@gmx.net
 * 
 */
public abstract class EnergyFunction {

	/**
	 * Return the Energy between the pair of passed Atoms. You dont't have to
	 * care about periodic borders here, this will be handled automatically by
	 * the algorithm.
	 * 
	 * The distance is the atomDistance in angstrom () no need to calculate this
	 * one twice
	 * 
	 * @param referenceAtom
	 *            the first atom
	 * @param otherAtom
	 *            the second atom
	 * @param bondFactor
	 *            the constraintFactor set in the controlPanel
	 * @param distance
	 *            the distance between these atoms
	 * @param currentTemperature
	 *            the current temperature of the simulated annealing process
	 * @return the result of your Energy function
	 */
	public abstract double calculateEnergy(Atom referenceAtom, Atom otherAtom,
			double bondFactor, float distance, float currentTemperature);

	/**
	 * Return a scale factor for your Energy Values. Can be used to refine the
	 * functionality. Make sure to return a value>0, or 1 if you dont want to
	 * scale anything
	 * 
	 * @return the desired scale factor or 1 if unscaled
	 */
	public abstract double getScaleFactor();

	/**
	 * Return the name for the RadioButton of this Algortihm here
	 * 
	 * @return any string that is not empty
	 */
	public abstract String getButtonName();

	/**
	 * Return a tooltip for this algorithm's Button
	 * 
	 * @return any string that is not empty or null for no tooltip
	 */
	public abstract String getTooltip();

	/**
	 * Here you may return the gui for this. It can be opened by the user during
	 * runtime.
	 * 
	 * @return your specific JComponent or null if no Gui is available
	 */
	public abstract JComponent getGui();

	/**
	 * Gets called at startup and passes a reference to the underLying
	 * MembraneAlgortihm
	 * 
	 * @param algo
	 *            the instance of the currently performing MembranePackingAlgorithm
	 */
	public abstract void setMembranePackingAlgorithm(MembranePackingAlgorithm algo);

	/**
	 * Gets called everytime when any atoms have been added, or removed from the
	 * model during runtime
	 * 
	 * @param allAtoms
	 *            the new list of atoms that should be used by now
	 */
	public abstract void MembersUpdated(Atom[] allAtoms);

	/**
	 * Used to get the cutoff between this pair of atoms if "Autocutoff" is
	 * activated. You can return the defaultCutoff if you don't want to specify
	 * individual cutoffs
	 * 
	 * @param reference
	 *            the first atom
	 * @param other
	 *            the second atom
	 * @return the specified cutoff distance
	 */
	public abstract float getCutoff(Atom reference, Atom other,
			float defaultCutoff);

	/**
	 * Should return the lowest value possible that could be a result of the
	 * calculateEnergy method. This is vital for proper functioning. If you
	 * should only return positive values, then you may return 0 here.
	 * 
	 * @return the lowetst possible value of calculateEnergy
	 */
	public abstract double getLowestPossibleReturnValue();

}
